@@ -1,4 +1,4 @@
-/* $Id: Str.c,v 1.7 2001/12/21 21:37:12 ukai Exp $ */
+/* vi: set sw=4 ts=8 ai sm noet : */
 /* 
  * String manipulation library for Boehm GC
  *
@@ -15,7 +15,7 @@
  */
 #include <stdio.h>
 #include <stdlib.h>
-#include <gc.h>
+#include <gc/gc.h>
 #include <stdarg.h>
 #include <string.h>
 #ifdef __EMX__			/* or include "fm.h" for HAVE_BCOPY? */
@@ -25,55 +25,45 @@
 #include "myctype.h"
 
 #define INITIAL_STR_SIZE 32
+#define STR_SIZE_MAX (STR_LEN_MAX + 1)
 
-#ifdef STR_DEBUG
-/* This is obsolete, because "Str" can handle a '\0' character now. */
-#define STR_LENGTH_CHECK(x) if (((x)->ptr==0&&(x)->length!=0)||(strlen((x)->ptr)!=(x)->length))abort();
-#else				/* not STR_DEBUG */
-#define STR_LENGTH_CHECK(x)
-#endif				/* not STR_DEBUG */
+static Str Strgrow_n(Str s, int n);
 
 Str
-Strnew()
+Strnew(void)
 {
-    Str x = GC_MALLOC(sizeof(struct _Str));
-    x->ptr = GC_MALLOC_ATOMIC(INITIAL_STR_SIZE);
-    x->ptr[0] = '\0';
-    x->area_size = INITIAL_STR_SIZE;
-    x->length = 0;
-    return x;
+    return Strnew_size(INITIAL_STR_SIZE - 1);
 }
 
 Str
 Strnew_size(int n)
-{
-    Str x = GC_MALLOC(sizeof(struct _Str));
-    x->ptr = GC_MALLOC_ATOMIC(n + 1);
-    x->ptr[0] = '\0';
-    x->area_size = n + 1;
-    x->length = 0;
-    return x;
-}
-
-Str
-Strnew_charp(char *p)
 {
     Str x;
-    int n;
 
-    if (p == NULL)
-	return Strnew();
-    x = GC_MALLOC(sizeof(struct _Str));
-    n = strlen(p) + 1;
-    x->ptr = GC_MALLOC_ATOMIC(n);
+    if (n < 0 || n > STR_LEN_MAX)
+	n = STR_SIZE_MAX;
+    else if (n + 1 < INITIAL_STR_SIZE)
+	n = INITIAL_STR_SIZE;
+    else
+	n++;
+
+    if (!(x = GC_MALLOC(sizeof(struct _Str)))
+	|| !(x->ptr = GC_MALLOC_ATOMIC(n)))
+	exit(1);
     x->area_size = n;
-    x->length = n - 1;
-    bcopy((void *)p, (void *)x->ptr, n);
+    x->length = 0;
+    x->ptr[x->length] = '\0';
     return x;
 }
 
 Str
-Strnew_m_charp(char *p, ...)
+Strnew_charp(const char *p)
+{
+    return p ? Strnew_charp_n(p, strlen(p)) : Strnew();
+}
+
+Str
+Strnew_m_charp(const char *p, ...)
 {
     va_list ap;
     Str r = Strnew();
@@ -83,22 +73,18 @@ Strnew_m_charp(char *p, ...)
 	Strcat_charp(r, p);
 	p = va_arg(ap, char *);
     }
+    va_end(ap);
     return r;
 }
 
 Str
-Strnew_charp_n(char *p, int n)
+Strnew_charp_n(const char *p, int n)
 {
     Str x;
 
-    if (p == NULL)
-	return Strnew_size(n);
-    x = GC_MALLOC(sizeof(struct _Str));
-    x->ptr = GC_MALLOC_ATOMIC(n + 1);
-    x->area_size = n + 1;
-    x->length = n;
-    bcopy((void *)p, (void *)x->ptr, n);
-    x->ptr[n] = '\0';
+    x = Strnew_size(n);
+    if (p)
+	Strcopy_charp_n(x, p, n);
     return x;
 }
 
@@ -106,117 +92,99 @@ Str
 Strdup(Str s)
 {
     Str n = Strnew_size(s->length);
-    STR_LENGTH_CHECK(s);
     Strcopy(n, s);
     return n;
 }
 
-void
+Str
 Strclear(Str s)
 {
     s->length = 0;
-    s->ptr[0] = '\0';
+    s->ptr[s->length] = '\0';
+    return s;
 }
 
 void
 Strfree(Str x)
 {
+    if (!x) return;
     GC_free(x->ptr);
     GC_free(x);
 }
 
-void
-Strcopy(Str x, Str y)
+Str
+Strcopy(Str dst, Str src)
 {
-    STR_LENGTH_CHECK(x);
-    STR_LENGTH_CHECK(y);
-    if (x->area_size < y->length + 1) {
-	GC_free(x->ptr);
-	x->ptr = GC_MALLOC_ATOMIC(y->length + 1);
-	x->area_size = y->length + 1;
-    }
-    bcopy((void *)y->ptr, (void *)x->ptr, y->length + 1);
-    x->length = y->length;
+    return Strcopy_charp_n(dst, src->ptr, src->length);
 }
 
-void
-Strcopy_charp(Str x, char *y)
+Str
+Strcopy_charp(Str x, const char *y)
 {
     int len;
 
-    STR_LENGTH_CHECK(x);
-    if (y == NULL) {
-	x->length = 0;
-	return;
-    }
+    if (!y)
+	return Strtruncate(x, 0);
+
     len = strlen(y);
-    if (x->area_size < len + 1) {
-	GC_free(x->ptr);
-	x->ptr = GC_MALLOC_ATOMIC(len + 1);
-	x->area_size = len + 1;
-    }
-    bcopy((void *)y, (void *)x->ptr, len + 1);
-    x->length = len;
+    return Strcopy_charp_n(x, y, len);
 }
 
-void
-Strcopy_charp_n(Str x, char *y, int n)
+Str
+Strcopy_charp_n(Str x, const char *y, int n)
 {
-    int len = n;
+    if (!y)
+	return Strtruncate(x, 0);
 
-    STR_LENGTH_CHECK(x);
-    if (y == NULL) {
-	x->length = 0;
-	return;
-    }
-    if (x->area_size < len + 1) {
-	GC_free(x->ptr);
-	x->ptr = GC_MALLOC_ATOMIC(len + 1);
-	x->area_size = len + 1;
-    }
-    bcopy((void *)y, (void *)x->ptr, n);
-    x->ptr[n] = '\0';
+    if (n > STR_LEN_MAX)
+	n = STR_LEN_MAX;
+    if (x->area_size <= n)
+	Strgrow_n(x, n);
+    memmove(x->ptr, y, n);
     x->length = n;
+    x->ptr[x->length] = '\0';
+    return x;
 }
 
-void
-Strcat_charp_n(Str x, char *y, int n)
+Str
+Strcat_charp_n(Str x, const char *y, int n)
 {
     int newlen;
 
-    STR_LENGTH_CHECK(x);
-    if (y == NULL)
-	return;
-    newlen = x->length + n + 1;
-    if (x->area_size < newlen) {
-	char *old = x->ptr;
-	newlen = newlen * 3 / 2;
-	x->ptr = GC_MALLOC_ATOMIC(newlen);
-	x->area_size = newlen;
-	bcopy((void *)old, (void *)x->ptr, x->length);
-	GC_free(old);
+    if (!y || !n || x->length == STR_LEN_MAX)
+	return x;
+
+    if (n < 0)
+	n = strlen(y);
+    newlen = x->length + n;
+    if (newlen > STR_LEN_MAX) {
+	newlen = STR_LEN_MAX;
+	n = STR_LEN_MAX - x->length;
     }
-    bcopy((void *)y, (void *)&x->ptr[x->length], n);
+
+    if (newlen >= x->area_size)
+	Strgrow_n(x, newlen);
+    memmove(&x->ptr[x->length], y, n);
     x->length += n;
     x->ptr[x->length] = '\0';
+    return x;
 }
 
-void
+Str
 Strcat(Str x, Str y)
 {
-    STR_LENGTH_CHECK(y);
-    Strcat_charp_n(x, y->ptr, y->length);
+    return Strcat_charp_n(x, y->ptr, y->length);
 }
 
-void
-Strcat_charp(Str x, char *y)
+Str
+Strcat_charp(Str x, const char *y)
 {
-    if (y == NULL)
-	return;
-    Strcat_charp_n(x, y, strlen(y));
+    if (!y)
+	return x;
+    return Strcat_charp_n(x, y, strlen(y));
 }
 
-void
+Str
 Strcat_m_charp(Str x, ...)
 {
     va_list ap;
@@ -225,20 +193,45 @@ Strcat_m_charp(Str x, ...)
     va_start(ap, x);
     while ((p = va_arg(ap, char *)) != NULL)
 	 Strcat_charp_n(x, p, strlen(p));
+    va_end(ap);
+    return x;
 }
 
-void
+Str
+Strgrow_n(Str x, int n)
+{
+    if (n < 0)
+	n = STR_SIZE_MAX;
+    else
+	n = (n >= STR_SIZE_MAX) ? STR_SIZE_MAX : n + 1;
+
+    if (x->area_size >= n)
+	return x;
+
+    if (!(x->ptr = GC_REALLOC(x->ptr, n)))
+	exit(1);
+    x->area_size = n;
+    return x;
+}
+
+Str
 Strgrow(Str x)
 {
-    char *old = x->ptr;
-    int newlen;
-    newlen = x->length * 6 / 5;
-    if (newlen == x->length)
-	newlen += 2;
-    x->ptr = GC_MALLOC_ATOMIC(newlen);
-    x->area_size = newlen;
-    bcopy((void *)old, (void *)x->ptr, x->length);
-    GC_free(old);
+    int newlen, addlen;
+
+    if (x->area_size < 8192)
+	addlen = x->area_size;
+    else
+	addlen = x->area_size / 2;
+    if (addlen < INITIAL_STR_SIZE)
+	addlen = INITIAL_STR_SIZE;
+    newlen = x->area_size + addlen;
+    if (newlen <= 0 || newlen > STR_SIZE_MAX) {
+	newlen = STR_SIZE_MAX;
+	if (x->length + 1 >= newlen)
+	    x->length = newlen - 2;
+    }
+    return Strgrow_n(x, newlen - 1);
 }
 
 Str
@@ -247,7 +240,6 @@ Strsubstr(Str s, int beg, int len)
     Str new_s;
     int i;
 
-    STR_LENGTH_CHECK(s);
     new_s = Strnew();
     if (beg >= s->length)
 	return new_s;
@@ -256,117 +248,125 @@ Strsubstr(Str s, int beg, int len)
     return new_s;
 }
 
-void
+Str
 Strlower(Str s)
 {
     int i;
-    STR_LENGTH_CHECK(s);
     for (i = 0; i < s->length; i++)
-	s->ptr[i] = tolower(s->ptr[i]);
+	s->ptr[i] = TOLOWER(s->ptr[i]);
+    return s;
 }
 
-void
+Str
 Strupper(Str s)
 {
     int i;
-    STR_LENGTH_CHECK(s);
     for (i = 0; i < s->length; i++)
-	s->ptr[i] = toupper(s->ptr[i]);
+	s->ptr[i] = TOUPPER(s->ptr[i]);
+    return s;
 }
 
-void
+Str
 Strchop(Str s)
 {
-    STR_LENGTH_CHECK(s);
-    while ((s->ptr[s->length - 1] == '\n' || s->ptr[s->length - 1] == '\r') &&
-	   s->length > 0) {
+    while (s->length > 0 &&
+	   (s->ptr[s->length - 1] == '\n' || s->ptr[s->length - 1] == '\r')) {
 	s->length--;
     }
     s->ptr[s->length] = '\0';
+    return s;
 }
 
-void
+Str
 Strinsert_char(Str s, int pos, char c)
 {
     int i;
-    STR_LENGTH_CHECK(s);
-    if (pos < 0 || s->length < pos)
-	return;
-    if (s->length + 2 > s->area_size)
+    if (pos < 0 || s->length < pos || s->length == STR_LEN_MAX)
+	return s;
+    if (s->length + 1 >= s->area_size)
 	Strgrow(s);
     for (i = s->length; i > pos; i--)
 	s->ptr[i] = s->ptr[i - 1];
-    s->ptr[++s->length] = '\0';
+    s->length++;
+    s->ptr[s->length] = '\0';
     s->ptr[pos] = c;
+    return s;
 }
 
-void
-Strinsert_charp(Str s, int pos, char *p)
+Str
+Strinsert_charp_n(Str s, int pos, const char *p, int n)
+{
+    while (n--)
+	Strinsert_char(s, pos++, *(p++));
+    return s;
+}
+
+Str
+Strinsert_charp(Str s, int pos, const char *p)
 {
-    STR_LENGTH_CHECK(s);
     while (*p)
 	Strinsert_char(s, pos++, *(p++));
+    return s;
 }
 
-void
+Str
 Strdelete(Str s, int pos, int n)
 {
     int i;
-    STR_LENGTH_CHECK(s);
+    if (pos < 0 || s->length < pos)
+	return s;
+    if (n < 0)
+	n = STR_LEN_MAX - pos;
     if (s->length <= pos + n) {
-	s->ptr[pos] = '\0';
 	s->length = pos;
-	return;
+	s->ptr[s->length] = '\0';
+	return s;
     }
     for (i = pos; i < s->length - n; i++)
 	s->ptr[i] = s->ptr[i + n];
-    s->ptr[i] = '\0';
     s->length = i;
+    s->ptr[s->length] = '\0';
+    return s;
 }
 
-void
+Str
 Strtruncate(Str s, int pos)
 {
-    STR_LENGTH_CHECK(s);
-    s->ptr[pos] = '\0';
+    if (pos < 0 || s->length < pos)
+	return s;
     s->length = pos;
+    s->ptr[s->length] = '\0';
+    return s;
 }
 
-void
+Str
 Strshrink(Str s, int n)
 {
-    STR_LENGTH_CHECK(s);
-    if (n >= s->length) {
-	s->length = 0;
-	s->ptr[0] = '\0';
-    }
-    else {
-	s->length -= n;
-	s->ptr[s->length] = '\0';
-    }
+    s->length = (n >= s->length) ? 0 : (s->length - n);
+    s->ptr[s->length] = '\0';
+    return s;
 }
 
-void
+Str
 Strremovefirstspaces(Str s)
 {
     int i;
 
-    STR_LENGTH_CHECK(s);
     for (i = 0; i < s->length && IS_SPACE(s->ptr[i]); i++) ;
     if (i == 0)
-	return;
-    Strdelete(s, 0, i);
+	return s;
+    return Strdelete(s, 0, i);
 }
 
-void
+Str
 Strremovetrailingspaces(Str s)
 {
     int i;
 
-    STR_LENGTH_CHECK(s);
     for (i = s->length - 1; i >= 0 && IS_SPACE(s->ptr[i]); i--) ;
     s->length = i + 1;
-    s->ptr[i + 1] = '\0';
+    s->ptr[s->length] = '\0';
+    return s;
 }
 
 Str
@@ -375,7 +375,6 @@ Stralign_left(Str s, int width)
     Str n;
     int i;
 
-    STR_LENGTH_CHECK(s);
     if (s->length >= width)
 	return Strdup(s);
     n = Strnew_size(width);
@@ -391,7 +390,6 @@ Stralign_right(Str s, int width)
     Str n;
     int i;
 
-    STR_LENGTH_CHECK(s);
     if (s->length >= width)
 	return Strdup(s);
     n = Strnew_size(width);
@@ -407,7 +405,6 @@ Stralign_center(Str s, int width)
     Str n;
     int i, w;
 
-    STR_LENGTH_CHECK(s);
     if (s->length >= width)
 	return Strdup(s);
     n = Strnew_size(width);
@@ -425,12 +422,12 @@ Stralign_center(Str s, int width)
 #define SP_PREC2  2
 
 Str
-Sprintf(char *fmt, ...)
+Sprintf(const char *fmt, ...)
 {
     int len = 0;
     int status = SP_NORMAL;
     int p = 0;
-    char *f;
+    const char *f;
     Str s;
     va_list ap;
 
@@ -449,10 +446,8 @@ Sprintf(char *fmt, ...)
 	case SP_PREC:
 	    if (IS_ALPHA(*f)) {
 		/* conversion char. */
-		double vd;
 		int vi;
 		char *vs;
-		void *vp;
 
 		switch (*f) {
 		case 'l':
@@ -466,7 +461,7 @@ Sprintf(char *fmt, ...)
 		case 'x':
 		case 'X':
 		case 'u':
-		    vi = va_arg(ap, int);
+		    va_arg(ap, int);
 		    len += (p > 0) ? p : 10;
 		    break;
 		case 'f':
@@ -474,12 +469,12 @@ Sprintf(char *fmt, ...)
 		case 'e':
 		case 'G':
 		case 'E':
-		    vd = va_arg(ap, double);
+		    va_arg(ap, double);
 		    len += (p > 0) ? p : 15;
 		    break;
 		case 'c':
 		    len += 1;
-		    vi = va_arg(ap, int);
+		    va_arg(ap, int);
 		    break;
 		case 's':
 		    vs = va_arg(ap, char *);
@@ -487,11 +482,11 @@ Sprintf(char *fmt, ...)
 		    len += (p > vi) ? p : vi;
 		    break;
 		case 'p':
-		    vp = va_arg(ap, void *);
+		    va_arg(ap, void *);
 		    len += 10;
 		    break;
 		case 'n':
-		    vp = va_arg(ap, void *);
+		    va_arg(ap, void *);
 		    break;
 		}
 		status = SP_NORMAL;
@@ -530,11 +525,8 @@ Str
 Strfgets(FILE * f)
 {
     Str s = Strnew();
-    char c;
-    while (1) {
-	c = fgetc(f);
-	if (feof(f) || ferror(f))
-	    break;
+    int c;
+    while ((c = fgetc(f)) != EOF) {
 	Strcat_char(s, c);
 	if (c == '\n')
 	    break;
@@ -546,11 +538,8 @@ Str
 Strfgetall(FILE * f)
 {
     Str s = Strnew();
-    char c;
-    while (1) {
-	c = fgetc(f);
-	if (feof(f) || ferror(f))
-	    break;
+    int c;
+    while ((c = fgetc(f)) != EOF) {
 	Strcat_char(s, c);
     }
     return s;
