@@ -1,8 +1,10 @@
-/* $Id: anchor.c,v 1.33 2006/04/08 11:33:16 inu Exp $ */
+/* vi: set sw=4 ts=8 ai sm noet : */
 #include "fm.h"
 #include "myctype.h"
 #include "regex.h"
 
+#include <strings.h>
+
 #define FIRST_ANCHOR_SIZE 30
 
 AnchorList *
@@ -11,7 +13,7 @@ putAnchor(AnchorList *al, char *url, char *target, Anchor **anchor_return,
 {
     int n, i, j;
     Anchor *a;
-    BufferPoint bp;
+    BufferPoint bp = { 0 };
     if (al == NULL) {
 	al = New(AnchorList);
 	al->anchors = NULL;
@@ -191,7 +193,7 @@ searchURLLabel(Buffer *buf, char *url)
 
 #ifdef USE_NNTP
 static Anchor *
-_put_anchor_news(Buffer *buf, char *p1, char *p2, int line, int pos)
+_put_anchor_news(Buffer *buf, const char *p1, const char *p2, int line, int pos)
 {
     Str tmp;
 
@@ -200,27 +202,29 @@ _put_anchor_news(Buffer *buf, char *p1, char *p2, int line, int pos)
 	if (*(p2 - 1) == '>')
 	    p2--;
     }
-    tmp = wc_Str_conv_strict(Strnew_charp_n(p1, p2 - p1), InnerCharset,
-			     buf->document_charset);
-    tmp = Sprintf("news:%s", file_quote(tmp->ptr));
-    return registerHref(buf, tmp->ptr, NULL, NO_REFERER, NULL, '\0', line,
+    tmp = Strnew_charp("news:");
+    Strcat_charp_n(tmp, p1, p2 - p1);
+    return registerHref(buf, url_encode(tmp->ptr, baseURL(buf),
+					buf->document_charset),
+			NULL, NO_REFERER, NULL, '\0', line,
 			pos);
 }
 #endif				/* USE_NNTP */
 
 static Anchor *
-_put_anchor_all(Buffer *buf, char *p1, char *p2, int line, int pos)
+_put_anchor_all(Buffer *buf, const char *p1, const char *p2, int line, int pos)
 {
     Str tmp;
 
-    tmp = wc_Str_conv_strict(Strnew_charp_n(p1, p2 - p1), InnerCharset,
-			     buf->document_charset);
-    return registerHref(buf, url_quote(tmp->ptr), NULL, NO_REFERER, NULL,
+    tmp = Strnew_charp_n(p1, p2 - p1);
+    return registerHref(buf, url_encode(tmp->ptr, baseURL(buf),
+					buf->document_charset),
+			NULL, NO_REFERER, NULL,
 			'\0', line, pos);
 }
 
 static void
-reseq_anchor0(AnchorList *al, short *seqmap)
+reseq_anchor0(AnchorList *al, const short *seqmap)
 {
     int i;
     Anchor *a;
@@ -292,9 +296,9 @@ reseq_anchor(Buffer *buf)
     reseq_anchor0(buf->formitem, seqmap);
 }
 
-static char *
-reAnchorPos(Buffer *buf, Line *l, char *p1, char *p2,
-	    Anchor *(*anchorproc) (Buffer *, char *, char *, int, int))
+static const char *
+reAnchorPos(Buffer *buf, Line *l, const char *p1, const char *p2,
+	    Anchor *(*anchorproc) (Buffer *, const char *, const char *, int, int))
 {
     Anchor *a;
     int spos, epos;
@@ -343,12 +347,12 @@ reAnchorWord(Buffer *buf, Line *l, int spos, int epos)
 
 /* search regexp and register them as anchors */
 /* returns error message if any               */
-static char *
-reAnchorAny(Buffer *buf, char *re,
-	    Anchor *(*anchorproc) (Buffer *, char *, char *, int, int))
+static const char *
+reAnchorAny(Buffer *buf, const char *re,
+	    Anchor *(*anchorproc) (Buffer *, const char *, const char *, int, int))
 {
     Line *l;
-    char *p = NULL, *p1, *p2;
+    const char *p = NULL, *p1, *p2;
 
     if (re == NULL || *re == '\0') {
 	return NULL;
@@ -378,14 +382,14 @@ reAnchorAny(Buffer *buf, char *re,
     return NULL;
 }
 
-char *
+const char *
 reAnchor(Buffer *buf, char *re)
 {
     return reAnchorAny(buf, re, _put_anchor_all);
 }
 
 #ifdef USE_NNTP
-char *
+const char *
 reAnchorNews(Buffer *buf, char *re)
 {
     return reAnchorAny(buf, re, _put_anchor_news);
@@ -395,7 +399,7 @@ char *
 reAnchorNewsheader(Buffer *buf)
 {
     Line *l;
-    char *p, *p1, *p2;
+    const char *p, *p1, *p2;
     static char *header_mid[] = {
 	"Message-Id:", "References:", "In-Reply-To:", NULL
     };
@@ -549,7 +553,8 @@ shiftAnchorPosition(AnchorList *al, HmarkerList *hl, int line, int pos,
 	    break;
 	if (a->start.pos > pos) {
 	    a->start.pos += shift;
-	    if (hl->marks[a->hseq].line == line)
+	    if (hl && hl->marks &&
+		a->hseq >= 0 && hl->marks[a->hseq].line == line)
 		hl->marks[a->hseq].pos = a->start.pos;
 	}
 	if (a->end.pos >= pos)
@@ -611,6 +616,13 @@ addMultirowsImg(Buffer *buf, AnchorList *al)
 	    a->slave = TRUE;
 	    a->image = img;
 	    a->end.pos = pos + ecol - col;
+	    /* TODO:
+	     * This is a hack to avoid adding images positioned on the wrong
+	     * line outside line bounds (which would cause a buffer overrun).
+	     * The actual bug is most likely in the image placement code.
+	     */
+	    if (pos < 0 || a->end.pos > l->size)
+		continue;
 	    for (k = pos; k < a->end.pos; k++)
 		l->propBuf[k] |= PE_IMAGE;
 	    if (a_href.url) {
@@ -641,7 +653,6 @@ addMultirowsForm(Buffer *buf, AnchorList *al)
 {
     int i, j, k, col, ecol, pos;
     Anchor a_form, *a;
-    FormItemList *fi;
     Line *l, *ls;
 
     if (al == NULL || al->nanchor == 0)
@@ -668,7 +679,6 @@ addMultirowsForm(Buffer *buf, AnchorList *al)
 	    if (!ls)
 		continue;
 	}
-	fi = (FormItemList *)a_form.url;
 	col = COLPOS(ls, a_form.start.pos);
 	ecol = COLPOS(ls, a_form.end.pos);
 	for (j = 0; l && j < a_form.rows; l = l->next, j++) {
@@ -685,6 +695,8 @@ addMultirowsForm(Buffer *buf, AnchorList *al)
 	    a->hseq = a_form.hseq;
 	    a->y = a_form.y;
 	    a->end.pos = pos + ecol - col;
+	    if (pos < 1 || a->end.pos >= l->size)
+		continue;
 	    l->lineBuf[pos - 1] = '[';
 	    l->lineBuf[a->end.pos] = ']';
 	    for (k = pos; k < a->end.pos; k++)
@@ -756,7 +768,7 @@ link_list_panel(Buffer *buf)
 		p = parsedURL2Str(&pu)->ptr;
 		u = html_quote(p);
 		if (DecodeURL)
-		    p = html_quote(url_unquote_conv(p, buf->document_charset));
+		    p = html_quote(url_decode2(p, buf));
 		else
 		    p = u;
 	    }
@@ -787,7 +799,7 @@ link_list_panel(Buffer *buf)
 	    p = parsedURL2Str(&pu)->ptr;
 	    u = html_quote(p);
 	    if (DecodeURL)
-		p = html_quote(url_unquote_conv(p, buf->document_charset));
+		p = html_quote(url_decode2(p, buf));
 	    else
 		p = u;
 	    t = getAnchorText(buf, al, a);
@@ -809,16 +821,13 @@ link_list_panel(Buffer *buf)
 	    p = parsedURL2Str(&pu)->ptr;
 	    u = html_quote(p);
 	    if (DecodeURL)
-		p = html_quote(url_unquote_conv(p, buf->document_charset));
+		p = html_quote(url_decode2(p, buf));
 	    else
 		p = u;
 	    if (a->title && *a->title)
 		t = html_quote(a->title);
-	    else if (DecodeURL)
-		t = html_quote(url_unquote_conv
-			       (a->url, buf->document_charset));
 	    else
-		t = html_quote(a->url);
+		t = html_quote(url_decode2(a->url, buf));
 	    Strcat_m_charp(tmp, "<li><a href=\"", u, "\">", t, "</a><br>", p,
 			   "\n", NULL);
 	    a = retrieveAnchor(buf->formitem, a->start.line, a->start.pos);
@@ -842,19 +851,13 @@ link_list_panel(Buffer *buf)
 		    p = parsedURL2Str(&pu)->ptr;
 		    u = html_quote(p);
 		    if (DecodeURL)
-			p = html_quote(url_unquote_conv(p,
-							buf->
-							document_charset));
+			p = html_quote(url_decode2(p, buf));
 		    else
 			p = u;
 		    if (m->alt && *m->alt)
 			t = html_quote(m->alt);
-		    else if (DecodeURL)
-			t = html_quote(url_unquote_conv(m->url,
-							buf->
-							document_charset));
 		    else
-			t = html_quote(m->url);
+			t = html_quote(url_decode2(m->url, buf));
 		    Strcat_m_charp(tmp, "<li><a href=\"", u, "\">", t,
 				   "</a><br>", p, "\n", NULL);
 		}
