@@ -1,12 +1,33 @@
-/* $Id: buffer.c,v 1.30 2010/07/18 14:10:09 htrb Exp $ */
+/* vi: set sw=4 ts=8 ai sm noet : */
+#include "buffer.h"
+
+#include "Str.h"
+#include "alloc.h"
+#include "config.h"
+#include "ctrlcode.h"
 #include "fm.h"
+#include "html.h"
+#include "terms.h"
+#include "util.h"
+
+#ifdef USE_M17N
+#include "wc.h"
+#include "wc_types.h"
+#endif
+
+#include <stdio.h>
+#include <stdlib.h>
+#include <string.h>
+#include <strings.h>
+#include <unistd.h>
+
 
 #ifdef USE_MOUSE
 #ifdef USE_GPM
 #include <gpm.h>
 #endif
 #if defined(USE_GPM) || defined(USE_SYSMOUSE)
-extern int do_getch();
+extern int do_getch(void);
 #define getch()	do_getch()
 #endif				/* USE_GPM */
 #endif				/* USE_MOUSE */
@@ -15,7 +36,8 @@ extern int do_getch();
 #include <sys/kbdscan.h>
 #include <strings.h>
 #endif
-char *NullLine = "";
+
+const char *NullLine = "";
 Lineprop NullProp[] = { 0 };
 
 /* 
@@ -28,8 +50,8 @@ newBuffer(int width)
 
     n = New(Buffer);
     if (n == NULL)
-	return NULL;
-    bzero((void *)n, sizeof(Buffer));
+	exit(3);
+    bzero(n, sizeof(Buffer));
     n->width = width;
     n->COLS = COLS;
     n->LINES = LASTLINE;
@@ -47,6 +69,8 @@ newBuffer(int width)
 #ifdef USE_M17N
     n->auto_detect = WcOption.auto_detect;
 #endif
+    n->check_url = MarkAllPages; /* use default from -o mark_all_pages */
+    n->need_reshape = MarkAllPages;
     return n;
 }
 
@@ -114,25 +138,6 @@ discardBuffer(Buffer *buf)
     }
 }
 
-/* 
- * namedBuffer: Select buffer which have specified name
- */
-Buffer *
-namedBuffer(Buffer *first, char *name)
-{
-    Buffer *buf;
-
-    if (!strcmp(first->buffername, name)) {
-	return first;
-    }
-    for (buf = first; buf->nextBuffer != NULL; buf = buf->nextBuffer) {
-	if (!strcmp(buf->nextBuffer->buffername, name)) {
-	    return buf->nextBuffer;
-	}
-    }
-    return NULL;
-}
-
 /* 
  * deleteBuffer: delete buffer
  */
@@ -171,7 +176,7 @@ replaceBuffer(Buffer *first, Buffer *delbuf, Buffer *newbuf)
 	discardBuffer(delbuf);
 	return newbuf;
     }
-    if (delbuf && (buf = prevBuffer(first, delbuf))) {
+    if ((buf = prevBuffer(first, delbuf))) {
 	buf->nextBuffer = newbuf;
 	newbuf->nextBuffer = delbuf->nextBuffer;
 	discardBuffer(delbuf);
@@ -207,8 +212,7 @@ writeBufferName(Buffer *buf, int n)
     if (all == 0 && buf->lastLine != NULL)
 	all = buf->lastLine->linenumber;
     move(n, 0);
-    /* FIXME: gettextize? */
-    msg = Sprintf("<%s> [%d lines]", buf->buffername, all);
+    msg = Sprintf(N_("<%s> [%d lines]"), buf->buffername, all);
     if (buf->filename != NULL) {
 	switch (buf->currentURL.scheme) {
 	case SCM_LOCAL:
@@ -237,7 +241,7 @@ writeBufferName(Buffer *buf, int n)
 void
 gotoLine(Buffer *buf, int n)
 {
-    char msg[32];
+    char msg[36];
     Line *l = buf->firstLine;
 
     if (l == NULL)
@@ -249,16 +253,14 @@ gotoLine(Buffer *buf, int n)
 	       (getNextPage(buf, 1) != NULL)) ;
     }
     if (l->linenumber > n) {
-	/* FIXME: gettextize? */
-	sprintf(msg, "First line is #%ld", l->linenumber);
+	sprintf(msg, _("First line is #%ld"), l->linenumber);
 	set_delayed_message(msg);
 	buf->topLine = buf->currentLine = l;
 	return;
     }
     if (buf->lastLine->linenumber < n) {
 	l = buf->lastLine;
-	/* FIXME: gettextize? */
-	sprintf(msg, "Last line is #%ld", buf->lastLine->linenumber);
+	sprintf(msg, _("Last line is #%ld"), buf->lastLine->linenumber);
 	set_delayed_message(msg);
 	buf->currentLine = l;
 	buf->topLine = lineSkip(buf, buf->currentLine, -(buf->LINES - 1),
@@ -282,7 +284,7 @@ gotoLine(Buffer *buf, int n)
 void
 gotoRealLine(Buffer *buf, int n)
 {
-    char msg[32];
+    char msg[36];
     Line *l = buf->firstLine;
 
     if (l == NULL)
@@ -294,16 +296,14 @@ gotoRealLine(Buffer *buf, int n)
 	       (getNextPage(buf, 1) != NULL)) ;
     }
     if (l->real_linenumber > n) {
-	/* FIXME: gettextize? */
-	sprintf(msg, "First line is #%ld", l->real_linenumber);
+	sprintf(msg, _("First line is #%ld"), l->real_linenumber);
 	set_delayed_message(msg);
 	buf->topLine = buf->currentLine = l;
 	return;
     }
     if (buf->lastLine->real_linenumber < n) {
 	l = buf->lastLine;
-	/* FIXME: gettextize? */
-	sprintf(msg, "Last line is #%ld", buf->lastLine->real_linenumber);
+	sprintf(msg, _("Last line is #%ld"), buf->lastLine->real_linenumber);
 	set_delayed_message(msg);
 	buf->currentLine = l;
 	buf->topLine = lineSkip(buf, buf->currentLine, -(buf->LINES - 1),
@@ -323,7 +323,7 @@ gotoRealLine(Buffer *buf, int n)
 
 
 static Buffer *
-listBuffer(Buffer *top, Buffer *current)
+listBuffer(Buffer *top, const Buffer *current)
 {
     int i, c = 0;
     Buffer *buf = top;
@@ -360,12 +360,9 @@ listBuffer(Buffer *top, Buffer *current)
 	buf = buf->nextBuffer;
     }
     standout();
-    /* FIXME: gettextize? */
-    message("Buffer selection mode: SPC for select / D for delete buffer", 0,
+    message(_("Buffer selection mode: SPC for select / D for delete buffer"), 0,
 	    0);
     standend();
-    /* 
-     * move(LASTLINE, COLS - 1); */
     move(c, 0);
     refresh();
     return buf->nextBuffer;
@@ -489,9 +486,6 @@ selectBuffer(Buffer *firstbuf, Buffer *currentbuf, char *selectchar)
 	    *selectchar = c;
 	    return currentbuf;
 	}
-	/* 
-	 * move(LASTLINE, COLS - 1);
-	 */
 	move(spoint, 0);
 	refresh();
     }
@@ -509,8 +503,6 @@ reshapeBuffer(Buffer *buf)
     wc_uint8 old_auto_detect = WcOption.auto_detect;
 #endif
 
-    if (!buf->need_reshape)
-	return;
     buf->need_reshape = FALSE;
     buf->width = INIT_BUFFER_WIDTH;
     if (buf->sourcefile == NULL)
@@ -612,11 +604,11 @@ void
 copyBuffer(Buffer *a, Buffer *b)
 {
     readBufferCache(b);
-    bcopy((void *)b, (void *)a, sizeof(Buffer));
+    memmove(a, b, sizeof(Buffer));
 }
 
 Buffer *
-prevBuffer(Buffer *first, Buffer *buf)
+prevBuffer(Buffer *first, const Buffer *buf)
 {
     Buffer *b;
 
@@ -705,6 +697,8 @@ readBufferCache(Buffer *buf)
 
     cache = fopen(buf->savecache, "r");
     if (cache == NULL || fread1(clnum, cache) || fread1(tlnum, cache)) {
+	if (cache != NULL)
+	    fclose(cache);
 	buf->savecache = NULL;
 	return -1;
     }
@@ -760,10 +754,56 @@ readBufferCache(Buffer *buf)
 	}
 #endif
     }
-    buf->lastLine = prevl;
-    buf->lastLine->next = NULL;
+    if (prevl) {
+	    buf->lastLine = prevl;
+	    buf->lastLine->next = NULL;
+    }
     fclose(cache);
     unlink(buf->savecache);
     buf->savecache = NULL;
     return 0;
 }
+
+char *
+getCurWord(Buffer *buf, int *spos, int *epos)
+{
+    char *p;
+    Line *l = buf->currentLine;
+    int b, e;
+
+    *spos = 0;
+    *epos = 0;
+    if (l == NULL)
+	return NULL;
+    p = l->lineBuf;
+    e = buf->pos;
+    while (e > 0 && !is_wordchar(getChar(&p[e])))
+	prevChar(e, l);
+    if (!is_wordchar(getChar(&p[e])))
+	return NULL;
+    b = e;
+    while (b > 0) {
+	int tmp = b;
+	prevChar(tmp, l);
+	if (!is_wordchar(getChar(&p[tmp])))
+	    break;
+	b = tmp;
+    }
+    while (e < l->len && is_wordchar(getChar(&p[e])))
+	nextChar(e, l);
+    *spos = b;
+    *epos = e;
+    return &p[b];
+}
+
+char *
+GetWord(Buffer *buf)
+{
+    int b, e;
+    char *p;
+
+    if ((p = getCurWord(buf, &b, &e)) != NULL) {
+	return Strnew_charp_n(p, e - b)->ptr;
+    }
+    return NULL;
+}
