@@ -1,10 +1,27 @@
-/* $Id: display.c,v 1.70 2007/05/29 12:07:02 inu Exp $ */
-#include <signal.h>
+/* vi: set sw=4 ts=8 ai sm noet : */
+#include "display.h"
+
+#include "buffer.h"
 #include "fm.h"
 
+#include <signal.h>
+
 /* *INDENT-OFF* */
 #ifdef USE_COLOR
 
+static void effect_anchor_start(void);
+static void effect_anchor_end(void);
+static void effect_image_start(void);
+static void effect_image_end(void);
+static void effect_form_start(void);
+static void effect_form_end(void);
+static void effect_active_start(void);
+static void effect_active_end(void);
+static void effect_visited_start(void);
+static void effect_visited_end(void);
+static void effect_mark_start(void);
+static void effect_mark_end(void);
+
 #define EFFECT_ANCHOR_START       effect_anchor_start()
 #define EFFECT_ANCHOR_END         effect_anchor_end()
 #define EFFECT_IMAGE_START        effect_image_start()
@@ -70,18 +87,22 @@
 static void name_start { if (useColor) { color_start; } else { mono_start; }}\
 static void name_end { if (useColor) { color_end; } else { mono_end; }}
 
-define_effect(EFFECT_ANCHOR_START, EFFECT_ANCHOR_END, EFFECT_ANCHOR_START_C,
-	      EFFECT_ANCHOR_END_C, EFFECT_ANCHOR_START_M, EFFECT_ANCHOR_END_M)
-define_effect(EFFECT_IMAGE_START, EFFECT_IMAGE_END, EFFECT_IMAGE_START_C,
-	      EFFECT_IMAGE_END_C, EFFECT_IMAGE_START_M, EFFECT_IMAGE_END_M)
-define_effect(EFFECT_FORM_START, EFFECT_FORM_END, EFFECT_FORM_START_C,
-	      EFFECT_FORM_END_C, EFFECT_FORM_START_M, EFFECT_FORM_END_M)
-define_effect(EFFECT_MARK_START, EFFECT_MARK_END, EFFECT_MARK_START_C,
-	      EFFECT_MARK_END_C, EFFECT_MARK_START_M, EFFECT_MARK_END_M)
+define_effect(effect_anchor_start(void), effect_anchor_end(void),
+	      EFFECT_ANCHOR_START_C, EFFECT_ANCHOR_END_C,
+	      EFFECT_ANCHOR_START_M, EFFECT_ANCHOR_END_M)
+define_effect(effect_image_start(void), effect_image_end(void),
+	      EFFECT_IMAGE_START_C, EFFECT_IMAGE_END_C,
+	      EFFECT_IMAGE_START_M, EFFECT_IMAGE_END_M)
+define_effect(effect_form_start(void), effect_form_end(void),
+	      EFFECT_FORM_START_C, EFFECT_FORM_END_C,
+	      EFFECT_FORM_START_M, EFFECT_FORM_END_M)
+define_effect(effect_mark_start(void), effect_mark_end(void),
+	      EFFECT_MARK_START_C, EFFECT_MARK_END_C,
+	      EFFECT_MARK_START_M, EFFECT_MARK_END_M)
 
 /*****************/
 static void
-EFFECT_ACTIVE_START
+effect_active_start(void)
 {
     if (useColor) {
 	if (useActiveColor) {
@@ -102,7 +123,7 @@ EFFECT_ACTIVE_START
 }
 
 static void
-EFFECT_ACTIVE_END
+effect_active_end(void)
 {
     if (useColor) {
 	if (useActiveColor) {
@@ -116,7 +137,7 @@ EFFECT_ACTIVE_END
 }
 
 static void
-EFFECT_VISITED_START
+effect_visited_start(void)
 {
     if (useVisitedColor) {
 	if (useColor) {
@@ -128,7 +149,7 @@ EFFECT_VISITED_START
 }
 
 static void
-EFFECT_VISITED_END
+effect_visited_end(void)
 {
     if (useVisitedColor) {
 	if (useColor) {
@@ -257,7 +278,7 @@ make_lastline_link(Buffer *buf, char *title, char *url)
     parseURL2(url, &pu, baseURL(buf));
     u = parsedURL2Str(&pu);
     if (DecodeURL)
-	u = Strnew_charp(url_unquote_conv(u->ptr, buf->document_charset));
+	u = Strnew_charp(url_decode2(u->ptr, buf));
 #ifdef USE_M17N
     u = checkType(u, &pr, NULL);
 #endif
@@ -328,11 +349,17 @@ make_lastline_message(Buffer *buf)
 	int cl = buf->currentLine->real_linenumber;
 	int ll = buf->lastLine->real_linenumber;
 	int r = (int)((double)cl * 100.0 / (double)(ll ? ll : 1) + 0.5);
-	Strcat(msg, Sprintf("%d/%d (%d%%)", cl, ll, r));
+	int cc;
+
+	if (displayColumnNumber) {
+	    cc = buf->currentColumn + buf->cursorX + 1;
+	    Strcat(msg, Sprintf("%d/%d,%d (%d%%)", cl, ll, cc, r));
+	}
+	else
+	    Strcat(msg, Sprintf("%d/%d (%d%%)", cl, ll, r));
     }
     else
-	/* FIXME: gettextize? */
-	Strcat_charp(msg, "Viewing");
+	Strcat_charp(msg, _("Viewing"));
 #ifdef USE_SSL
     if (buf->ssl_certificate)
 	Strcat_charp(msg, "[SSL]");
@@ -380,9 +407,8 @@ displayBuffer(Buffer *buf, int mode)
     if (buf->height == 0)
 	buf->height = LASTLINE + 1;
     if ((buf->width != INIT_BUFFER_WIDTH &&
-	 ((buf->type && !strcmp(buf->type, "text/html")) || FoldLine))
+	 (is_html_type(buf->type) || FoldLine))
 	|| buf->need_reshape) {
-	buf->need_reshape = TRUE;
 	reshapeBuffer(buf);
     }
     if (showLineNum) {
@@ -430,11 +456,6 @@ displayBuffer(Buffer *buf, int mode)
 		scroll(n);
 	    }
 	    else if (n < 0 && n > -buf->LINES) {
-#if 0 /* defined(__CYGWIN__) */
-		move(LASTLINE + n + 1, 0);
-		clrtoeolx();
-		refresh();
-#endif				/* defined(__CYGWIN__) */
 		rscroll(-n);
 	    }
 	    redrawNLine(buf, n);
@@ -473,8 +494,7 @@ displayBuffer(Buffer *buf, int mode)
 
     msg = make_lastline_message(buf);
     if (buf->firstLine == NULL) {
-	/* FIXME: gettextize? */
-	Strcat_charp(msg, "\tNo Line");
+	Strcat_charp(msg, _("\tNo Line"));
     }
     if (delayed_msg != NULL) {
 	disp_message(delayed_msg, FALSE);
@@ -487,7 +507,7 @@ displayBuffer(Buffer *buf, int mode)
     term_title(conv_to_system(buf->buffername));
     refresh();
 #ifdef USE_IMAGE
-    if (activeImage && displayImage && buf->img) {
+    if (activeImage && displayImage && buf->img && buf->image_loaded) {
 	drawImage();
     }
 #endif
@@ -497,6 +517,10 @@ displayBuffer(Buffer *buf, int mode)
 	save_current_buf = buf;
     }
 #endif
+    if (mode == B_FORCE_REDRAW &&  (buf->check_url & CHK_URL) ) {
+	chkURLBuffer(buf);
+	displayBuffer(buf, B_NORMAL);
+    }
 }
 
 static void
@@ -521,7 +545,15 @@ drawAnchorCursor0(Buffer *buf, AnchorList *al, int hseq, int prevhseq,
 		break;
 	}
 	if (hseq >= 0 && an->hseq == hseq) {
+	    int start_pos = an->start.pos;
+	    int end_pos = an->end.pos;
 	    for (i = an->start.pos; i < an->end.pos; i++) {
+	        if (enable_inline_image && (l->propBuf[i] & PE_IMAGE)) {
+		    if (start_pos == i)
+			start_pos = i + 1;
+		    else if (end_pos == an->end.pos)
+		        end_pos = i - 1;
+		}
 		if (l->propBuf[i] & (PE_IMAGE | PE_ANCHOR | PE_FORM)) {
 		    if (active)
 			l->propBuf[i] |= PE_ACTIVE;
@@ -529,9 +561,9 @@ drawAnchorCursor0(Buffer *buf, AnchorList *al, int hseq, int prevhseq,
 			l->propBuf[i] &= ~PE_ACTIVE;
 		}
 	    }
-	    if (active)
+	    if (active && start_pos < end_pos)
 		redrawLineRegion(buf, l, l->linenumber - tline + buf->rootY,
-				 an->start.pos, an->end.pos);
+				 start_pos, end_pos);
 	}
 	else if (prevhseq >= 0 && an->hseq == prevhseq) {
 	    if (active)
@@ -625,10 +657,6 @@ redrawNLine(Buffer *buf, int n)
 	    if (t == CurrentTab)
 		boldend();
 	}
-#if 0
-	move(0, COLS - 2);
-	addstr(" x");
-#endif
 	move(LastTab->y + 1, 0);
 	for (i = 0; i < COLS; i++)
 	    addch('~');
@@ -707,7 +735,6 @@ redrawLine(Buffer *buf, Line *l, int i)
 	clrtoeolx();
 	return l;
     }
-    /* need_clrtoeol(); */
     pos = columnPos(l, column);
     p = &(l->lineBuf[pos]);
     pr = &(l->propBuf[pos]);
@@ -855,14 +882,16 @@ redrawLineImage(Buffer *buf, Line *l, int i)
 		y = (int)(i * pixel_per_line);
 		sx = (int)((rcol - COLPOS(l, a->start.pos)) * pixel_per_char);
 		sy = (int)((l->linenumber - image->y) * pixel_per_line);
-		if (sx == 0 && x + image->xoffset >= 0)
-		    x += image->xoffset;
-		else
-		    sx -= image->xoffset;
-		if (sy == 0 && y + image->yoffset >= 0)
-		    y += image->yoffset;
-		else
-		    sy -= image->yoffset;
+		if (! enable_inline_image) {
+		    if (sx == 0 && x + image->xoffset >= 0)
+			x += image->xoffset;
+		    else
+			sx -= image->xoffset;
+		    if (sy == 0 && y + image->yoffset >= 0)
+			y += image->yoffset;
+		    else
+			sy -= image->yoffset;
+		}
 		if (image->width > 0)
 		    w = image->width - sx;
 		else
@@ -1119,18 +1148,18 @@ addChar(char c, Lineprop mode)
 	    }
 #ifdef USE_M17N
 	    if (w == 2 && WcOption.use_wide)
-		addstr(graph2_symbol[(int)c]);
+		addstr(graph2_symbol[(unsigned char)c % N_GRAPH_SYMBOL]);
 	    else
 #endif
-		addch(*graph_symbol[(int)c]);
+		addch(*graph_symbol[(unsigned char)c % N_GRAPH_SYMBOL]);
 	}
 	else {
 #ifdef USE_M17N
 	    symbol = get_symbol(DisplayCharset, &w);
-	    addstr(symbol[(int)c]);
+	    addstr(symbol[(unsigned char)c % N_SYMBOL]);
 #else
 	    symbol = get_symbol();
-	    addch(*symbol[(int)c]);
+	    addch(*symbol[(unsigned char)c % N_SYMBOL]);
 #endif
 	}
     }
@@ -1173,7 +1202,7 @@ addChar(char c, Lineprop mode)
 static GeneralList *message_list = NULL;
 
 void
-record_err_message(char *s)
+record_err_message(const char *s)
 {
     if (fmInitialized) {
 	if (!message_list)
@@ -1208,7 +1237,7 @@ message_list_panel(void)
 }
 
 void
-message(char *s, int return_x, int return_y)
+message(const char *s, int return_x, int return_y)
 {
     if (!fmInitialized)
 	return;
@@ -1219,14 +1248,14 @@ message(char *s, int return_x, int return_y)
 }
 
 void
-disp_err_message(char *s, int redraw_current)
+disp_err_message(const char *s, int redraw_current)
 {
     record_err_message(s);
     disp_message(s, redraw_current);
 }
 
 void
-disp_message_nsec(char *s, int redraw_current, int sec, int purge, int mouse)
+disp_message_nsec(const char *s, int redraw_current, int sec, int purge, int mouse)
 {
     if (QuietMessage)
 	return;
@@ -1254,7 +1283,7 @@ disp_message_nsec(char *s, int redraw_current, int sec, int purge, int mouse)
 }
 
 void
-disp_message(char *s, int redraw_current)
+disp_message(const char *s, int redraw_current)
 {
     disp_message_nsec(s, redraw_current, 10, FALSE, TRUE);
 }
@@ -1267,7 +1296,7 @@ disp_message_nomouse(char *s, int redraw_current)
 #endif
 
 void
-set_delayed_message(char *s)
+set_delayed_message(const char *s)
 {
     delayed_msg = allocStr(s, -1);
 }
@@ -1354,15 +1383,14 @@ cursorRight(Buffer *buf, int n)
 {
     int i, delta = 1, cpos, vpos2;
     Line *l = buf->currentLine;
-    Lineprop *p;
 
     if (buf->firstLine == NULL)
 	return;
     if (buf->pos == l->len && !(l->next && l->next->bpos))
 	return;
     i = buf->pos;
-    p = l->propBuf;
 #ifdef USE_M17N
+    Lineprop *p = l->propBuf;
     while (i + delta < l->len && p[i + delta] & PC_WCHAR2)
 	delta++;
 #endif
@@ -1405,13 +1433,12 @@ cursorLeft(Buffer *buf, int n)
 {
     int i, delta = 1, cpos;
     Line *l = buf->currentLine;
-    Lineprop *p;
 
     if (buf->firstLine == NULL)
 	return;
     i = buf->pos;
-    p = l->propBuf;
 #ifdef USE_M17N
+    Lineprop *p = l->propBuf;
     while (i - delta > 0 && p[i - delta] & PC_WCHAR2)
 	delta++;
 #endif
@@ -1436,14 +1463,6 @@ cursorLeft(Buffer *buf, int n)
     buf->cursorX = buf->visualpos - l->bwidth;
 }
 
-void
-cursorHome(Buffer *buf)
-{
-    buf->visualpos = 0;
-    buf->cursorX = buf->cursorY = 0;
-}
-
-
 /* 
  * Arrange line,column and cursor position according to current line and
  * current position.
@@ -1458,9 +1477,6 @@ arrangeCursor(Buffer *buf)
     /* Arrange line */
     if (buf->currentLine->linenumber - buf->topLine->linenumber >= buf->LINES
 	|| buf->currentLine->linenumber < buf->topLine->linenumber) {
-	/*
-	 * buf->topLine = buf->currentLine;
-	 */
 	buf->topLine = lineSkip(buf, buf->currentLine, 0, FALSE);
     }
     /* Arrange column */
