@@ -1,4 +1,4 @@
-/* $Id: fm.h,v 1.149 2010/08/20 09:47:09 htrb Exp $ */
+/* vi: set sw=4 ts=8 ai sm noet : */
 /* 
  * w3m: WWW wo Miru utility
  * 
@@ -19,6 +19,7 @@
 #include <stdlib.h>
 #include <string.h>
 #include <unistd.h>
+#include <limits.h>
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <unistd.h>
@@ -37,13 +38,11 @@
 
 #include "ctrlcode.h"
 #include "html.h"
-#include <gc.h>
+#include <gc/gc.h>
 #include "Str.h"
 #ifdef USE_M17N
 #include "wc.h"
 #include "wtf.h"
-#else
-typedef int wc_ces;	/* XXX: not used */
 #endif
 
 #ifdef HAVE_LOCALE_H
@@ -74,16 +73,8 @@ typedef int wc_ces;	/* XXX: not used */
 #include "func.h"
 #include "menu.h"
 #include "textlist.h"
-#include "funcname1.h"
 #include "terms.h"
-
-#ifndef HAVE_BCOPY
-void bcopy(const void *, void *, int);
-void bzero(void *, int);
-#endif				/* HAVE_BCOPY */
-#ifdef __EMX__
-#include <strings.h>		/* for bzero() and bcopy() */
-#endif
+#include "istream.h"
 
 #ifdef MAINPROGRAM
 #define global
@@ -101,7 +92,6 @@ void bzero(void *, int);
 #define LINELEN	256		/* Initial line length */
 #define PAGER_MAX_LINE	10000	/* Maximum line kept as pager */
 
-#define MAXIMUM_COLS 1024
 #define DEFAULT_COLS 80
 
 #ifdef USE_IMAGE
@@ -134,6 +124,10 @@ void bzero(void *, int);
 #define DICTBUFFERNAME "*dictionary*"
 #endif				/* USE_DICT */
 
+#ifndef HOST_NAME_MAX
+#define HOST_NAME_MAX 255
+#endif
+
 /* 
  * Line Property
  */
@@ -264,6 +258,20 @@ extern int REV_LB[];
 #define IMG_FLAG_ERROR		2
 #define IMG_FLAG_DONT_REMOVE	4
 
+#define IS_EMPTY_PARSED_URL(pu) ((pu)->scheme == SCM_UNKNOWN && !(pu)->file)
+#define SCONF_RESERVED		0
+#define SCONF_SUBSTITUTE_URL	1
+#define SCONF_URL_CHARSET	2
+#define SCONF_NO_REFERER_FROM	3
+#define SCONF_NO_REFERER_TO	4
+#define SCONF_USER_AGENT	5
+#define SCONF_N_FIELD		6
+#define query_SCONF_SUBSTITUTE_URL(pu) ((char *)querySiteconf(pu, SCONF_SUBSTITUTE_URL))
+#define query_SCONF_USER_AGENT(pu) ((const char *)querySiteconf(pu, SCONF_USER_AGENT))
+#define query_SCONF_URL_CHARSET(pu) ((const wc_ces *)querySiteconf(pu, SCONF_URL_CHARSET))
+#define query_SCONF_NO_REFERER_FROM(pu) ((const int *)querySiteconf(pu, SCONF_NO_REFERER_FROM))
+#define query_SCONF_NO_REFERER_TO(pu) ((const int *)querySiteconf(pu, SCONF_NO_REFERER_TO))
+
 /* 
  * Macros.
  */
@@ -276,8 +284,6 @@ extern int REV_LB[];
 #define inputFilenameHist(p,d,h)	inputLineHist(p,d,IN_FILENAME,h)
 #define inputChar(p)		inputLine(p,"",IN_CHAR)
 
-#define free(x)  GC_free(x)	/* let GC do it. */
-
 #ifdef __EMX__
 #define HAVE_STRCASECMP
 #define strcasecmp	stricmp
@@ -285,8 +291,8 @@ extern int REV_LB[];
 #endif				/* __EMX__ */
 
 
-#define SKIP_BLANKS(p) {while(*(p)&&IS_SPACE(*(p)))(p)++;}
-#define SKIP_NON_BLANKS(p) {while(*(p)&&!IS_SPACE(*(p)))(p)++;}
+#define SKIP_BLANKS(p) do{while(*(p)&&IS_SPACE(*(p)))(p)++;}while(0)
+#define SKIP_NON_BLANKS(p) do{while(*(p)&&!IS_SPACE(*(p)))(p)++;}while(0)
 #define IS_ENDL(c) ((c)=='\0'||(c)=='\r'||(c)=='\n')
 #define IS_ENDT(c) (IS_ENDL(c)||(c)==';')
 
@@ -299,6 +305,12 @@ extern int REV_LB[];
 #define EOL(l) (&(l)->ptr[(l)->length])
 #define IS_EOL(p,l) ((p)==&(l)->ptr[(l)->length])
 
+#define INLINE_IMG_NONE		0
+#define INLINE_IMG_OSC5379	1
+#define INLINE_IMG_SIXEL	2
+#define INLINE_IMG_ITERM2	3
+#define INLINE_IMG_KITTY	4
+
 /* 
  * Types.
  */
@@ -362,6 +374,8 @@ typedef struct _imageCache {
     int index;
     short width;
     short height;
+    short a_width;
+    short a_height;
 } ImageCache;
 
 typedef struct _image {
@@ -491,6 +505,7 @@ typedef struct _Buffer {
 #ifdef USE_ALARM
     struct _AlarmEvent *event;
 #endif
+    int mainline;
 } Buffer;
 
 typedef struct _BufferPos {
@@ -518,7 +533,7 @@ typedef struct _DownloadList {
     char *url;
     char *save;
     char *lock;
-    clen_t size;
+    size_t size;
     time_t time;
     int running;
     int err;
@@ -557,11 +572,19 @@ typedef struct _DownloadList {
 #define FONT_STACK_SIZE 5
 
 #define FONTSTAT_SIZE 7
+#define FONTSTAT_MAX 127
 
 #define _INIT_BUFFER_WIDTH (COLS - (showLineNum ? 6 : 1))
 #define INIT_BUFFER_WIDTH ((_INIT_BUFFER_WIDTH > 0) ? _INIT_BUFFER_WIDTH : 0)
 #define FOLD_BUFFER_WIDTH (FoldLine ? (INIT_BUFFER_WIDTH + 1) : -1)
 
+struct input_alt_attr {
+  int hseq;
+  int fid;
+  int in;
+  Str type, name, value;
+};
+
 typedef struct {
     int pos;
     int len;
@@ -569,6 +592,7 @@ typedef struct {
     long flag;
     Anchor anchor;
     Str img_alt;
+    struct input_alt_attr input_alt;
     char fontstat[FONTSTAT_SIZE];
     short nobr_level;
     Lineprop prev_ctype;
@@ -587,10 +611,12 @@ struct readbuffer {
     int flag_sp;
     int status;
     unsigned char end_tag;
+    unsigned char q_level;
     short table_level;
     short nobr_level;
     Anchor anchor;
     Str img_alt;
+    struct input_alt_attr input_alt;
     char fontstat[FONTSTAT_SIZE];
     char fontstat_stack[FONT_STACK_SIZE][FONTSTAT_SIZE];
     int fontstat_sp;
@@ -632,14 +658,12 @@ struct readbuffer {
 #define RB_SPECIAL	(RB_PRE | RB_PRE_INT | RB_SCRIPT | RB_STYLE | RB_PLAIN | RB_NOBR)
 #define RB_PLAIN_PRE	0x40000
 
-#ifdef FORMAT_NICE
-#define RB_FILL		0x80000
-#endif				/* FORMAT_NICE */
 #define RB_DEL		0x100000
 #define RB_S		0x200000
+#define RB_HTML5	0x400000
 
 #define RB_GET_ALIGN(obuf) ((obuf)->flag&RB_ALIGN)
-#define RB_SET_ALIGN(obuf,align) {(obuf)->flag &= ~RB_ALIGN; (obuf)->flag |= (align); }
+#define RB_SET_ALIGN(obuf,align) do{(obuf)->flag &= ~RB_ALIGN; (obuf)->flag |= (align); }while(0)
 #define RB_SAVE_FLAG(obuf) {\
   if ((obuf)->flag_sp < RB_STACK_SIZE) \
     (obuf)->flag_stack[(obuf)->flag_sp++] = RB_GET_ALIGN(obuf); \
@@ -649,7 +673,7 @@ struct readbuffer {
    RB_SET_ALIGN(obuf,(obuf)->flag_stack[--(obuf)->flag_sp]); \
 }
 
-/* status flags */
+/* state of token scanning finite state machine */
 #define R_ST_NORMAL 0		/* normal */
 #define R_ST_TAG0   1		/* within tag, just after < */
 #define R_ST_TAG    2		/* within tag */
@@ -693,49 +717,6 @@ struct html_feed_environ {
     int blank_lines;
 };
 
-#ifdef USE_COOKIE
-struct portlist {
-    unsigned short port;
-    struct portlist *next;
-};
-
-struct cookie {
-    ParsedURL url;
-    Str name;
-    Str value;
-    time_t expires;
-    Str path;
-    Str domain;
-    Str comment;
-    Str commentURL;
-    struct portlist *portl;
-    char version;
-    char flag;
-    struct cookie *next;
-};
-#define COO_USE		1
-#define COO_SECURE	2
-#define COO_DOMAIN	4
-#define COO_PATH	8
-#define COO_DISCARD	16
-#define COO_OVERRIDE	32	/* user chose to override security checks */
-
-#define COO_OVERRIDE_OK	32	/* flag to specify that an error is overridable */
-						/* version 0 refers to the original cookie_spec.html */
-						/* version 1 refers to RFC 2109 */
-						/* version 1' refers to the Internet draft to obsolete RFC 2109 */
-#define COO_EINTERNAL	(1)	/* unknown error; probably forgot to convert "return 1" in cookie.c */
-#define COO_ETAIL	(2 | COO_OVERRIDE_OK)	/* tail match failed (version 0) */
-#define COO_ESPECIAL	(3)	/* special domain check failed (version 0) */
-#define COO_EPATH	(4)	/* Path attribute mismatch (version 1 case 1) */
-#define COO_ENODOT	(5 | COO_OVERRIDE_OK)	/* no embedded dots in Domain (version 1 case 2.1) */
-#define COO_ENOTV1DOM	(6 | COO_OVERRIDE_OK)	/* Domain does not start with a dot (version 1 case 2.2) */
-#define COO_EDOM	(7 | COO_OVERRIDE_OK)	/* domain-match failed (version 1 case 3) */
-#define COO_EBADHOST	(8 | COO_OVERRIDE_OK)	/* dot in matched host name in FQDN (version 1 case 4) */
-#define COO_EPORT	(9)	/* Port match failed (version 1' case 5) */
-#define COO_EMAX	COO_EPORT
-#endif				/* USE_COOKIE */
-
 /* modes for align() */
 
 #define ALIGN_CENTER 0
@@ -774,7 +755,8 @@ typedef struct http_request {
 #define TMPF_FRAME	2
 #define TMPF_CACHE	3
 #define TMPF_COOKIE	4
-#define MAX_TMPF_TYPE	5
+#define TMPF_HIST	5
+#define MAX_TMPF_TYPE	6
 
 #define set_no_proxy(domains) (NO_proxy_domains=make_domain_list(domains))
 
@@ -782,7 +764,6 @@ typedef struct http_request {
  * Globals.
  */
 
-extern int LINES, COLS;
 #if defined(__CYGWIN__)
 extern int LASTLINE;
 #else				/* not defined(__CYGWIN__) */
@@ -791,8 +772,10 @@ extern int LASTLINE;
 
 global int Tabstop init(8);
 global int IndentIncr init(4);
+global int MaxCols init(0);
 global int ShowEffect init(TRUE);
 global int PagerMax init(PAGER_MAX_LINE);
+global int MessageDelay init(2);
 
 global char SearchHeader init(FALSE);
 global char *DefaultType init(NULL);
@@ -802,8 +785,10 @@ global char PermitSaveToPipe init(FALSE);
 global char DecodeCTE init(FALSE);
 global char AutoUncompress init(FALSE);
 global char PreserveTimestamp init(TRUE);
-global char ArgvIsURL init(FALSE);
+global char ArgvIsURL init(TRUE);
 global char MetaRefresh init(FALSE);
+global char LocalhostOnly init(FALSE);
+global char* HostName init(NULL);
 
 global char fmInitialized init(FALSE);
 global char QuietMessage init(FALSE);
@@ -874,10 +859,9 @@ global char *index_file init(NULL);
 
 global char *CurrentDir;
 global int CurrentPid;
-/*
- * global Buffer *Currentbuf;
- * global Buffer *Firstbuf;
- */
+#if defined(DONT_CALL_GC_AFTER_FORK) && defined(USE_IMAGE)
+global char *MyProgramName init("w3m");
+#endif /* defined(DONT_CALL_GC_AFTER_FORK) && defined(USE_IMAGE) */
 global TabBuffer *CurrentTab;
 global TabBuffer *FirstTab;
 global TabBuffer *LastTab;
@@ -895,7 +879,7 @@ global int CurrentKey;
 global char *CurrentKeyData;
 global char *CurrentCmdData;
 global char *w3m_reqlog;
-extern char *w3m_version;
+extern int enable_inline_image;
 
 #define DUMP_BUFFER   0x01
 #define DUMP_HEAD     0x02
@@ -909,9 +893,11 @@ global int w3m_dump init(0);
 global int w3m_halfload init(FALSE);
 global Str header_string init(NULL);
 global int override_content_type init(FALSE);
+global int override_user_agent init(FALSE);
 
 #ifdef USE_COLOR
 global int useColor init(TRUE);
+global int highIntensityColors init(FALSE);
 global int basic_color init(8);	/* don't change */
 global int anchor_color init(4);	/* blue  */
 global int image_color init(2);	/* green */
@@ -926,17 +912,22 @@ global int useVisitedColor init(FALSE);
 global int visited_color init(5);	/* magenta  */
 #endif				/* USE_COLOR */
 global int confirm_on_quit init(TRUE);
+global int exit_on_last init(FALSE);
 #ifdef USE_MARK
 global int use_mark init(FALSE);
 #endif
 global int emacs_like_lineedit init(FALSE);
+global int rl_paste init(FALSE);
+global int space_autocomplete init(FALSE);
 global int vi_prec_num init(FALSE);
 global int label_topline init(FALSE);
 global int nextpage_topline init(FALSE);
 global char *displayTitleTerm init(NULL);
 global int displayLink init(FALSE);
 global int displayLinkNumber init(FALSE);
+global int zeroBasedLinkNo init(FALSE);
 global int displayLineInfo init(FALSE);
+global int displayColumnNumber init(FALSE);
 global int DecodeURL init(FALSE);
 global int retryAsHttp init(TRUE);
 global int showLineNum init(FALSE);
@@ -953,7 +944,7 @@ global int image_map_list init(TRUE);
 global int displayImage init(FALSE);	/* XXX: emacs-w3m use display_image=off */
 #endif
 global int pseudoInlines init(TRUE);
-global char *Editor init(DEF_EDITOR);
+global const char *Editor init(DEF_EDITOR);
 #ifdef USE_W3MMAILER
 global char *Mailer init(NULL);
 #else
@@ -968,10 +959,17 @@ global int MailtoOptions init(MAILTO_OPTIONS_IGNORE);
 global char *ExtBrowser init(DEF_EXT_BROWSER);
 global char *ExtBrowser2 init(NULL);
 global char *ExtBrowser3 init(NULL);
+global char *ExtBrowser4 init(NULL);
+global char *ExtBrowser5 init(NULL);
+global char *ExtBrowser6 init(NULL);
+global char *ExtBrowser7 init(NULL);
+global char *ExtBrowser8 init(NULL);
+global char *ExtBrowser9 init(NULL);
 global int BackgroundExtViewer init(TRUE);
 global int disable_secret_security_check init(FALSE);
 global char *passwd_file init(PASSWD_FILE);
 global char *pre_form_file init(PRE_FORM_FILE);
+global char *siteconf_file init(SITECONF_FILE);
 global char *ftppasswd init(NULL);
 global int ftppass_hostnamegen init(TRUE);
 global int do_download init(FALSE);
@@ -980,19 +978,22 @@ global char *image_source init(NULL);
 #endif
 global char *UserAgent init(NULL);
 global int NoSendReferer init(FALSE);
+global int CrossOriginReferer init(TRUE);
 global char *AcceptLang init(NULL);
 global char *AcceptEncoding init(NULL);
 global char *AcceptMedia init(NULL);
 global int WrapDefault init(FALSE);
 global int IgnoreCase init(TRUE);
+global int SmartCase init(FALSE);
 global int WrapSearch init(FALSE);
 global int squeezeBlankLine init(FALSE);
 global char *BookmarkFile init(NULL);
 global int UseExternalDirBuffer init(TRUE);
-global char *DirBufferCommand init("file:///$LIB/dirlist" CGI_EXTENSION);
+global const char *DirBufferCommand init("file:///$LIB/dirlist" CGI_EXTENSION);
 #ifdef USE_DICT
-global int UseDictCommand init(FALSE);
-global char *DictCommand init("file:///$LIB/w3mdict" CGI_EXTENSION);
+global int UseDictCommand init(TRUE);
+global const char *DictCommand init("file:///$LIB/w3mdict" CGI_EXTENSION);
+global const char *DictPrompt init("(dictionary)!");
 #endif				/* USE_DICT */
 global int ignore_null_img_alt init(TRUE);
 #define DISPLAY_INS_DEL_SIMPLE	0
@@ -1000,11 +1001,12 @@ global int ignore_null_img_alt init(TRUE);
 #define DISPLAY_INS_DEL_FONTIFY	2
 global int displayInsDel init(DISPLAY_INS_DEL_NORMAL);
 global int FoldTextarea init(FALSE);
+global int FoldPre init(FALSE);
 global int FoldLine init(FALSE);
 #define DEFAULT_URL_EMPTY	0
 #define DEFAULT_URL_CURRENT	1
 #define DEFAULT_URL_LINK	2
-global int DefaultURLString init(DEFAULT_URL_EMPTY);
+global int DefaultURLString init(DEFAULT_URL_CURRENT);
 global int MarkAllPages init(FALSE);
 
 #ifdef USE_MIGEMO
@@ -1013,11 +1015,6 @@ global int migemo_active init(0);
 global char *migemo_command init(DEF_MIGEMO_COMMAND);
 #endif				/* USE_MIGEMO */
 
-global struct auth_cookie *Auth_cookie init(NULL);
-#ifdef USE_COOKIE
-global struct cookie *First_cookie init(NULL);
-#endif				/* USE_COOKIE */
-
 global char *mailcap_files init(USER_MAILCAP ", " SYS_MAILCAP);
 global char *mimetypes_files init(USER_MIMETYPES ", " SYS_MIMETYPES);
 #ifdef USE_EXTERNAL_URI_LOADER
@@ -1065,31 +1062,38 @@ global char SimplePreserveSpace init(FALSE);
 #define wc_Str_conv(x,charset0,charset1) (x)
 #define wc_Str_conv_strict(x,charset0,charset1) (x)
 #endif
-global char UseAltEntity init(TRUE);
+global char UseAltEntity init(FALSE);
 #define GRAPHIC_CHAR_ASCII 2
 #define GRAPHIC_CHAR_DEC 1
 #define GRAPHIC_CHAR_CHARSET 0
 global char UseGraphicChar init(GRAPHIC_CHAR_CHARSET);
+global char DisplayBorders init(FALSE);
+global char DisableCenter init(FALSE);
 extern char *graph_symbol[];
 extern char *graph2_symbol[];
 extern int symbol_width;
 extern int symbol_width0;
 #define N_GRAPH_SYMBOL 32
+#define N_SYMBOL (N_GRAPH_SYMBOL + 14)
 #define SYMBOL_BASE 0x20
 global int no_rc_dir init(FALSE);
 global char *rc_dir init(NULL);
 global char *tmp_dir;
+global char *param_tmp_dir init(NULL);
+global char *param_dl_dir init(NULL);
+#ifdef HAVE_MKDTEMP
+global char *mkd_tmp_dir init(NULL);
+#endif
 global char *config_file init(NULL);
 
 #ifdef USE_MOUSE
 global int use_mouse init(TRUE);
-extern int mouseActive;
 global int reverse_mouse init(FALSE);
 global int relative_wheel_scroll init(FALSE);
 global int fixed_wheel_scroll_count init(5);
 global int relative_wheel_scroll_ratio init(30);
 typedef struct _MouseActionMap {
-    void (*func) ();
+    void (*func) (void);
     char *data;
 } MouseActionMap;
 typedef struct _MouseAction {
@@ -1111,23 +1115,6 @@ global MouseAction mouse_action;
 #define LIMIT_MOUSE_MENU 100
 #endif				/* USE_MOUSE */
 
-#ifdef USE_COOKIE
-global int default_use_cookie init(TRUE);
-global int use_cookie init(FALSE);
-global int show_cookie init(TRUE);
-global int accept_cookie init(FALSE);
-#define ACCEPT_BAD_COOKIE_DISCARD	0
-#define ACCEPT_BAD_COOKIE_ACCEPT	1
-#define ACCEPT_BAD_COOKIE_ASK		2
-global int accept_bad_cookie init(ACCEPT_BAD_COOKIE_DISCARD);
-global char *cookie_reject_domains init(NULL);
-global char *cookie_accept_domains init(NULL);
-global char *cookie_avoid_wrong_number_of_dots init(NULL);
-global TextList *Cookie_reject_domains;
-global TextList *Cookie_accept_domains;
-global TextList *Cookie_avoid_wrong_number_of_dots_domains;
-#endif				/* USE_COOKIE */
-
 #ifdef USE_IMAGE
 global int view_unseenobject init(FALSE);
 #else
@@ -1135,24 +1122,35 @@ global int view_unseenobject init(TRUE);
 #endif
 
 #if defined(USE_SSL) && defined(USE_SSL_VERIFY)
-global int ssl_verify_server init(FALSE);
+global int ssl_verify_server init(TRUE);
 global char *ssl_cert_file init(NULL);
 global char *ssl_key_file init(NULL);
 global char *ssl_ca_path init(NULL);
-global char *ssl_ca_file init(NULL);
+global char *ssl_ca_file init(DEF_CAFILE);
+global int ssl_ca_default init(TRUE);
 global int ssl_path_modified init(FALSE);
 #endif				/* defined(USE_SSL) &&
 				 * defined(USE_SSL_VERIFY) */
 #ifdef USE_SSL
-global char *ssl_forbid_method init(NULL);
+global char *ssl_forbid_method init("2, 3, t, 5");
+#ifdef SSL_CTX_set_min_proto_version
+global char *ssl_min_version init(NULL);
 #endif
+#if (OPENSSL_VERSION_NUMBER < 0x10100000L) || defined(LIBRESSL_VERSION_NUMBER)
+global char *ssl_cipher init("DEFAULT:!LOW:!RC4:!EXP");
+#else
+global char *ssl_cipher init(NULL);
+#endif
+#endif				/* USE_SSL */
 
 global int is_redisplay init(FALSE);
 global int clear_buffer init(TRUE);
 global double pixel_per_char init(DEFAULT_PIXEL_PER_CHAR);
+global int pixel_per_char_i init(DEFAULT_PIXEL_PER_CHAR);
 global int set_pixel_per_char init(FALSE);
 #ifdef USE_IMAGE
 global double pixel_per_line init(DEFAULT_PIXEL_PER_LINE);
+global int pixel_per_line_i init(DEFAULT_PIXEL_PER_LINE);
 global int set_pixel_per_line init(FALSE);
 global double image_scale init(100);
 #endif
@@ -1161,10 +1159,10 @@ global int use_lessopen init(FALSE);
 global char *keymap_file init(KEYMAP_FILE);
 
 #ifdef USE_M17N
-#define get_mctype(c) ((Lineprop)wtf_type((wc_uchar *)(c)) << 8)
-#define get_mclen(c) wtf_len1((wc_uchar *)(c))
-#define get_mcwidth(c) wtf_width((wc_uchar *)(c))
-#define get_strwidth(c) wtf_strwidth((wc_uchar *)(c))
+#define get_mctype(c) ((Lineprop)wtf_type((const wc_uchar *)(c)) << 8)
+#define get_mclen(c) wtf_len1((const wc_uchar *)(c))
+#define get_mcwidth(c) wtf_width((const wc_uchar *)(c))
+#define get_strwidth(c) wtf_strwidth((const wc_uchar *)(c))
 #define get_Str_strwidth(c) wtf_strwidth((wc_uchar *)((c)->ptr))
 #else
 #define get_mctype(c) (IS_CNTRL(*(c)) ? PC_CTRL : PC_ASCII)
