@@ -1,20 +1,16 @@
-/* $Id: frame.c,v 1.32 2003/05/12 16:33:18 ukai Exp $ */
+/* vi: set sw=4 ts=8 ai sm noet : */
+#include "buffer.h"
 #include "fm.h"
 #include "parsetagx.h"
 #include "myctype.h"
 #include <signal.h>
 #include <setjmp.h>
-
-#ifdef KANJI_SYMBOLS
-#define RULE_WIDTH 2
-#else				/* not KANJI_SYMBOLS */
-#define RULE_WIDTH 1
-#endif				/* not KANJI_SYMBOLS */
+#include <strings.h>
 
 static JMP_BUF AbortLoading;
 struct frameset *renderFrameSet = NULL;
 
-static MySignalHandler
+static void
 KeyAbort(SIGNAL_ARG)
 {
     LONGJMP(AbortLoading, 1);
@@ -91,15 +87,16 @@ newFrame(struct parsed_tag *tag, Buffer *buf)
     char *p;
 
     body = New(struct frame_body);
-    bzero((void *)body, sizeof(*body));
+    bzero(body, sizeof(*body));
     body->attr = F_UNLOADED;
     body->flags = 0;
     body->baseURL = baseURL(buf);
     if (tag) {
 	if (parsedtag_get_value(tag, ATTR_SRC, &p))
-	    body->url = url_quote_conv(remove_space(p), buf->document_code);
+	    body->url = url_encode(remove_space(p), body->baseURL,
+				   buf->document_charset);
 	if (parsedtag_get_value(tag, ATTR_NAME, &p) && *p != '_')
-	    body->name = url_quote_conv(p, buf->document_code);
+	    body->name = url_quote_conv(p, buf->document_charset);
     }
     return body;
 }
@@ -116,7 +113,7 @@ deleteFrame(struct frame_body *b)
     if (b == NULL)
 	return;
     unloadFrame(b);
-    bzero((void *)b, sizeof(*b));
+    bzero(b, sizeof(*b));
 }
 
 void
@@ -174,7 +171,7 @@ copyFrame(struct frame_body *ob)
     struct frame_body *rb;
 
     rb = New(struct frame_body);
-    bcopy((const void *)ob, (void *)rb, sizeof(struct frame_body));
+    memmove(rb, ob, sizeof(struct frame_body));
     return rb;
 }
 
@@ -186,13 +183,11 @@ copyFrameSet(struct frameset *of)
 
     rf = New(struct frameset);
     n = of->col * of->row;
-    bcopy((const void *)of, (void *)rf, sizeof(struct frameset));
+    memmove(rf, of, sizeof(struct frameset));
     rf->width = New_N(char *, rf->col);
-    bcopy((const void *)of->width,
-	  (void *)rf->width, sizeof(char *) * rf->col);
+    memmove(rf->width, of->width, sizeof(char *) * rf->col);
     rf->height = New_N(char *, rf->row);
-    bcopy((const void *)of->height,
-	  (void *)rf->height, sizeof(char *) * rf->row);
+    memmove(rf->height, of->height, sizeof(char *) * rf->row);
     rf->frame = New_N(union frameset_element, n);
     while (n) {
 	n--;
@@ -215,7 +210,7 @@ copyFrameSet(struct frameset *of)
     return rf;
 }
 
-void
+static void
 flushFrameSet(struct frameset *fs)
 {
     int n = fs->i;
@@ -287,7 +282,7 @@ popFrameTree(struct frameset_queue **fqpp)
 	(rfq = cfq->back)->next = cfq->next;
     }
     *fqpp = rfq;
-    bzero((void *)cfq, sizeof(struct frameset_queue));
+    bzero(cfq, sizeof(struct frameset_queue));
     return rfs;
 }
 
@@ -342,9 +337,6 @@ frame_download_source(struct frame_body *b, ParsedURL *currentURL,
     parseURL2(b->url, &url, currentURL);
     switch (url.scheme) {
     case SCM_LOCAL:
-#if 0
-	b->source = url.real_file;
-#endif
 	b->flags = 0;
     default:
 	is_redisplay = TRUE;
@@ -410,13 +402,12 @@ createFrameFile(struct frameset *f, FILE * f1, Buffer *current, int level,
 {
     int r, c, t_stack;
     URLFile f2;
-    char code;
-#ifdef JP_CHARSET
-    char charset[2];
-#endif				/* JP_CHARSET */
+#ifdef USE_M17N
+    wc_ces charset, doc_charset;
+#endif
     char *d_target, *p_target, *s_target, *t_target;
     ParsedURL *currentURL, base;
-    MySignalHandler(*volatile prevtrap) (SIGNAL_ARG) = NULL;
+    void (*volatile prevtrap) (SIGNAL_ARG) = NULL;
     int flag;
 
     if (f == NULL)
@@ -445,9 +436,6 @@ createFrameFile(struct frameset *f, FILE * f1, Buffer *current, int level,
 	fputs("<table hborder>\n", f1);
 
     currentURL = f->currentURL ? f->currentURL : &current->currentURL;
-#ifdef JP_CHARSET
-    charset[1] = '\0';
-#endif
     for (r = 0; r < f->row; r++) {
 	fputs("<tr valign=top>\n", f1);
 	for (c = 0; c < f->col; c++) {
@@ -482,7 +470,7 @@ createFrameFile(struct frameset *f, FILE * f1, Buffer *current, int level,
 	    }
 	    switch (frame.element->attr) {
 	    default:
-		fprintf(f1, "Frameset \"%s\" frame %d: type unrecognized",
+		fprintf(f1, _("Frameset \"%s\" frame %d: type unrecognized"),
 			html_quote(f->name), i + 1);
 		break;
 	    case F_UNLOADED:
@@ -508,12 +496,13 @@ createFrameFile(struct frameset *f, FILE * f1, Buffer *current, int level,
 		if (f2.stream == NULL) {
 		    frame.body->attr = F_UNLOADED;
 		    if (frame.body->flags & FB_NO_BUFFER)
-			fprintf(f1, "Open %s with other method",
+			fprintf(f1, _("Open %s with other method"),
 				html_quote(frame.body->url));
 		    else if (frame.body->url)
-			fprintf(f1, "Can't open %s",
+			fprintf(f1, _("Can't open %s"),
 				html_quote(frame.body->url));
 		    else
+			/* FIXME: gettextize? */
 			fprintf(f1,
 				"This frame (%s) contains no src attribute",
 				frame.body->name ? html_quote(frame.body->name)
@@ -525,16 +514,21 @@ createFrameFile(struct frameset *f, FILE * f1, Buffer *current, int level,
 		s_target = frame.body->name;
 		t_target = "_blank";
 		d_target = TargetSelf ? s_target : t_target;
-#ifdef JP_CHARSET
-		code = DocumentCode;
+#ifdef USE_M17N
+		charset = WC_CES_US_ASCII;
+		if (current->document_charset != WC_CES_US_ASCII)
+		    doc_charset = current->document_charset;
+		else
+		    doc_charset = DocumentCharset;
 #endif
 		t_stack = 0;
 		if (frame.body->type &&
 		    !strcasecmp(frame.body->type, "text/plain")) {
 		    Str tmp;
 		    fprintf(f1, "<pre>\n");
-		    while ((tmp = StrmyUFgets(&f2))->length) {
-			tmp = convertLine(NULL, tmp, &code, HTML_MODE);
+		    while ((tmp = StrmyUFgets(&f2)) && tmp->length) {
+			tmp = convertLine(NULL, tmp, HTML_MODE, &charset,
+					  doc_charset);
 			fprintf(f1, "%s", html_quote(tmp->ptr));
 		    }
 		    fprintf(f1, "</pre>\n");
@@ -549,9 +543,10 @@ createFrameFile(struct frameset *f, FILE * f1, Buffer *current, int level,
 		    do {
 			if (*p == '\0') {
 			    Str tmp = StrmyUFgets(&f2);
-			    if (tmp->length == 0)
+			    if (!tmp || tmp->length == 0)
 				break;
-			    tmp = convertLine(NULL, tmp, &code, HTML_MODE);
+			    tmp = convertLine(NULL, tmp, HTML_MODE, &charset,
+					      doc_charset);
 			    p = tmp->ptr;
 			}
 			read_token(tok, &p, &status, 1, status != R_ST_NORMAL);
@@ -639,7 +634,7 @@ createFrameFile(struct frameset *f, FILE * f1, Buffer *current, int level,
 			case HTML_BASE:
 			    /* "BASE" is prohibit tag */
 			    if (parsedtag_get_value(tag, ATTR_HREF, &q)) {
-				q = url_quote_conv(remove_space(q), code);
+				q = url_encode(remove_space(q), NULL, charset);
 				parseURL(q, &base, NULL);
 			    }
 			    if (parsedtag_get_value(tag, ATTR_TARGET, &q)) {
@@ -648,7 +643,7 @@ createFrameFile(struct frameset *f, FILE * f1, Buffer *current, int level,
 				else if (!strcasecmp(q, "_parent"))
 				    d_target = p_target;
 				else
-				    d_target = url_quote_conv(q, code);
+				    d_target = url_quote_conv(q, charset);
 			    }
 			    Strshrinkfirst(tok, 1);
 			    Strshrink(tok, 1);
@@ -670,6 +665,25 @@ createFrameFile(struct frameset *f, FILE * f1, Buffer *current, int level,
 				    }
 				}
 			    }
+#ifdef USE_M17N
+			    if (UseContentCharset &&
+				parsedtag_get_value(tag, ATTR_HTTP_EQUIV, &q)
+				&& !strcasecmp(q, "Content-Type")
+				&& parsedtag_get_value(tag, ATTR_CONTENT, &q)
+				&& (q = strcasestr(q, "charset")) != NULL) {
+				q += 7;
+				SKIP_BLANKS(q);
+				if (*q == '=') {
+				    wc_ces c;
+				    q++;
+				    SKIP_BLANKS(q);
+				    if ((c = wc_guess_charset(q, 0)) != 0) {
+					doc_charset = c;
+					charset = WC_CES_US_ASCII;
+				    }
+				}
+			    }
+#endif
 			    /* fall thru, "META" is prohibit tag */
 			case HTML_HEAD:
 			case HTML_N_HEAD:
@@ -749,8 +763,8 @@ createFrameFile(struct frameset *f, FILE * f1, Buffer *current, int level,
 				if (!tag->value[j])
 				    break;
 				tag->value[j] =
-				    url_quote_conv(remove_space(tag->value[j]),
-						   code);
+				    url_encode(remove_space(tag->value[j]),
+					       &base, charset);
 				tag->need_reconstruct = TRUE;
 				parseURL2(tag->value[j], &url, &base);
 				if (url.scheme == SCM_UNKNOWN ||
@@ -764,12 +778,13 @@ createFrameFile(struct frameset *f, FILE * f1, Buffer *current, int level,
 				parsedtag_set_value(tag,
 						    ATTR_REFERER,
 						    parsedURL2Str(&base)->ptr);
-#ifdef JP_CHARSET
-				if (code != '\0') {
-				    charset[0] = code;
+#ifdef USE_M17N
+				if (tag->attrid[j] == ATTR_ACTION &&
+				    charset != WC_CES_US_ASCII)
 				    parsedtag_set_value(tag,
-							ATTR_CHARSET, charset);
-				}
+							ATTR_CHARSET,
+							wc_ces_to_charset
+							(charset));
 #endif
 				break;
 			    case ATTR_TARGET:
@@ -863,29 +878,38 @@ renderFrame(Buffer *Cbuf, int force_reload)
     Buffer *buf;
     int flag;
     struct frameset *fset;
+#ifdef USE_M17N
+    wc_ces doc_charset = DocumentCharset;
+#endif
 
     tmp = tmpfname(TMPF_FRAME, ".html");
     f = fopen(tmp->ptr, "w");
     if (f == NULL)
 	return NULL;
-    /* 
-     * if (Cbuf->frameQ != NULL) fset = Cbuf->frameQ->frameset; else */
     fset = Cbuf->frameset;
-    if (fset == NULL || createFrameFile(fset, f, Cbuf, 0, force_reload) < 0)
+    if (fset == NULL || createFrameFile(fset, f, Cbuf, 0, force_reload) < 0) {
+	fclose(f);
 	return NULL;
+    }
     fclose(f);
     flag = RG_FRAME;
     if ((Cbuf->currentURL).is_nocache)
 	flag |= RG_NOCACHE;
     renderFrameSet = Cbuf->frameset;
     flushFrameSet(renderFrameSet);
+#ifdef USE_M17N
+    DocumentCharset = InnerCharset;
+#endif
     buf = loadGeneralFile(tmp->ptr, NULL, NULL, flag, NULL);
+#ifdef USE_M17N
+    DocumentCharset = doc_charset;
+#endif
     renderFrameSet = NULL;
     if (buf == NULL || buf == NO_BUFFER)
 	return NULL;
     buf->sourcefile = tmp->ptr;
-#ifdef JP_CHARSET
-    buf->document_code = Cbuf->document_code;
+#ifdef USE_M17N
+    buf->document_charset = Cbuf->document_charset;
 #endif
     copyParsedURL(&buf->currentURL, &Cbuf->currentURL);
     preFormUpdateBuffer(buf);
