@@ -1,20 +1,20 @@
-/* $Id: html.h,v 1.31 2010/08/14 01:29:40 htrb Exp $ */
+/* vi: set sw=4 ts=8 ai sm noet : */
 #ifndef _HTML_H
 #define _HTML_H
+#include "config.h"
 #ifdef USE_SSL
 #include <openssl/bio.h>
 #include <openssl/x509.h>
 #include <openssl/ssl.h>
 #endif				/* USE_SSL */
 
-#include "istream.h"
+#include <time.h>
 
 #define StrUFgets(f) StrISgets((f)->stream)
 #define StrmyUFgets(f) StrmyISgets((f)->stream)
 #define UFgetc(f) ISgetc((f)->stream)
 #define UFundogetc(f) ISundogetc((f)->stream)
-#define UFread(f,buf,len) ISread((f)->stream,buf,len)
-#define UFclose(f) (void)(ISclose((f)->stream) == 0 && ((f)->stream = NULL))
+#define UFclose(f) if (ISclose((f)->stream) == 0) {(f)->stream = NULL ;}
 #define UFfileno(f) ISfileno((f)->stream)
 
 struct cmdtable {
@@ -62,11 +62,12 @@ typedef struct _ParsedURL {
     int is_nocache;
 } ParsedURL;
 
+union input_stream;
 typedef struct {
     unsigned char scheme;
     char is_cgi;
     char encoding;
-    InputStream stream;
+    union input_stream *stream;
     char *ext;
     int compression;
     int content_encoding;
@@ -83,6 +84,7 @@ typedef struct {
 #define CMP_GZIP         2
 #define CMP_BZIP2        3
 #define CMP_DEFLATE      4
+#define CMP_BROTLI       5
 
 #define ENC_7BIT	0
 #define ENC_BASE64	1
@@ -214,21 +216,34 @@ typedef struct {
 #define HTML_BIG        122
 #define HTML_N_BIG      123
 #define HTML_BUTTON     124
-#define HTML_FIELDSET   125
-#define HTML_N_FIELDSET 126
-#define HTML_IFRAME     127
-#define HTML_LABEL      128
-#define HTML_N_LABEL    129
-#define HTML_LEGEND     130
-#define HTML_N_LEGEND   131
-#define HTML_NOSCRIPT   132
-#define HTML_N_NOSCRIPT 133
-#define HTML_OBJECT     134
-#define HTML_OPTGROUP   135
-#define HTML_N_OPTGROUP 136
-#define HTML_PARAM      137
-#define HTML_SMALL      138
-#define HTML_N_SMALL    139
+#define HTML_N_BUTTON   125
+#define HTML_FIELDSET   126
+#define HTML_N_FIELDSET 127
+#define HTML_IFRAME     128
+#define HTML_LABEL      129
+#define HTML_N_LABEL    130
+#define HTML_LEGEND     131
+#define HTML_N_LEGEND   132
+#define HTML_NOSCRIPT   133
+#define HTML_N_NOSCRIPT 134
+#define HTML_OBJECT     135
+#define HTML_OPTGROUP   136
+#define HTML_N_OPTGROUP 137
+#define HTML_PARAM      138
+#define HTML_SMALL      139
+#define HTML_N_SMALL    140
+#define HTML_FIGURE     141
+#define HTML_N_FIGURE   142
+#define HTML_FIGCAPTION 143
+#define HTML_N_FIGCAPTION   144
+#define HTML_SECTION    145
+#define HTML_N_SECTION  146
+#define HTML_N_DT       147
+#define HTML_N_DD       148
+#define HTML_DFN        149
+#define HTML_N_DFN      150
+#define HTML_MAIN	151
+#define HTML_N_MAIN	152
 
    /* pseudo tag */
 #define HTML_SELECT_INT     160
@@ -312,6 +327,8 @@ typedef struct {
 #define ATTR_REV		48
 #define ATTR_TITLE		49
 #define ATTR_ACCESSKEY		50
+#define ATTR_PUBLIC		51
+#define ATTR_ARIALBL		52
 
 /* Internal attribute */
 #define ATTR_XOFFSET		60
@@ -335,10 +352,10 @@ typedef struct {
 /* HTML Tag Information Table */
 
 typedef struct html_tag_info {
-    char *name;
-    unsigned char *accept_attribute;
-    unsigned char max_attribute;
-    unsigned char flag;
+    const char *name;
+    const unsigned char *accept_attribute;
+    const unsigned char max_attribute;
+    const unsigned char flag;
 } TagInfo;
 
 #define TFLG_END	1
@@ -347,9 +364,9 @@ typedef struct html_tag_info {
 /* HTML Tag Attribute Information Table */
 
 typedef struct tag_attribute_info {
-    char *name;
-    unsigned char vtype;
-    unsigned char flag;
+    const char *name;
+    const unsigned char vtype;
+    const unsigned char flag;
 } TagAttrInfo;
 
 #define AFLG_INT	1
@@ -387,6 +404,10 @@ struct environment {
 
 #define INDENT_INCR IndentIncr
 
+/*
+ * Those SCM_ define the indeces of DefaultPort in url.c and scheme_str[] in
+ * _parsedURL2Str in url.c
+ */
 #define SCM_UNKNOWN	255
 #define SCM_MISSING	254
 #define SCM_HTTP	0
@@ -402,8 +423,9 @@ struct environment {
 #define SCM_NEWS_GROUP	10
 #define SCM_DATA	11
 #define SCM_MAILTO      12
+#define SCM_GOPHERS	13
 #ifdef USE_SSL
-#define SCM_HTTPS       13
+#define SCM_HTTPS       14
 #endif				/* USE_SSL */
 
 #endif				/* _HTML_H */
