@@ -1,10 +1,11 @@
-/* $Id: image.c,v 1.37 2010/12/21 10:13:55 htrb Exp $ */
+/* vi: set sw=4 ts=8 ai sm noet : */
 
 #include "fm.h"
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <signal.h>
 #include <errno.h>
+#include <strings.h>
 #include <unistd.h>
 #ifdef HAVE_WAITPID
 #include <sys/wait.h>
@@ -31,12 +32,12 @@ static int n_terminal_image = 0;
 static int max_terminal_image = 0;
 static FILE *Imgdisplay_rf = NULL, *Imgdisplay_wf = NULL;
 static pid_t Imgdisplay_pid = 0;
-static int openImgdisplay();
-static void closeImgdisplay();
-int getCharSize();
+static int openImgdisplay(void);
+static void closeImgdisplay(void);
+static int getCharSize(void);
 
 void
-initImage()
+initImage(void)
 {
     if (activeImage)
 	return;
@@ -44,14 +45,32 @@ initImage()
 	activeImage = TRUE;
 }
 
-int
-getCharSize()
+static int
+getCharSize(void)
 {
     FILE *f;
     Str tmp;
     int w = 0, h = 0;
 
     set_environ("W3M_TTY", ttyname_tty());
+
+    if (enable_inline_image) {
+	int ppc, ppl;
+
+	if (get_pixel_per_cell(&ppc,&ppl)) {
+	    pixel_per_char_i = ppc ;
+	    pixel_per_line_i = ppl ;
+	    pixel_per_char = (double)ppc;
+	    pixel_per_line = (double)ppl;
+	}
+	else {
+	    pixel_per_char_i = (int)pixel_per_char;
+	    pixel_per_line_i = (int)pixel_per_line;
+	}
+
+	return  TRUE;
+    }
+
     tmp = Strnew();
     if (!strchr(Imgdisplay, '/'))
 	Strcat_m_charp(tmp, w3m_auxbin_dir(), "/", NULL);
@@ -75,7 +94,7 @@ getCharSize()
 }
 
 void
-termImage()
+termImage(void)
 {
     if (!activeImage)
 	return;
@@ -88,19 +107,20 @@ termImage()
 }
 
 static int
-openImgdisplay()
+openImgdisplay(void)
 {
+    char *cmd;
+
+    if (!strchr(Imgdisplay, '/'))
+	cmd = Strnew_m_charp(w3m_auxbin_dir(), "/", Imgdisplay, NULL)->ptr;
+    else
+	cmd = Imgdisplay;
     Imgdisplay_pid = open_pipe_rw(&Imgdisplay_rf, &Imgdisplay_wf);
     if (Imgdisplay_pid < 0)
 	goto err0;
     if (Imgdisplay_pid == 0) {
 	/* child */
-	char *cmd;
 	setup_child(FALSE, 2, -1);
-	if (!strchr(Imgdisplay, '/'))
-	    cmd = Strnew_m_charp(w3m_auxbin_dir(), "/", Imgdisplay, NULL)->ptr;
-	else
-	    cmd = Imgdisplay;
 	myExec(cmd);
 	/* XXX: ifdef __EMX__, use start /f ? */
     }
@@ -113,7 +133,7 @@ openImgdisplay()
 }
 
 static void
-closeImgdisplay()
+closeImgdisplay(void)
 {
     if (Imgdisplay_wf)
 	fclose(Imgdisplay_wf);
@@ -155,6 +175,10 @@ addImage(ImageCache * cache, int x, int y, int sx, int sy, int w, int h)
 static void
 syncImage(void)
 {
+    if (enable_inline_image) {
+	return;
+    }
+
     fputs("3;\n", Imgdisplay_wf);	/* XSync() */
     fputs("4;\n", Imgdisplay_wf);	/* put '\n' */
     while (fflush(Imgdisplay_wf) != 0) {
@@ -171,11 +195,12 @@ syncImage(void)
 }
 
 void
-drawImage()
+drawImage(void)
 {
     static char buf[64];
     int j, draw = FALSE;
     TerminalImage *i;
+    struct stat st ;
 
     if (!activeImage)
 	return;
@@ -183,6 +208,50 @@ drawImage()
 	return;
     for (j = 0; j < n_terminal_image; j++) {
 	i = &terminal_image[j];
+
+	if (enable_inline_image) {
+	    /*
+	     * So this shouldn't ever happen, but if it does then at least let's
+	     * not have external programs fetch images from the Internet...
+	     */
+	    if (!i->cache->touch || stat(i->cache->file,&st))
+	      return;
+
+	    char *url = i->cache->file;
+
+	    int x = i->x / pixel_per_char_i;
+	    int y = i->y / pixel_per_line_i;
+
+	    int w = i->cache->a_width > 0 ? (
+		    (i->cache->width + i->x % pixel_per_char_i + pixel_per_char_i - 1) /
+		    pixel_per_char_i) : 0;
+	    int h = i->cache->a_height > 0 ? (
+		    (i->cache->height + i->y % pixel_per_line_i + pixel_per_line_i - 1) /
+		    pixel_per_line_i) : 0;
+
+	    int sx = i->sx / pixel_per_char_i;
+	    int sy = i->sy / pixel_per_line_i;
+
+	    int sw = (i->width + i->sx % pixel_per_char_i + pixel_per_char_i - 1) /
+		      pixel_per_char_i;
+	    int sh = (i->height + i->sy % pixel_per_line_i + pixel_per_line_i - 1) /
+		      pixel_per_line_i;
+
+	    if (enable_inline_image == INLINE_IMG_SIXEL) {
+		w = i->cache->a_width > 0 ? i->width : 0;
+		h = i->cache->a_height > 0 ? i->height : 0;
+		put_image_sixel(url, x, y, w, h, i->sx, i->sy, sw * pixel_per_char, sh * pixel_per_line_i, n_terminal_image);
+	    } else if (enable_inline_image == INLINE_IMG_OSC5379) {
+		put_image_osc5379(url, x, y, w, h, sx, sy, sw, sh);
+	    } else if (enable_inline_image == INLINE_IMG_ITERM2) {
+		put_image_iterm2(url, x, y, sw, sh);
+	    } else if (enable_inline_image == INLINE_IMG_KITTY) {
+		put_image_kitty(url, x, y, i->width, i->height, i->sx, i->sy, sw * pixel_per_char, sh * pixel_per_line_i, sw, sh);
+	    }
+
+	    continue ;
+	}
+
 	if (!(i->cache->loaded & IMG_FLAG_LOADED &&
 	      i->width > 0 && i->height > 0))
 	    continue;
@@ -206,15 +275,21 @@ drawImage()
 	fputs("\n", Imgdisplay_wf);
 	draw = TRUE;
     }
-    if (!draw)
-	return;
-    syncImage();
+
+    if (!enable_inline_image) {
+	if (!draw)
+	    return;
+	syncImage();
+    }
+    else
+	n_terminal_image = 0;
+
     touch_cursor();
     refresh();
 }
 
 void
-clearImage()
+clearImage(void)
 {
     static char buf[64];
     int j;
@@ -300,7 +375,7 @@ getAllImage(Buffer *buf)
     }
 }
 
-void
+static void
 showImageProgress(Buffer *buf)
 {
     AnchorList *al;
@@ -320,6 +395,8 @@ showImageProgress(Buffer *buf)
 	}
     }
     if (n) {
+        if (enable_inline_image && n == l)
+	    drawImage();
 	message(Sprintf("%d/%d images loaded", l, n)->ptr,
 		buf->cursorX + buf->rootX, buf->cursorY + buf->rootY);
 	refresh();
@@ -333,6 +410,9 @@ loadImage(Buffer *buf, int flag)
     struct stat st;
     int i, draw = FALSE;
     /* int wait_st; */
+#ifdef DONT_CALL_GC_AFTER_FORK
+    char *loadargs[7];
+#endif
 
     if (maxLoadImage > MAX_LOAD_IMAGE)
 	maxLoadImage = MAX_LOAD_IMAGE;
@@ -346,19 +426,12 @@ loadImage(Buffer *buf, int flag)
     }
     for (i = 0; i < n_load_image; i++) {
 	cache = image_cache[i];
-	if (!cache)
+	if (!cache || !cache->touch)
 	    continue;
 	if (lstat(cache->touch, &st))
 	    continue;
 	if (cache->pid) {
 	    kill(cache->pid, SIGKILL);
-	    /*
-	     * #ifdef HAVE_WAITPID
-	     * waitpid(cache->pid, &wait_st, 0);
-	     * #else
-	     * wait(&wait_st);
-	     * #endif
-	     */
 	    cache->pid = 0;
 	}
 	if (!stat(cache->file, &st)) {
@@ -377,20 +450,12 @@ loadImage(Buffer *buf, int flag)
 
     for (i = (buf != image_buffer) ? 0 : maxLoadImage; i < n_load_image; i++) {
 	cache = image_cache[i];
-	if (!cache)
+	if (!cache || !cache->touch)
 	    continue;
 	if (cache->pid) {
 	    kill(cache->pid, SIGKILL);
-	    /*
-	     * #ifdef HAVE_WAITPID
-	     * waitpid(cache->pid, &wait_st, 0);
-	     * #else
-	     * wait(&wait_st);
-	     * #endif
-	     */
 	    cache->pid = 0;
 	}
-	unlink(cache->touch);
 	image_cache[i] = NULL;
     }
 
@@ -403,7 +468,8 @@ loadImage(Buffer *buf, int flag)
     }
 
     if (draw && image_buffer) {
-	drawImage();
+        if (!enable_inline_image)
+	    drawImage();
 	showImageProgress(image_buffer);
     }
 
@@ -431,18 +497,33 @@ loadImage(Buffer *buf, int flag)
 		break;
 	}
 	image_cache[i] = cache;
+	if (!cache->touch) {
+	    continue;
+	}
 
 	flush_tty();
+#ifdef DONT_CALL_GC_AFTER_FORK
+	loadargs[0] = MyProgramName;
+	loadargs[1] = "-$$getimage";
+	loadargs[2] = conv_to_system(cache->url);
+	loadargs[3] = conv_to_system(parsedURL2Str(cache->current)->ptr);
+	loadargs[4] = cache->file;
+	loadargs[5] = cache->touch;
+	loadargs[6] = NULL;
+	if ((cache->pid = fork()) == 0) {
+	    setup_child(FALSE, 0, -1);
+	    execvp(MyProgramName, loadargs);
+	    exit(1);
+	}
+	else if (cache->pid < 0) {
+	    cache->pid = 0;
+	    return;
+	}
+#else /* !DONT_CALL_GC_AFTER_FORK */
 	if ((cache->pid = fork()) == 0) {
-	    Buffer *b;
-	    /*
-	     * setup_child(TRUE, 0, -1);
-	     */
 	    setup_child(FALSE, 0, -1);
 	    image_source = cache->file;
-	    b = loadGeneralFile(cache->url, cache->current, NULL, 0, NULL);
-	    if (!b || !b->real_type || strncasecmp(b->real_type, "image/", 6))
-		unlink(cache->file);
+	    loadGeneralFile(cache->url, cache->current, NULL, 0, NULL);
 #if defined(HAVE_SYMLINK) && defined(HAVE_LSTAT)
 	    symlink(cache->file, cache->touch);
 #else
@@ -458,6 +539,7 @@ loadImage(Buffer *buf, int flag)
 	    cache->pid = 0;
 	    return;
 	}
+#endif /* !DONT_CALL_GC_AFTER_FORK */
     }
 }
 
@@ -492,23 +574,33 @@ getImage(Image * image, ParsedURL *current, int flag)
 	cache->url = image->url;
 	cache->current = current;
 	cache->file = tmpfname(TMPF_DFL, image->ext)->ptr;
-	cache->touch = tmpfname(TMPF_DFL, NULL)->ptr;
 	cache->pid = 0;
 	cache->index = 0;
 	cache->loaded = IMG_FLAG_UNLOADED;
-	cache->width = image->width;
-	cache->height = image->height;
-	putHash_sv(image_hash, key->ptr, (void *)cache);
+	if (enable_inline_image == INLINE_IMG_OSC5379) {
+	    if (image->width > 0 && image->width % pixel_per_char_i > 0)
+		image->width += (pixel_per_char_i - image->width % pixel_per_char_i);
+
+	    if (image->height > 0 && image->height % pixel_per_line_i > 0)
+		image->height += (pixel_per_line_i - image->height % pixel_per_line_i);
+	}
+	cache->touch = tmpfname(TMPF_DFL, NULL)->ptr;
+
+	cache->width = image->width ;
+	cache->height = image->height ;
+	cache->a_width = image->width;
+	cache->a_height = image->height;
+	putHash_sv(image_hash, key->ptr, cache);
     }
     if (flag != IMG_FLAG_SKIP) {
 	if (cache->loaded == IMG_FLAG_UNLOADED) {
 	    if (!image_file)
 		image_file = newHash_sv(100);
 	    if (!getHash_sv(image_file, cache->file, NULL)) {
-		putHash_sv(image_file, cache->file, (void *)cache);
+		putHash_sv(image_file, cache->file, cache);
 		if (!image_list)
 		    image_list = newGeneralList();
-		pushValue(image_list, (void *)cache);
+		pushValue(image_list, cache);
 	    }
 	}
 	if (!cache->index)
@@ -519,18 +611,94 @@ getImage(Image * image, ParsedURL *current, int flag)
     return cache;
 }
 
+static int
+parseImageHeader(char *path, u_int *width, u_int *height)
+{
+    FILE *fp;
+    u_char buf[8];
+
+    if (!(fp = fopen(path, "r"))) return FALSE;
+
+    if (fread(buf, 1, 2, fp) != 2) goto error;
+
+    if (memcmp(buf, "\xff\xd8", 2) == 0) {
+        /* JPEG */
+	if (fseek(fp, 2, SEEK_CUR) < 0)	goto error;   /* 0xffe0 */
+	while (fread(buf, 1, 2, fp) == 2) {
+	    size_t len = ((buf[0] << 8) | buf[1]) - 2;
+	    if (fseek(fp, len, SEEK_CUR) < 0) goto error;
+	    if (fread(buf, 1, 2, fp) == 2 &&
+	        /* SOF0 or SOF2 */
+	        (memcmp(buf, "\xff\xc0", 2) == 0 || memcmp(buf, "\xff\xc2", 2) == 0)) {
+		fseek(fp, 3, SEEK_CUR);
+		if (fread(buf, 1, 2, fp) == 2) {
+		    *height = (buf[0] << 8) | buf[1];
+		    if (fread(buf, 1, 2, fp) == 2) {
+			*width = (buf[0] << 8) | buf[1];
+			goto success;
+		    }
+		}
+		break;
+	    }
+	}
+	goto error;
+    }
+
+    if (fread(buf + 2, 1, 1, fp) != 1) goto error;
+
+    if (memcmp(buf, "GIF", 3) == 0) {
+        /* GIF */
+	if (fseek(fp, 3, SEEK_CUR) < 0) goto error;
+	if (fread(buf, 1, 2, fp) == 2) {
+	    *width = (buf[1] << 8) | buf[0];
+	    if (fread(buf, 1, 2, fp) == 2) {
+		*height = (buf[1] << 8) | buf[0];
+		goto success;
+	    }
+	}
+	goto error;
+    }
+
+    if (fread(buf + 3, 1, 5, fp) != 5) goto error;
+
+    if (memcmp(buf, "\x89\x50\x4e\x47\x0d\x0a\x1a\x0a", 8) == 0) {
+	/* PNG */
+	if (fseek(fp, 8, SEEK_CUR) < 0) goto error;
+	if (fread(buf, 1, 4, fp) == 4) {
+	    *width = (buf[0] << 24) | (buf[1] << 16) | (buf[2] << 8) | buf[3];
+	    if (fread(buf, 1, 4, fp) == 4) {
+		*height = (buf[0] << 24) | (buf[1] << 16) | (buf[2] << 8) | buf[3];
+		goto success;
+	    }
+	}
+	goto error;
+    }
+
+error:
+    fclose(fp);
+    return FALSE;
+
+success:
+    fclose(fp);
+    return TRUE;
+}
+
 int
 getImageSize(ImageCache * cache)
 {
     Str tmp;
     FILE *f;
-    int w = 0, h = 0;
+    unsigned int w = 0, h = 0;
 
     if (!activeImage)
 	return FALSE;
     if (!cache || !(cache->loaded & IMG_FLAG_LOADED) ||
 	(cache->width > 0 && cache->height > 0))
 	return FALSE;
+
+    if (parseImageHeader(cache->file, &w, &h))
+	goto got_image_size;
+
     tmp = Strnew();
     if (!strchr(Imgdisplay, '/'))
 	Strcat_m_charp(tmp, w3m_auxbin_dir(), "/", NULL);
@@ -538,7 +706,7 @@ getImageSize(ImageCache * cache)
     f = popen(tmp->ptr, "r");
     if (!f)
 	return FALSE;
-    while (fscanf(f, "%d %d", &w, &h) < 0) {
+    while (fscanf(f, "%u %u", &w, &h) < 0) {
 	if (feof(f))
 	    break;
     }
@@ -546,6 +714,8 @@ getImageSize(ImageCache * cache)
 
     if (!(w > 0 && h > 0))
 	return FALSE;
+
+got_image_size:
     w = (int)(w * image_scale / 100 + 0.5);
     if (w == 0)
 	w = 1;
@@ -558,18 +728,18 @@ getImageSize(ImageCache * cache)
     }
     else if (cache->width < 0) {
 	int tmp = (int)((double)cache->height * w / h + 0.5);
-	cache->width = (tmp > MAX_IMAGE_SIZE) ? MAX_IMAGE_SIZE : tmp;
+	cache->a_width = cache->width = (tmp > MAX_IMAGE_SIZE) ? MAX_IMAGE_SIZE : tmp;
     }
     else if (cache->height < 0) {
 	int tmp = (int)((double)cache->width * h / w + 0.5);
-	cache->height = (tmp > MAX_IMAGE_SIZE) ? MAX_IMAGE_SIZE : tmp;
+	cache->a_height = cache->height = (tmp > MAX_IMAGE_SIZE) ? MAX_IMAGE_SIZE : tmp;
     }
     if (cache->width == 0)
 	cache->width = 1;
     if (cache->height == 0)
 	cache->height = 1;
     tmp = Sprintf("%d;%d;%s", cache->width, cache->height, cache->url);
-    putHash_sv(image_hash, tmp->ptr, (void *)cache);
+    putHash_sv(image_hash, tmp->ptr, cache);
     return TRUE;
 }
 #endif
