@@ -1,4 +1,4 @@
-/* $Id: istream.c,v 1.27 2010/07/18 13:43:23 htrb Exp $ */
+/* vi: set sw=4 ts=8 ai sm noet : */
 #include "fm.h"
 #include "myctype.h"
 #include "istream.h"
@@ -19,28 +19,29 @@
 
 #define POP_CHAR(bs) ((bs)->iseos?'\0':(bs)->stream.buf[(bs)->stream.cur++])
 
-static void basic_close(int *handle);
-static int basic_read(int *handle, char *buf, int len);
+static int basic_close(int *handle);
+static int basic_read(int *handle, unsigned char *buf, int len);
 
-static void file_close(struct file_handle *handle);
-static int file_read(struct file_handle *handle, char *buf, int len);
+static int file_read(FILE *handle, char *buf, int len);
 
 static int str_read(Str handle, char *buf, int len);
 
 #ifdef USE_SSL
-static void ssl_close(struct ssl_handle *handle);
+static int ssl_close(struct ssl_handle *handle);
 static int ssl_read(struct ssl_handle *handle, char *buf, int len);
 #endif
 
 static int ens_read(struct ens_handle *handle, char *buf, int len);
-static void ens_close(struct ens_handle *handle);
+static int ens_close(struct ens_handle *handle);
+
+static void memchop(char *p, int *len);
 
 static void
 do_update(BaseStream base)
 {
     int len;
     base->stream.cur = base->stream.next = 0;
-    len = base->read(base->handle, base->stream.buf, base->stream.size);
+    len = (*base->read) (base->handle, base->stream.buf, base->stream.size);
     if (len <= 0)
 	base->iseos = TRUE;
     else
@@ -48,30 +49,30 @@ do_update(BaseStream base)
 }
 
 static int
-buffer_read(StreamBuffer sb, char *obuf, int count)
+buffer_read(StreamBuffer sb, unsigned char *obuf, int count)
 {
     int len = sb->next - sb->cur;
     if (len > 0) {
 	if (len > count)
 	    len = count;
-	bcopy((const void *)&sb->buf[sb->cur], obuf, len);
+	memmove(obuf, (const void *)&sb->buf[sb->cur], len);
 	sb->cur += len;
     }
     return len;
 }
 
 static void
-init_buffer(BaseStream base, char *buf, int bufsize)
+init_buffer(BaseStream base, const char *buf, int bufsize)
 {
     StreamBuffer sb = &base->stream;
     sb->size = bufsize;
     sb->cur = 0;
+    sb->buf = NewWithoutGC_N(uchar, bufsize);
     if (buf) {
-	sb->buf = (uchar *) buf;
+	memcpy(sb->buf, buf, bufsize);
 	sb->next = bufsize;
     }
     else {
-	sb->buf = NewAtom_N(uchar, bufsize);
 	sb->next = 0;
     }
     base->iseos = FALSE;
@@ -95,33 +96,29 @@ newInputStream(int des)
     InputStream stream;
     if (des < 0)
 	return NULL;
-    stream = New(union input_stream);
+    stream = NewWithoutGC(union input_stream);
     init_base_stream(&stream->base, STREAM_BUF_SIZE);
     stream->base.type = IST_BASIC;
-    stream->base.handle = New(int);
+    stream->base.handle = NewWithoutGC(int);
+    /* TODO(rkta): Check cast from int to void ptr */
     *(int *)stream->base.handle = des;
-    stream->base.read = (int (*)())basic_read;
-    stream->base.close = (void (*)())basic_close;
+    stream->base.read = basic_read;
+    stream->base.close = basic_close;
     return stream;
 }
 
 InputStream
-newFileStream(FILE * f, void (*closep) ())
+newFileStream(FILE * f, int (*closep)(FILE *))
 {
     InputStream stream;
     if (f == NULL)
 	return NULL;
-    stream = New(union input_stream);
+    stream = NewWithoutGC(union input_stream);
     init_base_stream(&stream->base, STREAM_BUF_SIZE);
     stream->file.type = IST_FILE;
-    stream->file.handle = New(struct file_handle);
-    stream->file.handle->f = f;
-    if (closep)
-	stream->file.handle->close = closep;
-    else
-	stream->file.handle->close = (void (*)())fclose;
-    stream->file.read = (int (*)())file_read;
-    stream->file.close = (void (*)())file_close;
+    stream->file.handle = f;
+    stream->file.close = closep;
+    stream->file.read = file_read;
     return stream;
 }
 
@@ -131,11 +128,11 @@ newStrStream(Str s)
     InputStream stream;
     if (s == NULL)
 	return NULL;
-    stream = New(union input_stream);
+    stream = NewWithoutGC(union input_stream);
     init_str_stream(&stream->base, s);
     stream->str.type = IST_STR;
-    stream->str.handle = s;
-    stream->str.read = (int (*)())str_read;
+    stream->str.handle = NULL;
+    stream->str.read = str_read;
     stream->str.close = NULL;
     return stream;
 }
@@ -147,14 +144,14 @@ newSSLStream(SSL * ssl, int sock)
     InputStream stream;
     if (sock < 0)
 	return NULL;
-    stream = New(union input_stream);
+    stream = NewWithoutGC(union input_stream);
     init_base_stream(&stream->base, SSL_BUF_SIZE);
     stream->ssl.type = IST_SSL;
-    stream->ssl.handle = New(struct ssl_handle);
+    stream->ssl.handle = NewWithoutGC(struct ssl_handle);
     stream->ssl.handle->ssl = ssl;
     stream->ssl.handle->sock = sock;
-    stream->ssl.read = (int (*)())ssl_read;
-    stream->ssl.close = (void (*)())ssl_close;
+    stream->ssl.read = ssl_read;
+    stream->ssl.close = ssl_close;
     return stream;
 }
 #endif
@@ -166,29 +163,35 @@ newEncodedStream(InputStream is, char encoding)
     if (is == NULL || (encoding != ENC_QUOTE && encoding != ENC_BASE64 &&
 		       encoding != ENC_UUENCODE))
 	return is;
-    stream = New(union input_stream);
+    stream = NewWithoutGC(union input_stream);
     init_base_stream(&stream->base, STREAM_BUF_SIZE);
     stream->ens.type = IST_ENCODED;
-    stream->ens.handle = New(struct ens_handle);
+    stream->ens.handle = NewWithoutGC(struct ens_handle);
     stream->ens.handle->is = is;
     stream->ens.handle->pos = 0;
     stream->ens.handle->encoding = encoding;
-    stream->ens.handle->s = NULL;
-    stream->ens.read = (int (*)())ens_read;
-    stream->ens.close = (void (*)())ens_close;
+    growbuf_init_without_GC(&stream->ens.handle->gb);
+    stream->ens.read = ens_read;
+    stream->ens.close = ens_close;
     return stream;
 }
 
 int
 ISclose(InputStream stream)
 {
-    MySignalHandler(*prevtrap) ();
-    if (stream == NULL || stream->base.close == NULL ||
-	stream->base.type & IST_UNCLOSE)
-	return -1;
-    prevtrap = mySignal(SIGINT, SIG_IGN);
-    stream->base.close(stream->base.handle);
-    mySignal(SIGINT, prevtrap);
+    void (*prevtrap) (SIGNAL_ARG);
+    if (stream == NULL)
+        return -1;
+    if (stream->base.close != NULL) {
+        if (stream->base.type & IST_UNCLOSE) {
+            return -1;
+        }
+        prevtrap = mySignal(SIGINT, SIG_IGN);
+        stream->base.close(stream->base.handle);
+        mySignal(SIGINT, prevtrap);
+    }
+    free(stream->base.stream.buf);
+    free(stream);
     return 0;
 }
 
@@ -218,122 +221,77 @@ ISundogetc(InputStream stream)
     return -1;
 }
 
-#define MARGIN_STR_SIZE 10
 Str
-StrISgets(InputStream stream)
+StrISgets2(InputStream stream, char crnl)
 {
-    BaseStream base;
-    StreamBuffer sb;
-    Str s = NULL;
-    uchar *p;
-    int len;
+    struct growbuf gb;
 
     if (stream == NULL)
-	return '\0';
-    base = &stream->base;
-    sb = &base->stream;
-
-    while (!base->iseos) {
-	if (MUST_BE_UPDATED(base)) {
-	    do_update(base);
-	}
-	else {
-	    if ((p = memchr(&sb->buf[sb->cur], '\n', sb->next - sb->cur))) {
-		len = p - &sb->buf[sb->cur] + 1;
-		if (s == NULL)
-		    s = Strnew_size(len);
-		Strcat_charp_n(s, (char *)&sb->buf[sb->cur], len);
-		sb->cur += len;
-		return s;
-	    }
-	    else {
-		if (s == NULL)
-		    s = Strnew_size(sb->next - sb->cur + MARGIN_STR_SIZE);
-		Strcat_charp_n(s, (char *)&sb->buf[sb->cur],
-			       sb->next - sb->cur);
-		sb->cur = sb->next;
-	    }
-	}
-    }
-
-    if (s == NULL)
-	return Strnew();
-    return s;
+	return NULL;
+    growbuf_init(&gb);
+    ISgets_to_growbuf(stream, &gb, crnl);
+    return growbuf_to_Str(&gb);
 }
 
-Str
-StrmyISgets(InputStream stream)
+void
+ISgets_to_growbuf(InputStream stream, struct growbuf *gb, char crnl)
 {
-    BaseStream base;
-    StreamBuffer sb;
-    Str s = NULL;
-    int i, len;
+    BaseStream base = &stream->base;
+    StreamBuffer sb = &base->stream;
+    int i;
 
-    if (stream == NULL)
-	return '\0';
-    base = &stream->base;
-    sb = &base->stream;
+    gb->length = 0;
 
     while (!base->iseos) {
 	if (MUST_BE_UPDATED(base)) {
 	    do_update(base);
+	    continue;
 	}
-	else {
-	    if (s && Strlastchar(s) == '\r') {
-		if (sb->buf[sb->cur] == '\n')
-		    Strcat_char(s, (char)sb->buf[sb->cur++]);
-		return s;
+	if (crnl && gb->length > 0  && gb->ptr[gb->length - 1] == '\r') {
+	    if (sb->buf[sb->cur] == '\n') {
+		GROWBUF_ADD_CHAR(gb, '\n');
+		++sb->cur;
 	    }
-	    for (i = sb->cur;
-		 i < sb->next && sb->buf[i] != '\n' && sb->buf[i] != '\r';
-		 i++) ;
-	    if (i < sb->next) {
-		len = i - sb->cur + 1;
-		if (s == NULL)
-		    s = Strnew_size(len + MARGIN_STR_SIZE);
-		Strcat_charp_n(s, (char *)&sb->buf[sb->cur], len);
-		sb->cur = i + 1;
-		if (sb->buf[i] == '\n')
-		    return s;
-	    }
-	    else {
-		if (s == NULL)
-		    s = Strnew_size(sb->next - sb->cur + MARGIN_STR_SIZE);
-		Strcat_charp_n(s, (char *)&sb->buf[sb->cur],
-			       sb->next - sb->cur);
-		sb->cur = sb->next;
+	    break;
+	}
+	for (i = sb->cur; i < sb->next; ++i) {
+	    if (sb->buf[i] == '\n' || (crnl && sb->buf[i] == '\r')) {
+		++i;
+		break;
 	    }
 	}
+	growbuf_append(gb, &sb->buf[sb->cur], i - sb->cur);
+	sb->cur = i;
+	if (gb->length > 0 && gb->ptr[gb->length - 1] == '\n')
+	    break;
     }
 
-    if (s == NULL)
-	return Strnew();
-    return s;
+    growbuf_reserve(gb, gb->length + 1);
+    gb->ptr[gb->length] = '\0';
+    return;
 }
 
 int
-ISread(InputStream stream, Str buf, int count)
+ISread_n(InputStream stream, unsigned char *dst, int count)
 {
-    int rest, len;
+    int len, l;
     BaseStream base;
 
-    if (stream == NULL || (base = &stream->base)->iseos)
+    if (stream == NULL || count <= 0)
+	return -1;
+    if ((base = &stream->base)->iseos)
 	return 0;
 
-    len = buffer_read(&base->stream, buf->ptr, count);
-    rest = count - len;
+    len = buffer_read(&base->stream, dst, count);
     if (MUST_BE_UPDATED(base)) {
-	len = base->read(base->handle, &buf->ptr[len], rest);
-	if (len <= 0) {
+	l = (*base->read) (base->handle, &dst[len], count - len);
+	if (l <= 0) {
 	    base->iseos = TRUE;
-	    len = 0;
+	} else {
+	    len += l;
 	}
-	rest -= len;
     }
-    Strtruncate(buf, count - rest);
-    if (buf->length > 0)
-	return 1;
-    return 0;
+    return len;
 }
 
 int
@@ -345,7 +303,7 @@ ISfileno(InputStream stream)
     case IST_BASIC:
 	return *(int *)stream->base.handle;
     case IST_FILE:
-	return fileno(stream->file.handle->f);
+	return fileno(stream->file.handle);
 #ifdef USE_SSL
     case IST_SSL:
 	return stream->ssl.handle->sock;
@@ -357,15 +315,6 @@ ISfileno(InputStream stream)
     }
 }
 
-int
-ISeos(InputStream stream)
-{
-    BaseStream base = &stream->base;
-    if (!base->iseos && MUST_BE_UPDATED(base))
-	do_update(base);
-    return base->iseos;
-}
-
 #ifdef USE_SSL
 static Str accept_this_site;
 
@@ -379,7 +328,7 @@ ssl_accept_this_site(char *hostname)
 }
 
 static int
-ssl_match_cert_ident(char *ident, int ilen, char *hostname)
+ssl_match_cert_ident(const char *ident, int ilen, char *hostname)
 {
     /* RFC2818 3.1.  Server Identity
      * Names may contain the wildcard
@@ -435,39 +384,52 @@ ssl_check_cert_ident(X509 * x, char *hostname)
 	if (alt) {
 	    int n;
 	    GENERAL_NAME *gn;
-	    X509V3_EXT_METHOD *method;
 	    Str seen_dnsname = NULL;
 
 	    n = sk_GENERAL_NAME_num(alt);
 	    for (i = 0; i < n; i++) {
 		gn = sk_GENERAL_NAME_value(alt, i);
 		if (gn->type == GEN_DNS) {
-		    char *sn = ASN1_STRING_data(gn->d.ia5);
+#if (OPENSSL_VERSION_NUMBER < 0x10100000L) || defined(LIBRESSL_VERSION_NUMBER)
+		    unsigned char *sn = ASN1_STRING_data(gn->d.ia5);
+#else
+		    const unsigned char *sn = ASN1_STRING_get0_data(gn->d.ia5);
+#endif
 		    int sl = ASN1_STRING_length(gn->d.ia5);
 
+		    /*
+		     * sn is a pointer to internal data and not guaranteed to
+		     * be null terminated. Ensure we have a null terminated
+		     * string that we can modify.
+		     */
+		    char *asn = GC_MALLOC(sl + 1);
+		    if (!asn)
+			exit(1);
+		    memmove(asn, sn, sl);
+		    asn[sl] = '\0';
+
 		    if (!seen_dnsname)
 			seen_dnsname = Strnew();
 		    /* replace \0 to make full string visible to user */
-		    if (sl != strlen(sn)) {
+		    if (sl != strlen(asn)) {
 			int i;
 			for (i = 0; i < sl; ++i) {
-			    if (!sn[i])
-				sn[i] = '!';
+			    if (!asn[i])
+				asn[i] = '!';
 			}
 		    }
-		    Strcat_m_charp(seen_dnsname, sn, " ", NULL);
-		    if (sl == strlen(sn) /* catch \0 in SAN */
-			&& ssl_match_cert_ident(sn, sl, hostname))
+		    Strcat_m_charp(seen_dnsname, asn, " ", NULL);
+		    if (sl == strlen(asn) /* catch \0 in SAN */
+			&& ssl_match_cert_ident(asn, sl, hostname))
 			break;
 		}
 	    }
-	    method = X509V3_EXT_get(ex);
+	    X509V3_EXT_get(ex);
 	    sk_GENERAL_NAME_free(alt);
 	    if (i < n)		/* Found a match */
 		match_ident = TRUE;
 	    else if (seen_dnsname)
-		/* FIXME: gettextize? */
-		ret = Sprintf("Bad cert ident from %s: dNSName=%s", hostname,
+		ret = Sprintf(_("Bad cert ident from %s: dNSName=%s"), hostname,
 			      seen_dnsname->ptr);
 	}
     }
@@ -481,8 +443,7 @@ ssl_check_cert_ident(X509 * x, char *hostname)
 
 	slen = X509_NAME_get_text_by_NID(xn, NID_commonName, buf, sizeof(buf));
 	if ( slen == -1)
-	    /* FIXME: gettextize? */
-	    ret = Strnew_charp("Unable to get common name from peer cert");
+	    ret = Strnew_charp(_("Unable to get common name from peer cert"));
 	else if (slen != strlen(buf)
 		|| !ssl_match_cert_ident(buf, strlen(buf), hostname)) {
 	    /* replace \0 to make full string visible to user */
@@ -493,11 +454,8 @@ ssl_check_cert_ident(X509 * x, char *hostname)
 			buf[i] = '!';
 		}
 	    }
-	    /* FIXME: gettextize? */
-	    ret = Sprintf("Bad cert ident %s from %s", buf, hostname);
+	    ret = Sprintf(_("Bad cert ident %s from %s"), buf, hostname);
 	}
-	else
-	    match_ident = TRUE;
     }
     return ret;
 }
@@ -514,7 +472,7 @@ ssl_get_certificate(SSL * ssl, char *hostname)
     char buf[2048];
     Str amsg = NULL;
     Str emsg;
-    char *ans;
+    int ans;
 
     if (ssl == NULL)
 	return NULL;
@@ -522,19 +480,15 @@ ssl_get_certificate(SSL * ssl, char *hostname)
     if (x == NULL) {
 	if (accept_this_site
 	    && strcasecmp(accept_this_site->ptr, hostname) == 0)
-	    ans = "y";
-	else {
-	    /* FIXME: gettextize? */
-	    emsg = Strnew_charp("No SSL peer certificate: accept? (y/n)");
-	    ans = inputAnswer(emsg->ptr);
-	}
-	if (ans && TOLOWER(*ans) == 'y')
-	    /* FIXME: gettextize? */
+	    ans = 1;
+	else
+	    ans = confirm(Strnew_charp(_("No SSL peer certificate: accept?")));
+	if (ans)
 	    amsg = Strnew_charp
-		("Accept SSL session without any peer certificate");
+		(_("Accept SSL session without any peer certificate"));
 	else {
 	    /* FIXME: gettextize? */
-	    char *e = "This SSL session was rejected "
+	    const char *e = "This SSL session was rejected "
 		"to prevent security violation: no peer certificate";
 	    disp_err_message(e, FALSE);
 	    free_ssl_ctx();
@@ -543,8 +497,7 @@ ssl_get_certificate(SSL * ssl, char *hostname)
 	if (amsg)
 	    disp_err_message(amsg->ptr, FALSE);
 	ssl_accept_this_site(hostname);
-	/* FIXME: gettextize? */
-	s = amsg ? amsg : Strnew_charp("valid certificate");
+	s = amsg ? amsg : Strnew_charp(_("valid certificate"));
 	return s;
     }
 #ifdef USE_SSL_VERIFY
@@ -559,21 +512,19 @@ ssl_get_certificate(SSL * ssl, char *hostname)
 	    const char *em = X509_verify_cert_error_string(verr);
 	    if (accept_this_site
 		&& strcasecmp(accept_this_site->ptr, hostname) == 0)
-		ans = "y";
+		ans = 1;
 	    else {
 		/* FIXME: gettextize? */
-		emsg = Sprintf("%s: accept? (y/n)", em);
-		ans = inputAnswer(emsg->ptr);
+		ans = confirm(Sprintf("%s: accept?", em));
 	    }
-	    if (ans && TOLOWER(*ans) == 'y') {
+	    if (ans) {
 		/* FIXME: gettextize? */
 		amsg = Sprintf("Accept unsecure SSL session: "
 			       "unverified: %s", em);
 	    }
 	    else {
-		/* FIXME: gettextize? */
 		char *e =
-		    Sprintf("This SSL session was rejected: %s", em)->ptr;
+		    Sprintf(_("This SSL session was rejected: %s"), em)->ptr;
 		disp_err_message(e, FALSE);
 		free_ssl_ctx();
 		return NULL;
@@ -585,22 +536,21 @@ ssl_get_certificate(SSL * ssl, char *hostname)
     if (emsg != NULL) {
 	if (accept_this_site
 	    && strcasecmp(accept_this_site->ptr, hostname) == 0)
-	    ans = "y";
+	    ans = 1;
 	else {
 	    Str ep = Strdup(emsg);
 	    if (ep->length > COLS - 16)
 		Strshrink(ep, ep->length - (COLS - 16));
-	    Strcat_charp(ep, ": accept? (y/n)");
-	    ans = inputAnswer(ep->ptr);
+	    Strcat_charp(ep, ": accept?");
+	    ans = confirm(ep);
 	}
-	if (ans && TOLOWER(*ans) == 'y') {
-	    /* FIXME: gettextize? */
-	    amsg = Strnew_charp("Accept unsecure SSL session:");
+	if (ans) {
+	    amsg = Strnew_charp(_("Accept unsecure SSL session:"));
 	    Strcat(amsg, emsg);
 	}
 	else {
 	    /* FIXME: gettextize? */
-	    char *e = "This SSL session was rejected "
+	    const char *e = "This SSL session was rejected "
 		"to prevent security violation";
 	    disp_err_message(e, FALSE);
 	    free_ssl_ctx();
@@ -610,8 +560,7 @@ ssl_get_certificate(SSL * ssl, char *hostname)
     if (amsg)
 	disp_err_message(amsg->ptr, FALSE);
     ssl_accept_this_site(hostname);
-    /* FIXME: gettextize? */
-    s = amsg ? amsg : Strnew_charp("valid certificate");
+    s = amsg ? amsg : Strnew_charp(_("valid certificate"));
     Strcat_charp(s, "\n");
     xn = X509_get_subject_name(x);
     if (X509_NAME_get_text_by_NID(xn, NID_commonName, buf, sizeof(buf)) == -1)
@@ -637,36 +586,32 @@ ssl_get_certificate(SSL * ssl, char *hostname)
 
 /* Raw level input stream functions */
 
-static void
+static int
 basic_close(int *handle)
 {
 #ifdef __MINGW32_VERSION
-    closesocket(*(int *)handle);
+    closesocket(*handle);
 #else
-    close(*(int *)handle);
+    close(*handle);
 #endif
+    free(handle);
+    return 0;
 }
 
 static int
-basic_read(int *handle, char *buf, int len)
+basic_read(int *handle, unsigned char *buf, int len)
 {
 #ifdef __MINGW32_VERSION
-    return recv(*(int *)handle, buf, len, 0);
+    return recv(*handle, buf, len, 0);
 #else
-    return read(*(int *)handle, buf, len);
+    return read(*handle, buf, len);
 #endif
 }
 
-static void
-file_close(struct file_handle *handle)
-{
-    handle->close(handle->f);
-}
-
 static int
-file_read(struct file_handle *handle, char *buf, int len)
+file_read(FILE *handle, char *buf, int len)
 {
-    return fread(buf, 1, len, handle->f);
+    return fread(buf, 1, len, handle);
 }
 
 static int
@@ -676,12 +621,14 @@ str_read(Str handle, char *buf, int len)
 }
 
 #ifdef USE_SSL
-static void
+static int
 ssl_close(struct ssl_handle *handle)
 {
     close(handle->sock);
     if (handle->ssl)
 	SSL_free(handle->ssl);
+    free(handle);
+    return 0;
 }
 
 static int
@@ -713,42 +660,65 @@ ssl_read(struct ssl_handle *handle, char *buf, int len)
 }
 #endif				/* USE_SSL */
 
-static void
+static int
 ens_close(struct ens_handle *handle)
 {
     ISclose(handle->is);
+    growbuf_clear(&handle->gb);
+    free(handle);
+    return 0;
 }
 
 static int
 ens_read(struct ens_handle *handle, char *buf, int len)
 {
-    if (handle->s == NULL || handle->pos == handle->s->length) {
+    if (handle->pos == handle->gb.length) {
 	char *p;
-	handle->s = StrmyISgets(handle->is);
-	if (handle->s->length == 0)
+	struct growbuf gbtmp;
+
+	ISgets_to_growbuf(handle->is, &handle->gb, TRUE);
+	if (handle->gb.length == 0)
 	    return 0;
-	cleanup_line(handle->s, PAGER_MODE);
 	if (handle->encoding == ENC_BASE64)
-	    Strchop(handle->s);
+	    memchop(handle->gb.ptr, &handle->gb.length);
 	else if (handle->encoding == ENC_UUENCODE) {
-	    if (!strncmp(handle->s->ptr, "begin", 5))
-		handle->s = StrmyISgets(handle->is);
-	    Strchop(handle->s);
+	    if (handle->gb.length >= 5 &&
+		!strncmp(handle->gb.ptr, "begin", 5))
+		ISgets_to_growbuf(handle->is, &handle->gb, TRUE);
+	    memchop(handle->gb.ptr, &handle->gb.length);
 	}
-	p = handle->s->ptr;
+	growbuf_init_without_GC(&gbtmp);
+	p = handle->gb.ptr;
 	if (handle->encoding == ENC_QUOTE)
-	    handle->s = decodeQP(&p);
+	    decodeQP_to_growbuf(&gbtmp, &p);
 	else if (handle->encoding == ENC_BASE64)
-	    handle->s = decodeB(&p);
+	    decodeB_to_growbuf(&gbtmp, &p);
 	else if (handle->encoding == ENC_UUENCODE)
-	    handle->s = decodeU(&p);
+	    decodeU_to_growbuf(&gbtmp, &p);
+	growbuf_clear(&handle->gb);
+	handle->gb = gbtmp;
 	handle->pos = 0;
     }
 
-    if (len > handle->s->length - handle->pos)
-	len = handle->s->length - handle->pos;
+    if (len > handle->gb.length - handle->pos)
+	len = handle->gb.length - handle->pos;
 
-    bcopy(&handle->s->ptr[handle->pos], buf, len);
+    memcpy(buf, &handle->gb.ptr[handle->pos], len);
     handle->pos += len;
     return len;
 }
+
+static void
+memchop(char *p, int *len)
+{
+    char *q;
+
+    for (q = p + *len; q > p; --q) {
+	if (q[-1] != '\n' && q[-1] != '\r')
+	    break;
+    }
+    if (q != p + *len)
+	*q = '\0';
+    *len = q - p;
+    return;
+}
