@@ -1,14 +1,12 @@
+/* vi: set sw=4 ts=8 ai sm noet : */
 
 #include <stdlib.h>
 #include <ctype.h>
-#include <gc.h>
-#define New_N(type,n) ((type*)GC_MALLOC((n)*sizeof(type)))
+#include "../alloc.h"
 
 #include "wc.h"
 
-#ifdef HAVE_LANGINFO_CODESET
 #include <langinfo.h>
-#endif
 
 wc_locale WcLocale = 0;
 
@@ -84,7 +82,7 @@ wc_guess_charset(char *charset, wc_ces orig)
 }
 
 wc_ces
-wc_guess_charset_short(char *charset, wc_ces orig)
+wc_guess_charset_short(const char *charset, wc_ces orig)
 {
     wc_ces guess;
 
@@ -106,17 +104,17 @@ wc_guess_locale_charset(char *locale, wc_ces orig)
 }
 
 wc_ces
-wc_charset_to_ces(char *charset)
+wc_charset_to_ces(const char *charset)
 {
-    char *p = charset;
+    const char *p = charset;
     char buf[16];
     int n;
 
-    if (tolower(*p) == 'x' && *(p+1) == '-')
+    if ((*p == 'x' || *p == 'X') && *(p+1) == '-')
 	p += 2;
     for (n = 0; *p && n < 15; p++) {
 	if ((unsigned char)*p > 0x20 && *p != '_' && *p != '-')
-	    buf[n++] = tolower(*p);
+	    buf[n++] = *p | 32; /* tolower(*p); */
     }
     buf[n] = 0;
     p = buf;
@@ -277,9 +275,9 @@ wc_charset_to_ces(char *charset)
 }
 
 wc_ces
-wc_charset_short_to_ces(char *charset)
+wc_charset_short_to_ces(const char *charset)
 {
-    char *p = charset;
+    const char *p = charset;
     char buf[16];
     wc_ces ces;
     int n;
@@ -290,7 +288,7 @@ wc_charset_short_to_ces(char *charset)
 
     for (n = 0; *p && n < 15; p++) {
 	if ((unsigned char)*p > 0x20 && *p != '_' && *p != '-')
-	    buf[n++] = tolower(*p);
+	    buf[n++] = *p | 32; /* tolower(*p); */
     }
     buf[n] = 0;
     p = buf;
@@ -373,16 +371,14 @@ wc_locale_to_ces(char *locale)
 
     if (*p == 'C' && *(p+1) == '\0')
 	return WC_CES_US_ASCII;
-#ifdef HAVE_LANGINFO_CODESET
     {
 	char *cs = nl_langinfo(CODESET);
 	if (cs && strcmp(cs, "US-ASCII"))
 	    return wc_charset_to_ces(cs);
     }
-#endif
     for (n = 0; *p && *p != '.' && n < 7; p++) {
 	if ((unsigned char)*p > 0x20)
-	    buf[n++] = tolower(*p);
+	    buf[n++] = *p | 32; /* tolower(*p); */
     }
     buf[n] = 0;
     if (*p == '.') {
@@ -411,8 +407,6 @@ wc_locale_to_ces(char *locale)
 	return wc_charset_to_ces(p);
     }
 
-    if (!strcmp(buf, "japanese"))
-	return WC_CES_SHIFT_JIS;
     if (!strcmp(buf, "zh_tw") ||
 	!strcmp(buf, "zh_hk"))
 	return WC_CES_BIG5;
@@ -469,7 +463,7 @@ wc_check_ces(wc_ces ces)
 static int
 wc_ces_list_cmp(const void *a, const void *b)
 {
-    return strcasecmp(((wc_ces_list *)a)->desc, ((wc_ces_list *)b)->desc);
+    return strcasecmp(((const wc_ces_list *)a)->desc, ((const wc_ces_list *)b)->desc);
 }
 
 static wc_ces_list *list = NULL;
