@@ -1,15 +1,18 @@
-/* $Id: menu.c,v 1.46 2007/05/23 12:34:20 inu Exp $ */
+/* vi: set sw=4 ts=8 ai sm noet : */
 /* 
  * w3m menu.c
  */
-#include <stdio.h>
 
+#include "buffer.h"
 #include "fm.h"
-#include "menu.h"
 #include "func.h"
+#include "funcname1.h"
+#include "menu.h"
 #include "myctype.h"
 #include "regex.h"
 
+#include <stdio.h>
+
 #ifdef USE_MOUSE
 #ifdef USE_GPM
 #include <gpm.h>
@@ -24,12 +27,13 @@ extern int sysm_process_mouse(int, int, int, int);
 #if defined(USE_GPM) || defined(USE_SYSMOUSE)
 #define X_MOUSE_SELECTED (char)0xff
 static int X_Mouse_Selection;
-extern int do_getch();
+extern int do_getch(void);
 #define getch()	do_getch()
 #endif				/* defined(USE_GPM) || defined(USE_SYSMOUSE) */
 #endif				/* USE_MOUSE */
 
 #ifdef USE_MENU
+static int menu_dispatch_with_mouse(int (*keymap[128]) (char c));
 
 static char **FRAME;
 static int FRAME_WIDTH;
@@ -57,6 +61,7 @@ static int mCancel(char c);
 static int mClose(char c);
 static int mSusp(char c);
 static int mMouse(char c);
+static int mSgrMouse(char c);
 static int mSrchF(char c);
 static int mSrchB(char c);
 static int mSrchN(char c);
@@ -108,7 +113,7 @@ static int (*MenuEscKeymap[128]) (char c) = {
     mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,
     mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,
     mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,
-    mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,
+    mNull,  mNull,  mNull,  mEsc,   mNull,  mNull,  mNull,  mNull,
 
     mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,
     mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,
@@ -116,15 +121,15 @@ static int (*MenuEscKeymap[128]) (char c) = {
     mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,
 
     mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,
-/*  O     */
+/*                                                          O     */
     mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mEscB,
     mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,
-/*  [                                     */
+/*                          [                                     */
     mNull,  mNull,  mNull,  mEscB,  mNull,  mNull,  mNull,  mNull,
 
     mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,
     mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,
-/*  v             */
+/*                                                  v             */
     mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mPrev,  mNull,
     mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,
 };
@@ -137,10 +142,11 @@ static int (*MenuEscBKeymap[128]) (char c) = {
     mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,
     mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,
     mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,
-    mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,
-/*  A       B       C       D       E                     */
+/*  8       9       :       ;       <       =       >       ?     */
+    mNull,  mNull,  mNull,  mNull,  mSgrMouse,mNull,mNull,  mNull,
+/*          A       B       C       D       E                     */
     mNull,  mUp,    mDown,  mOk,    mCancel,mClose, mNull, mNull,
-/*  L       M                     */
+/*                                  L       M                     */
     mNull,  mNull,  mNull,  mNull,  mClose, mMouse, mNull,  mNull,
     mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,
     mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,
@@ -240,6 +246,8 @@ static int SelectV = 0;
 static void initSelectMenu(void);
 static void smChBuf(void);
 static int smDelBuf(char c);
+static int smDnBuf(char c);
+static int smUpBuf(char c);
 
 /* --- SelectMenu (END) --- */
 
@@ -250,6 +258,8 @@ static int SelTabV = 0;
 static void initSelTabMenu(void);
 static void smChTab(void);
 static int smDelTab(char c);
+static int smDnTab(char c);
+static int smUpTab(char c);
 
 /* --- SelTabMenu (END) --- */
 
@@ -341,8 +351,8 @@ new_menu(Menu *menu, MenuItem *item)
     }
 }
 
-void
-geom_menu(Menu *menu, int x, int y, int mselect)
+static void
+geom_menu(Menu *menu, int mselect)
 {
     int win_x, win_y, win_w, win_h;
 
@@ -359,7 +369,6 @@ geom_menu(Menu *menu, int x, int y, int mselect)
 	if (win_w > COLS) {
 	    menu->width = COLS - 2 * FRAME_WIDTH;
 	    menu->width -= menu->width % FRAME_WIDTH;
-	    win_w = menu->width + 2 * FRAME_WIDTH;
 	}
     }
     menu->x = win_x + FRAME_WIDTH;
@@ -473,11 +482,11 @@ select_menu(Menu *menu, int mselect)
 	menu->select < menu->offset + menu->height)
 	draw_menu_item(menu, menu->select);
     menu->select = mselect;
-    standout();
-    draw_menu_item(menu, menu->select);
-    standend();
-    /* 
-     * move(menu->cursorY, menu->cursorX); */
+    if (menu->item[menu->select].type != MENU_NOP) {
+	standout();
+	draw_menu_item(menu, menu->select);
+	standend();
+    }
     move(menu->y + mselect - menu->offset, menu->x);
     toggle_stand();
     refresh();
@@ -543,7 +552,6 @@ down_menu(Menu *menu, int n)
 int
 action_menu(Menu *menu)
 {
-    char c;
     int mselect;
     MenuItem item;
 
@@ -556,27 +564,9 @@ action_menu(Menu *menu)
     select_menu(menu, menu->select);
 
     while (1) {
-#ifdef USE_MOUSE
-	if (use_mouse)
-	    mouse_active();
-#endif				/* USE_MOUSE */
-	c = getch();
-#ifdef USE_MOUSE
-	if (use_mouse)
-	    mouse_inactive();
-#if defined(USE_GPM) || defined(USE_SYSMOUSE)
-	if (c == X_MOUSE_SELECTED) {
-	    mselect = X_Mouse_Selection;
-	    if (mselect != MENU_NOTHING)
-		break;
-	}
-#endif				/* defined(USE_GPM) || defined(USE_SYSMOUSE) */
-#endif				/* USE_MOUSE */
-	if (IS_ASCII(c)) {	/* Ascii */
-	    mselect = (*menu->keymap[(int)c]) (c);
-	    if (mselect != MENU_NOTHING)
-		break;
-	}
+	mselect = menu_dispatch_with_mouse(menu->keymap);
+	if (mselect != MENU_NOTHING)
+	    break;
     }
     if (mselect >= 0 && mselect < menu->nitem) {
 	item = menu->item[mselect];
@@ -590,7 +580,6 @@ action_menu(Menu *menu)
 	    *item.variable = item.value;
 	if (item.type & MENU_FUNC) {
 	    CurrentKey = -1;
-	    CurrentKeyData = NULL;
 	    CurrentCmdData = item.data;
 	    (*item.func) ();
 	    CurrentCmdData = NULL;
@@ -630,7 +619,12 @@ popup_menu(Menu *parent, Menu *menu)
 	menu->cursorY = parent->cursorY;
 	guess_menu_xy(parent, menu->width, &menu->x, &menu->y);
     }
-    geom_menu(menu, menu->x, menu->y, menu->select);
+    while (menu->item[menu->select].type == MENU_NOP) {
+	if (menu->select >= menu->nitem - 1)
+	    break;
+	menu->select++;
+    }
+    geom_menu(menu, menu->select);
 
     CurrentMenu = menu;
     while (active) {
@@ -665,7 +659,7 @@ guess_menu_xy(Menu *parent, int width, int *x, int *y)
 }
 
 void
-new_option_menu(Menu *menu, char **label, int *variable, void (*func) ())
+new_option_menu(Menu *menu, char **label, int *variable, void (*func) (void))
 {
     int i, nitem;
     char **p;
@@ -696,6 +690,31 @@ new_option_menu(Menu *menu, char **label, int *variable, void (*func) ())
     new_menu(menu, item);
 }
 
+static int
+menu_dispatch_with_mouse(int (*keymap[128]) (char c))
+{
+    char c;
+
+#ifdef USE_MOUSE
+    if (use_mouse)
+	mouse_active();
+#endif			/* USE_MOUSE */
+    c = getch();
+#ifdef USE_MOUSE
+    if (use_mouse)
+	mouse_inactive();
+#if defined(USE_GPM) || defined(USE_SYSMOUSE)
+    if (c == X_MOUSE_SELECTED)
+	return X_Mouse_Selection;
+#endif			/* defined(USE_GPM) || defined(USE_SYSMOUSE) */
+#endif			/* USE_MOUSE */
+
+    if (IS_ASCII(c))	/* menu keymaps have 128 entries, so skip non-ascii */
+	return (*keymap[(int)c]) (c);
+
+    return (MENU_NOTHING);
+}
+
 static void
 set_menu_frame(void)
 {
@@ -725,15 +744,17 @@ static int
 mPc(char c)
 {
     c = getch();
-    return (MenuPcKeymap[(int)c] (c));
+    if (IS_ASCII(c))
+	return (MenuPcKeymap[(int)c] (c));
+    else
+	return (MENU_NOTHING);
 }
 #endif
 
 static int
 mEsc(char c)
 {
-    c = getch();
-    return (MenuEscKeymap[(int)c] (c));
+    return menu_dispatch_with_mouse(MenuEscKeymap);
 }
 
 static int
@@ -742,8 +763,10 @@ mEscB(char c)
     c = getch();
     if (IS_DIGIT(c))
 	return (mEscD(c));
-    else
+    else if (IS_ASCII(c))
 	return (MenuEscBKeymap[(int)c] (c));
+    else
+	return (MENU_NOTHING);
 }
 
 static int
@@ -921,15 +944,15 @@ mSusp(char c)
     return (MENU_NOTHING);
 }
 
-static char *SearchString = NULL;
+static const char *SearchString = NULL;
 
-int (*menuSearchRoutine) (Menu *, char *, int);
+int (*menuSearchRoutine) (Menu *, const char *, int);
 
 static int
-menuForwardSearch(Menu *menu, char *str, int from)
+menuForwardSearch(Menu *menu, const char *str, int from)
 {
     int i;
-    char *p;
+    const char *p;
     if ((p = regexCompile(str, IgnoreCase)) != NULL) {
 	message(p, 0, 0);
 	return -1;
@@ -946,7 +969,7 @@ menuForwardSearch(Menu *menu, char *str, int from)
 static int
 menu_search_forward(Menu *menu, int from)
 {
-    char *str;
+    const char *str;
     int found;
     str = inputStrHist("Forward: ", NULL, TextHist);
     if (str != NULL && *str == '\0')
@@ -976,10 +999,10 @@ mSrchF(char c)
 }
 
 static int
-menuBackwardSearch(Menu *menu, char *str, int from)
+menuBackwardSearch(Menu *menu, const char *str, int from)
 {
     int i;
-    char *p;
+    const char *p;
     if ((p = regexCompile(str, IgnoreCase)) != NULL) {
 	message(p, 0, 0);
 	return -1;
@@ -996,7 +1019,7 @@ menuBackwardSearch(Menu *menu, char *str, int from)
 static int
 menu_search_backward(Menu *menu, int from)
 {
-    char *str;
+    const char *str;
     int found;
     str = inputStrHist("Backward: ", NULL, TextHist);
     if (str != NULL && *str == '\0')
@@ -1029,9 +1052,9 @@ static int
 menu_search_next_previous(Menu *menu, int from, int reverse)
 {
     int found;
-    static int (*routine[2]) (Menu *, char *, int) = {
+    static int (*routine[2]) (Menu *, const char *, int) = {
     menuForwardSearch, menuBackwardSearch};
-    char *str;
+    const char *str;
 
     if (menuSearchRoutine == NULL) {
 	disp_message("No previous regular expression", TRUE);
@@ -1099,7 +1122,7 @@ process_mMouse(int btn, int x, int y)
 {
     Menu *menu;
     int mselect, i;
-    static int press_btn = MOUSE_BTN_RESET, press_x, press_y;
+    static int press_btn = MOUSE_BTN_RESET, press_y;
     char c = ' ';
 
     menu = CurrentMenu;
@@ -1168,7 +1191,6 @@ process_mMouse(int btn, int x, int y)
 
     if (btn != MOUSE_BTN4_DOWN_RXVT || press_btn == MOUSE_BTN_RESET) {
 	press_btn = btn;
-	press_x = x;
 	press_y = y;
     }
     else {
@@ -1183,14 +1205,6 @@ mMouse(char c)
     int btn, x, y;
 
     btn = (unsigned char)getch() - 32;
-#if defined(__CYGWIN__) && CYGWIN_VERSION_DLL_MAJOR < 1005
-    if (cygwin_mouse_btn_swapped) {
-	if (btn == MOUSE_BTN2_DOWN)
-	    btn = MOUSE_BTN3_DOWN;
-	else if (btn == MOUSE_BTN3_DOWN)
-	    btn = MOUSE_BTN2_DOWN;
-    }
-#endif
     x = (unsigned char)getch() - 33;
     if (x < 0)
 	x += 0x100;
@@ -1198,8 +1212,39 @@ mMouse(char c)
     if (y < 0)
 	y += 0x100;
 
-    /* 
-     * if (x < 0 || x >= COLS || y < 0 || y > LASTLINE) return; */
+    return process_mMouse(btn, x, y);
+}
+
+static int
+mSgrMouse(char c)
+{
+    int btn = 0, x = 0, y = 0;
+    unsigned char ch;
+
+    for (ch = getch(); IS_DIGIT(ch); ch = getch())
+	btn = btn * 10 + ch - '0';
+    if (ch != ';')
+	return MENU_NOTHING;
+
+    for (ch = getch(); IS_DIGIT(ch); ch = getch())
+	x = x * 10 + ch - '0';
+    if (ch != ';')
+	return MENU_NOTHING;
+    if (x > 0)
+	x--;
+
+    for (ch = getch(); IS_DIGIT(ch); ch = getch())
+	y = y * 10 + ch - '0';
+    if (ch == 'm')
+	btn |= 3;
+    else if (ch != 'M' && ch != ';')
+	return MENU_NOTHING;
+    if (y > 0)
+	y--;
+
+    if (x < 0 || x >= COLS || y < 0 || y > LASTLINE)
+	return MENU_NOTHING;
+
     return process_mMouse(btn, x, y);
 }
 
@@ -1261,6 +1306,12 @@ mMouse(char c)
 {
     return (MENU_NOTHING);
 }
+
+static int
+mSgrMouse(char c)
+{
+    return (MENU_NOTHING);
+}
 #endif				/* not USE_MOUSE */
 
 /* --- MenuFunctions (END) --- */
@@ -1272,9 +1323,6 @@ popupMenu(int x, int y, Menu *menu)
 {
     set_menu_frame();
 
-    initSelectMenu();
-    initSelTabMenu();
-
     menu->cursorX = Currentbuf->cursorX + Currentbuf->rootX;
     menu->cursorY = Currentbuf->cursorY + Currentbuf->rootY;
     menu->x = x + FRAME_WIDTH + 1;
@@ -1283,13 +1331,7 @@ popupMenu(int x, int y, Menu *menu)
     popup_menu(NULL, menu);
 }
 
-void
-mainMenu(int x, int y)
-{
-    popupMenu(x, y, &MainMenu);
-}
-
-DEFUN(mainMn, MAIN_MENU MENU, "Popup menu")
+DEFUN(mainMn, MAIN_MENU MENU, "Pop up menu")
 {
     Menu *menu = &MainMenu;
     char *data;
@@ -1317,7 +1359,7 @@ DEFUN(mainMn, MAIN_MENU MENU, "Popup menu")
 
 /* --- SelectMenu --- */
 
-DEFUN(selMn, SELECT_MENU, "Popup buffer selection menu")
+DEFUN(selMn, SELECT_MENU, "Pop up buffer-stack menu")
 {
     int x = Currentbuf->cursorX + Currentbuf->rootX,
 	y = Currentbuf->cursorY + Currentbuf->rootY;
@@ -1328,6 +1370,7 @@ DEFUN(selMn, SELECT_MENU, "Popup buffer selection menu")
 	y = mouse_action.cursorY;
     }
 #endif
+    initSelectMenu();
     popupMenu(x, y, &SelectMenu);
 }
 
@@ -1339,7 +1382,8 @@ initSelectMenu(void)
     Str str;
     char **label;
     char *p;
-    static char *comment = " SPC for select / D for delete buffer ";
+    static const char *comment =
+	    " SPC: select | D: delete | </>: move down/up ";
 
     SelectV = -1;
     for (i = 0, buf = Firstbuf; buf != NULL; i++, buf = buf->nextBuffer) {
@@ -1365,9 +1409,7 @@ initSelectMenu(void)
 		break;
 	    default:
 		Strcat_char(str, ' ');
-		p = parsedURL2Str(&buf->currentURL)->ptr;
-		if (DecodeURL)
-		    p = url_unquote_conv(p, 0);
+		p = url_decode2(parsedURL2Str(&buf->currentURL)->ptr, NULL);
 		Strcat_charp(str, p);
 		break;
 	    }
@@ -1396,6 +1438,8 @@ initSelectMenu(void)
     SelectMenu.cursorX = Currentbuf->cursorX + Currentbuf->rootX;
     SelectMenu.cursorY = Currentbuf->cursorY + Currentbuf->rootY;
     SelectMenu.keymap['D'] = smDelBuf;
+    SelectMenu.keymap['>'] = smUpBuf;
+    SelectMenu.keymap['<'] = smDnBuf;
     SelectMenu.item[nitem].type = MENU_NOP;
 }
 
@@ -1449,7 +1493,7 @@ smDelBuf(char c)
     CurrentMenu->x = x;
     CurrentMenu->y = y;
 
-    geom_menu(CurrentMenu, x, y, 0);
+    geom_menu(CurrentMenu, 0);
 
     CurrentMenu->select = (mselect <= CurrentMenu->nitem - 2) ? mselect
 	: (CurrentMenu->nitem - 2);
@@ -1460,11 +1504,82 @@ smDelBuf(char c)
     return (MENU_NOTHING);
 }
 
+static int
+smUpBuf(char c)
+{
+    Buffer *cur, *prev, *pprev;
+    int i, mselect, x, y;
+
+    if (!CurrentMenu->select) /* Top of menu */
+	return MENU_NOTHING;
+    for (i = 0, cur = Firstbuf; i < CurrentMenu->select;
+	 i++, cur = cur->nextBuffer) ;
+    prev = prevBuffer(Firstbuf, cur);
+    if ((pprev = prevBuffer(Firstbuf, prev)))
+	pprev->nextBuffer = cur;
+    prev->nextBuffer = cur->nextBuffer;
+    cur->nextBuffer = prev;
+    if (prev == Firstbuf)
+	Firstbuf = cur;
+
+    x = CurrentMenu->x;
+    y = CurrentMenu->y;
+    mselect = CurrentMenu->select;
+
+    initSelectMenu();
+    CurrentMenu->x = x;
+    CurrentMenu->y = y;
+    geom_menu(CurrentMenu, 0);
+    CurrentMenu->select = --mselect;
+    displayBuffer(Currentbuf, B_FORCE_REDRAW);
+    draw_all_menu(CurrentMenu);
+    select_menu(CurrentMenu, CurrentMenu->select);
+    return MENU_NOTHING;
+}
+
+static int
+smDnBuf(char c)
+{
+    Buffer *cur, *next, *prev;
+    int i, mselect, x, y;
+
+    /* Get selected buffer */
+    for (i = 0, cur = Firstbuf; i < CurrentMenu->select;
+	 i++, cur = cur->nextBuffer) ;
+
+    if (!cur->nextBuffer) /* Bottom of menu */
+	return MENU_NOTHING;
+
+    prev = prevBuffer(Firstbuf, cur);
+    next = cur->nextBuffer;
+
+    if (cur == Firstbuf)
+	Firstbuf = cur->nextBuffer;
+    else
+	prev->nextBuffer = next;
+    cur->nextBuffer = next->nextBuffer;
+    next->nextBuffer = cur;
+
+    x = CurrentMenu->x;
+    y = CurrentMenu->y;
+    mselect = CurrentMenu->select;
+
+    initSelectMenu();
+    CurrentMenu->x = x;
+    CurrentMenu->y = y;
+    geom_menu(CurrentMenu, 0);
+    CurrentMenu->select = ++mselect;
+    displayBuffer(Currentbuf, B_FORCE_REDRAW);
+    draw_all_menu(CurrentMenu);
+    select_menu(CurrentMenu, CurrentMenu->select);
+    return MENU_NOTHING;
+}
+
 /* --- SelectMenu (END) --- */
 
 /* --- SelTabMenu --- */
 
-DEFUN(tabMn, TAB_MENU, "Popup tab selection menu")
+DEFUN(tabMn, TAB_MENU, "Pop up tab selection menu")
 {
     int x = Currentbuf->cursorX + Currentbuf->rootX,
 	y = Currentbuf->cursorY + Currentbuf->rootY;
@@ -1475,6 +1590,7 @@ DEFUN(tabMn, TAB_MENU, "Popup tab selection menu")
 	y = mouse_action.cursorY;
     }
 #endif
+    initSelTabMenu();
     popupMenu(x, y, &SelTabMenu);
 }
 
@@ -1487,7 +1603,7 @@ initSelTabMenu(void)
     Str str;
     char **label;
     char *p;
-    static char *comment = " SPC for select / D for delete tab ";
+    static char *comment = "char  SPC for select / D for delete tab ";
 
     SelTabV = -1;
     for (i = 0, tab = LastTab; tab != NULL; i++, tab = tab->prevTab) {
@@ -1513,9 +1629,7 @@ initSelTabMenu(void)
 	    case SCM_MISSING:
 		break;
 	    default:
-		p = parsedURL2Str(&buf->currentURL)->ptr;
-		if (DecodeURL)
-		    p = url_unquote_conv(p, 0);
+		p = url_decode2(parsedURL2Str(&buf->currentURL)->ptr, NULL);
 		Strcat_charp(str, p);
 		break;
 	    }
@@ -1544,6 +1658,8 @@ initSelTabMenu(void)
     SelTabMenu.cursorX = Currentbuf->cursorX + Currentbuf->rootX;
     SelTabMenu.cursorY = Currentbuf->cursorY + Currentbuf->rootY;
     SelTabMenu.keymap['D'] = smDelTab;
+    SelTabMenu.keymap['>'] = smUpTab;
+    SelTabMenu.keymap['<'] = smDnTab;
     SelTabMenu.item[nitem].type = MENU_NOP;
 }
 
@@ -1592,7 +1708,7 @@ smDelTab(char c)
     CurrentMenu->x = x;
     CurrentMenu->y = y;
 
-    geom_menu(CurrentMenu, x, y, 0);
+    geom_menu(CurrentMenu, 0);
 
     CurrentMenu->select = (mselect <= CurrentMenu->nitem - 2) ? mselect
 	: (CurrentMenu->nitem - 2);
@@ -1603,13 +1719,108 @@ smDelTab(char c)
     return (MENU_NOTHING);
 }
 
+static int
+smUpTab(char c)
+{
+    int i, x, y, mselect;
+    TabBuffer *cur, *next, *nnext, *prev;
+
+    if (!CurrentMenu->select) /* Top of menu */
+	return MENU_NOTHING;
+    for (i = 0, cur = LastTab; i < CurrentMenu->select;
+	 i++, cur = cur->prevTab) ;
+
+    prev = cur->prevTab;
+    next = cur->nextTab;
+    nnext = next->nextTab;
+
+    if (nnext)
+	nnext->prevTab = cur;
+    next->prevTab = prev;
+    next->nextTab = cur;
+
+    if (prev)
+	prev->nextTab = next;
+    else
+	FirstTab = next;
+
+    cur->prevTab = next;
+    cur->nextTab = nnext;
+
+    if (next == LastTab)
+	LastTab = cur;
+
+    x = CurrentMenu->x;
+    y = CurrentMenu->y;
+    mselect = CurrentMenu->select;
+
+    initSelTabMenu();
+    CurrentMenu->x = x;
+    CurrentMenu->y = y;
+    geom_menu(CurrentMenu, 0);
+    CurrentMenu->select = --mselect;
+    displayBuffer(Currentbuf, B_FORCE_REDRAW);
+    draw_all_menu(CurrentMenu);
+    select_menu(CurrentMenu, CurrentMenu->select);
+    return (MENU_NOTHING);
+}
+
+static int
+smDnTab(char c)
+{
+    TabBuffer *cur, *next, *prev, *pprev;
+    int i, mselect, x, y;
+
+    /* Get selected tab */
+    for (i = 0, cur = LastTab; i < CurrentMenu->select;
+	 i++, cur = cur->prevTab) ;
+
+    if (!cur->prevTab) /* Bottom of menu */
+	return MENU_NOTHING;
+
+    prev = cur->prevTab;
+    pprev = prev->prevTab;
+    next = cur->nextTab;
+
+
+    if (pprev)
+	pprev->nextTab = cur;
+    prev->prevTab = cur;
+    prev->nextTab = next;
+    if (next)
+	next->prevTab = prev;
+
+    cur->prevTab = pprev;
+    cur->nextTab = prev;
+
+    if (prev == FirstTab)
+	FirstTab = cur;
+    if (cur == LastTab)
+	LastTab = prev;
+
+    x = CurrentMenu->x;
+    y = CurrentMenu->y;
+    mselect = CurrentMenu->select;
+
+    initSelTabMenu();
+    CurrentMenu->x = x;
+    CurrentMenu->y = y;
+    geom_menu(CurrentMenu, 0);
+    CurrentMenu->select = ++mselect;
+    displayBuffer(Currentbuf, B_FORCE_REDRAW);
+    draw_all_menu(CurrentMenu);
+    select_menu(CurrentMenu, CurrentMenu->select);
+    return MENU_NOTHING;
+}
+
+
 /* --- SelectMenu (END) --- */
 
 /* --- OptionMenu --- */
 
 void
 optionMenu(int x, int y, char **label, int *variable, int initial,
-	   void (*func) ())
+	   void (*func) (void))
 {
     Menu menu;
 
@@ -1696,7 +1907,7 @@ initMenu(void)
     FILE *mf;
     MenuList *list;
 
-    w3mMenuList = New_N(MenuList, 3);
+    w3mMenuList = New_N(MenuList, 4);
     w3mMenuList[0].id = "Main";
     w3mMenuList[0].menu = &MainMenu;
     w3mMenuList[0].item = MainMenuItem;
@@ -1845,10 +2056,8 @@ link_menu(Buffer *buf)
 	    Strcat_charp(str, " ");
 	if (!l->url)
 	    p = "";
-	else if (DecodeURL)
-	    p = url_unquote_conv(l->url, buf->document_charset);
 	else
-	    p = l->url;
+	    p = url_decode2(l->url, buf);
 	Strcat_charp(str, p);
 	label[i] = str->ptr;
 	if (len < str->length)
@@ -1913,6 +2122,7 @@ accesskey_menu(Buffer *buf)
     }
     label[nitem] = NULL;
 
+    set_menu_frame();
     new_option_menu(&menu, label, &key, NULL);
 
     menu.initial = 0;
@@ -2024,7 +2234,6 @@ list_menu(Buffer *buf)
     }
     label[nitem] = NULL;
 
-    set_menu_frame();
     set_menu_frame();
     new_option_menu(&menu, label, &key, NULL);
 
