@@ -1,3 +1,4 @@
+/* vi: set sw=4 ts=8 ai sm noet : */
 #ifndef PARSETAGX_H
 #define PARSETAGX_H
 
@@ -7,12 +8,11 @@
 /* Parsed Tag structure */
 
 struct parsed_tag {
-    unsigned char 	tagid;
-    char 		*tagname;
-    unsigned char 	*attrid;
-    char 		**value;
-    unsigned char 	*map;
-    char 		need_reconstruct;
+    unsigned char tagid;
+    unsigned char *attrid;
+    char **value;
+    unsigned char *map;
+    char need_reconstruct;
 };
 
 #define parsedtag_accepts(tag, id) ((tag)->map&&(tag)->map[id]!=MAX_TAGATTR)
