@@ -146,6 +146,7 @@ static int OptionEncode = FALSE;
 #define CMT_USE_MARK	N_("Enable mark operations")
 #endif
 #define CMT_EMACS_LIKE_LINEEDIT	N_("Enable Emacs-style line editing")
+#define CMT_READLINE_PASTE	N_("Use ^Y to paste deleted text")
 #define CMT_SPACE_AUTOCOMPLETE  N_("Space key triggers file completion while editing URLs")
 #define CMT_VI_PREC_NUM	 N_("Enable vi-like numeric prefix")
 #define CMT_LABEL_TOPLINE N_("Move cursor to top line when going to label")
@@ -215,6 +216,7 @@ static int OptionEncode = FALSE;
 #define CMT_NOSENDREFERER N_("Suppress `Referer:' header")
 #define CMT_CROSSORIGINREFERER N_("Exclude pathname and query string from `Referer:' header when cross domain communication")
 #define CMT_IGNORE_CASE N_("Search case-insensitively")
+#define CMT_SMART_CASE N_("Use vim-like smartcase for searching")
 #define CMT_USE_LESSOPEN N_("Use LESSOPEN")
 #ifdef USE_SSL
 #ifdef USE_SSL_VERIFY
@@ -527,6 +529,8 @@ struct param_ptr params3[] = {
 #endif
     {"emacs_like_lineedit", P_INT, PI_ONOFF, (void *)&emacs_like_lineedit,
      CMT_EMACS_LIKE_LINEEDIT, NULL},
+    {"rl_paste", P_INT, PI_ONOFF, (void *)&rl_paste,
+     CMT_READLINE_PASTE, NULL},
     {"space_autocomplete", P_INT, PI_ONOFF, (void *)&space_autocomplete,
      CMT_SPACE_AUTOCOMPLETE, NULL},
     {"vi_prec_num", P_INT, PI_ONOFF, (void *)&vi_prec_num, CMT_VI_PREC_NUM,
@@ -536,6 +540,8 @@ struct param_ptr params3[] = {
     {"wrap_search", P_INT, PI_ONOFF, (void *)&WrapDefault, CMT_WRAP, NULL},
     {"ignorecase_search", P_INT, PI_ONOFF, (void *)&IgnoreCase,
      CMT_IGNORE_CASE, NULL},
+    {"smartcase_search", P_INT, PI_ONOFF, (void *)&SmartCase, CMT_SMART_CASE,
+     NULL},
 #ifdef USE_MIGEMO
     {"use_migemo", P_INT, PI_ONOFF, (void *)&use_migemo, CMT_USE_MIGEMO,
      NULL},
