@@ -1,4 +1,4 @@
-/* $Id: table.c,v 1.58 2010/08/09 11:59:19 htrb Exp $ */
+/* vi: set sw=4 ts=8 ai sm noet : */
 /* 
  * HTML table
  */
@@ -12,6 +12,8 @@
 #include "Str.h"
 #include "myctype.h"
 
+#include <strings.h>
+
 int symbol_width = 0;
 int symbol_width0 = 0;
 
@@ -101,8 +103,8 @@ weight3(int x)
 #endif				/* not MATRIX */
 
 static int
-bsearch_2short(short e1, short *ent1, short e2, short *ent2, int base,
-	       short *indexarray, int nent)
+bsearch_2short(short e1, const short *ent1, short e2, const short *ent2,
+	       int base, const short *indexarray, int nent)
 {
     int n = nent;
     int k = 0;
@@ -128,7 +130,7 @@ bsearch_2short(short e1, short *ent1, short e2, short *ent2, int base,
 }
 
 static int
-bsearch_double(double e, double *ent, short *indexarray, int nent)
+bsearch_double(double e, const double *ent, const short *indexarray, int nent)
 {
     int n = nent;
     int k = 0;
@@ -188,7 +190,7 @@ dv2sv(double *dv, short *iv, int size)
     indexarray = NewAtom_N(short, size);
     edv = NewAtom_N(double, size);
     for (i = 0; i < size; i++) {
-	iv[i] = ceil(dv[i]);
+	iv[i] = (short) ceil(dv[i]);
 	edv[i] = (double)iv[i] - dv[i];
     }
 
@@ -205,7 +207,7 @@ dv2sv(double *dv, short *iv, int size)
 	indexarray[i] = k;
     }
     iw = min((int)(w + 0.5), size);
-    if (iw == 0)
+    if (iw <= 1)
 	return;
     x = edv[(int)indexarray[iw - 1]];
     for (i = 0; i < size; i++) {
@@ -269,7 +271,7 @@ table_border_width(struct table *t)
 }
 
 struct table *
-newTable()
+newTable(void)
 {
     struct table *t;
     int i, j;
@@ -278,7 +280,7 @@ newTable()
     t->max_rowsize = MAXROW;
     t->tabdata = New_N(GeneralList **, MAXROW);
     t->tabattr = New_N(table_attr *, MAXROW);
-    t->tabheight = NewAtom_N(short, MAXROW);
+    t->tabheight = NewAtom_N(int, MAXROW);
 #ifdef ID_EXT
     t->tabidvalue = New_N(Str *, MAXROW);
     t->tridvalue = New_N(Str, MAXROW);
@@ -307,12 +309,6 @@ newTable()
     t->matrix = NULL;
     t->vector = NULL;
 #endif				/* MATRIX */
-#if 0
-    t->tabcontentssize = 0;
-    t->indent = 0;
-    t->linfo.prev_ctype = PC_ASCII;
-    t->linfo.prev_spaces = -1;
-#endif
     t->linfo.prevchar = Strnew_size(8);
     set_prevchar(t->linfo.prevchar, "", 0);
     t->trattr = 0;
@@ -331,17 +327,21 @@ check_row(struct table *t, int row)
     int i, r;
     GeneralList ***tabdata;
     table_attr **tabattr;
-    short *tabheight;
+    int *tabheight;
 #ifdef ID_EXT
     Str **tabidvalue;
     Str *tridvalue;
 #endif				/* ID_EXT */
 
+    if (row < 0 || row >= MAXROW_LIMIT)
+	return;
     if (row >= t->max_rowsize) {
 	r = max(t->max_rowsize * 2, row + 1);
+	if (r <= 0 || r > MAXROW_LIMIT)
+	    r = MAXROW_LIMIT;
 	tabdata = New_N(GeneralList **, r);
 	tabattr = New_N(table_attr *, r);
-	tabheight = NewAtom_N(short, r);
+	tabheight = NewAtom_N(int, r);
 #ifdef ID_EXT
 	tabidvalue = New_N(Str *, r);
 	tridvalue = New_N(Str, r);
@@ -400,7 +400,7 @@ pushdata(struct table *t, int row, int col, char *data)
     pushText(t->tabdata[row][col], data ? data : "");
 }
 
-void
+static void
 suspend_or_pushdata(struct table *tbl, char *line)
 {
     if (tbl->flag & TBL_IN_COL)
@@ -415,21 +415,19 @@ suspend_or_pushdata(struct table *tbl, char *line)
 #ifdef USE_M17N
 #define PUSH_TAG(str,n) Strcat_charp_n(tagbuf, str, n)
 #else
-#define PUSH_TAG(str,n) Strcat_char(tagbuf, *str)
+#define PUSH_TAG(str,n) Strcat_char(tagbuf, *str), (void)n
 #endif
 
 int visible_length_offset = 0;
 int
-visible_length(char *str)
+visible_length(const char *str)
 {
     int len = 0, n, max_len = 0;
     int status = R_ST_NORMAL;
-    int prev_status = status;
+    int prev_status;
     Str tagbuf = Strnew();
     char *t, *r2;
-    int amp_len = 0;
 
-    t = str;
     while (*str) {
 	prev_status = status;
 	if (next_status(*str, &status)) {
@@ -457,11 +455,9 @@ visible_length(char *str)
 	    if (prev_status == R_ST_NORMAL) {
 		Strclear(tagbuf);
 		len--;
-		amp_len = 0;
 	    }
 	    else {
 		PUSH_TAG(str, n);
-		amp_len++;
 	    }
 	}
 	else if (status == R_ST_NORMAL && prev_status == R_ST_AMP) {
@@ -506,7 +502,7 @@ visible_length(char *str)
     return len > max_len ? len : max_len;
 }
 
-int
+static int
 visible_length_plain(char *str)
 {
     int len = 0, max_len = 0;
@@ -613,12 +609,17 @@ print_item(struct table *t, int row, int col, int width, Str buf)
 	    alignment = ALIGN_RIGHT;
 	else if ((t->tabattr[row][col] & HTT_ALIGN) == HTT_CENTER)
 	    alignment = ALIGN_CENTER;
+	if (DisableCenter && alignment == ALIGN_CENTER)
+	    alignment = ALIGN_LEFT;
 	align(lbuf, width, alignment);
 	Strcat(buf, lbuf->line);
     }
     else {
 	lbuf = newTextLine(NULL, 0);
-	align(lbuf, width, ALIGN_CENTER);
+	if (DisableCenter)
+	    align(lbuf, width, ALIGN_LEFT);
+	else
+	    align(lbuf, width, ALIGN_CENTER);
 	Strcat(buf, lbuf->line);
     }
 }
@@ -635,6 +636,8 @@ print_sep(struct table *t, int row, int type, int maxcol, Str buf)
     int rule_mode;
     int i, k, l, m;
 
+    if (row + 1 < 0 || row + 1 >= MAXROW_LIMIT)
+	return;
     if (row >= 0)
 	check_row(t, row);
     check_row(t, row + 1);
@@ -753,7 +756,7 @@ do_refill(struct table *tbl, int row, int col, int maxlimit)
     if (h_env.limit > maxlimit)
 	h_env.limit = maxlimit;
     if (tbl->border_mode != BORDER_NONE && tbl->vcellpadding > 0)
-	do_blankline(&h_env, &obuf, 0, 0, h_env.limit);
+	do_blankline(&h_env, &obuf, 0, h_env.limit);
     for (l = orgdata->first; l != NULL; l = l->next) {
 	if (TAG_IS(l->ptr, "<table_alt", 10)) {
 	    int id = -1;
@@ -761,7 +764,7 @@ do_refill(struct table *tbl, int row, int col, int maxlimit)
 	    struct parsed_tag *tag;
 	    if ((tag = parse_tag(&p, TRUE)) != NULL)
 		parsedtag_get_value(tag, ATTR_TID, &id);
-	    if (id >= 0 && id < tbl->ntable) {
+	    if (id >= 0 && id < tbl->ntable && tbl->tables[id].ptr) {
 		int alignment;
 		TextLineListItem *ti;
 		struct table *t = tbl->tables[id].ptr;
@@ -770,7 +773,7 @@ do_refill(struct table *tbl, int row, int col, int maxlimit)
 		save_fonteffect(&h_env, h_env.obuf);
 		flushline(&h_env, &obuf, 0, 2, h_env.limit);
 		if (t->vspace > 0 && !(obuf.flag & RB_IGNORE_P))
-		    do_blankline(&h_env, &obuf, 0, 0, h_env.limit);
+		    do_blankline(&h_env, &obuf, 0, h_env.limit);
 		if (RB_GET_ALIGN(h_env.obuf) == RB_CENTER)
 		    alignment = ALIGN_CENTER;
 		else if (RB_GET_ALIGN(h_env.obuf) == RB_RIGHT)
@@ -790,7 +793,7 @@ do_refill(struct table *tbl, int row, int col, int maxlimit)
 		obuf.flag &= ~RB_IGNORE_P;
 		h_env.blank_lines = 0;
 		if (t->vspace > 0) {
-		    do_blankline(&h_env, &obuf, 0, 0, h_env.limit);
+		    do_blankline(&h_env, &obuf, 0, h_env.limit);
 		    obuf.flag |= RB_IGNORE_P;
 		}
 	    }
@@ -808,7 +811,7 @@ do_refill(struct table *tbl, int row, int col, int maxlimit)
 	int rowspan = table_rowspan(tbl, row, col);
 	if (row + rowspan <= tbl->maxrow) {
 	    if (tbl->vcellpadding > 0 && !(obuf.flag & RB_IGNORE_P))
-		do_blankline(&h_env, &obuf, 0, 0, h_env.limit);
+		do_blankline(&h_env, &obuf, 0, h_env.limit);
 	}
 	else {
 	    if (tbl->vspace > 0)
@@ -818,7 +821,7 @@ do_refill(struct table *tbl, int row, int col, int maxlimit)
     else {
 	if (tbl->vcellpadding > 0) {
 	    if (!(obuf.flag & RB_IGNORE_P))
-		do_blankline(&h_env, &obuf, 0, 0, h_env.limit);
+		do_blankline(&h_env, &obuf, 0, h_env.limit);
 	}
 	else
 	    purgeline(&h_env);
@@ -847,9 +850,42 @@ table_rule_width(struct table *t)
 }
 
 static void
-check_cell_width(short *tabwidth, short *cellwidth,
-		 short *col, short *colspan, short maxcell,
-		 short *indexarray, int space, int dir)
+check_cell_height(int *tabheight, const int *cellheight,
+		 const short *row, const short *rowspan, short maxcell,
+		 const short *indexarray, int space, int dir)
+{
+    int i, j, k, brow, erow;
+    int sheight, height;
+
+    for (k = 0; k <= maxcell; k++) {
+	j = indexarray[k];
+	if (cellheight[j] <= 0)
+	    continue;
+	brow = row[j];
+	erow = brow + rowspan[j];
+	sheight = 0;
+	for (i = brow; i < erow; i++)
+	    sheight += tabheight[i];
+
+	height = cellheight[j] - (rowspan[j] - 1) * space;
+	if (height > sheight) {
+	    int w = (height - sheight) / rowspan[j];
+	    int r = (height - sheight) % rowspan[j];
+	    for (i = brow; i < erow; i++)
+		tabheight[i] += w;
+	    /* dir {0: horizontal, 1: vertical} */
+	    if (dir == 1 && r > 0)
+		r = rowspan[j];
+	    for (i = 1; i <= r; i++)
+		tabheight[erow - i]++;
+	}
+    }
+}
+
+static void
+check_cell_width(short *tabwidth, const short *cellwidth,
+		 const short *col, const short *colspan, short maxcell,
+		 const short *indexarray, int space, int dir)
 {
     int i, j, k, bcol, ecol;
     int swidth, width;
@@ -879,7 +915,7 @@ check_cell_width(short *tabwidth, short *cellwidth,
     }
 }
 
-void
+static void
 check_minimum_width(struct table *t, short *tabwidth)
 {
     int i;
@@ -894,7 +930,7 @@ check_minimum_width(struct table *t, short *tabwidth)
 		     cell->maxcell, cell->index, t->cellspacing, 0);
 }
 
-void
+static void
 check_maximum_width(struct table *t)
 {
     struct table_cell *cell = &t->cell;
@@ -961,7 +997,7 @@ set_integered_width(struct table *t, double *dwidth, short *iwidth)
     for (step = 0; step < 2; step++) {
 	for (i = 0; i <= t->maxcol; i += n) {
 	    int nn;
-	    char *idx;
+	    short *idx;
 	    double nsum;
 	    if (sum < 0.5)
 		return;
@@ -982,7 +1018,7 @@ set_integered_width(struct table *t, double *dwidth, short *iwidth)
 		    (double)rulewidth - mod[ii] > 0.5)
 		    fixed[ii] = 1;
 	    }
-	    idx = NewAtom_N(char, n);
+	    idx = NewAtom_N(short, n);
 	    for (k = 0; k < cell->maxcell; k++) {
 		int kk, w, width, m;
 		j = cell->index[k];
@@ -1212,21 +1248,18 @@ check_compressible_cell(struct table *t, MAT * minv,
 }
 
 #define MAX_ITERATION 10
-int
+static int
 check_table_width(struct table *t, double *newwidth, MAT * minv, int itr)
 {
     int i, j, k, m, bcol, ecol;
     int corr = 0;
     struct table_cell *cell = &t->cell;
-#ifdef __GNUC__
-    short orgwidth[t->maxcol + 1], corwidth[t->maxcol + 1];
-    short cwidth[cell->maxcell + 1];
-    double swidth[cell->maxcell + 1];
-#else				/* __GNUC__ */
-    short orgwidth[MAXCOL], corwidth[MAXCOL];
-    short cwidth[MAXCELL];
-    double swidth[MAXCELL];
-#endif				/* __GNUC__ */
+    int maxcol = t->maxcol >= 0 ? t->maxcol + 1 : 1;
+    int maxcell = cell->maxcell >= 0 ? cell->maxcell + 1 : 1;
+    short *orgwidth = GC_malloc(maxcol * sizeof(*orgwidth));
+    short *corwidth = GC_malloc(maxcol * sizeof(*corwidth));
+    short *cwidth = GC_malloc(maxcell * sizeof(*cwidth));
+    double *swidth = GC_malloc(maxcell * sizeof(*swidth));
     double twidth, sxy, *Sxx, stotal;
 
     twidth = 0.;
@@ -1469,7 +1502,7 @@ set_table_width(struct table *t, short *newwidth, int maxwidth)
 }
 #endif				/* not MATRIX */
 
-void
+static void
 check_table_height(struct table *t)
 {
     int i, j, k;
@@ -1479,8 +1512,8 @@ check_table_height(struct table *t)
 	short *indexarray;
 	short maxcell;
 	short size;
-	short *height;
-    } cell;
+	int *height;
+    } cell = {0};
     int space = 0;
 
     cell.size = 0;
@@ -1518,7 +1551,7 @@ check_table_height(struct table *t)
 			cell.row = NewAtom_N(short, cell.size);
 			cell.rowspan = NewAtom_N(short, cell.size);
 			cell.indexarray = NewAtom_N(short, cell.size);
-			cell.height = NewAtom_N(short, cell.size);
+			cell.height = NewAtom_N(int, cell.size);
 		    }
 		    else {
 			cell.size = max(cell.size + MAXCELL, c + 1);
@@ -1527,7 +1560,7 @@ check_table_height(struct table *t)
 						 cell.size);
 			cell.indexarray = New_Reuse(short, cell.indexarray,
 						    cell.size);
-			cell.height = New_Reuse(short, cell.height, cell.size);
+			cell.height = New_Reuse(int, cell.height, cell.size);
 		    }
 		}
 		if (c > cell.maxcell) {
@@ -1561,21 +1594,18 @@ check_table_height(struct table *t)
     case BORDER_NONE:
 	space = 0;
     }
-    check_cell_width(t->tabheight, cell.height, cell.row, cell.rowspan,
+    check_cell_height(t->tabheight, cell.height, cell.row, cell.rowspan,
 		     cell.maxcell, cell.indexarray, space, 1);
 }
 
 #define CHECK_MINIMUM	1
 #define CHECK_FIXED	2
 
-int
+static int
 get_table_width(struct table *t, short *orgwidth, short *cellwidth, int flag)
 {
-#ifdef __GNUC__
-    short newwidth[t->maxcol + 1];
-#else				/* not __GNUC__ */
-    short newwidth[MAXCOL];
-#endif				/* not __GNUC__ */
+    int maxcol = t->maxcol >= 0 ? t->maxcol + 1 : 1;
+    short *newwidth = GC_malloc(maxcol * sizeof(*newwidth));
     int i;
     int swidth;
     struct table_cell *cell = &t->cell;
@@ -1585,11 +1615,8 @@ get_table_width(struct table *t, short *orgwidth, short *cellwidth, int flag)
 	newwidth[i] = max(orgwidth[i], 0);
 
     if (flag & CHECK_FIXED) {
-#ifdef __GNUC__
-	short ccellwidth[cell->maxcell + 1];
-#else				/* not __GNUC__ */
-	short ccellwidth[MAXCELL];
-#endif				/* not __GNUC__ */
+	int maxcell = cell->maxcell >= 0 ? cell->maxcell + 1 : 1;
+	short *ccellwidth = GC_malloc(maxcell * sizeof(*ccellwidth));
 	for (i = 0; i <= t->maxcol; i++) {
 	    if (newwidth[i] < t->fixed_width[i])
 		newwidth[i] = t->fixed_width[i];
@@ -1624,7 +1651,16 @@ get_table_width(struct table *t, short *orgwidth, short *cellwidth, int flag)
 #define fixed_table_width(t)\
   (get_table_width(t,t->fixed_width,t->cell.fixed_width,CHECK_MINIMUM))
 
+#define MAX_COTABLE_LEVEL 100
+static int cotable_level;
+
 void
+initRenderTable(void)
+{
+    cotable_level = 0;
+}
+
+static void
 renderCoTable(struct table *tbl, int maxlimit)
 {
     struct readbuffer obuf;
@@ -1634,8 +1670,14 @@ renderCoTable(struct table *tbl, int maxlimit)
     int i, col, row;
     int indent, maxwidth;
 
+    if (cotable_level >= MAX_COTABLE_LEVEL)
+	return;	/* workaround to prevent infinite recursion */
+    cotable_level++;
+
     for (i = 0; i < tbl->ntable; i++) {
 	t = tbl->tables[i].ptr;
+	if (t == NULL)
+	    continue;
 	col = tbl->tables[i].col;
 	row = tbl->tables[i].row;
 	indent = tbl->tables[i].indent;
@@ -1691,7 +1733,7 @@ renderTable(struct table *t, int max_width, struct html_feed_environ *h_env)
 {
     int i, j, w, r, h;
     Str renderbuf;
-    short new_tabwidth[MAXCOL];
+    short new_tabwidth[MAXCOL] = { 0 };
 #ifdef MATRIX
     int itr;
     VEC *newwidth;
@@ -1724,6 +1766,10 @@ renderTable(struct table *t, int max_width, struct html_feed_environ *h_env)
     if (max_width < rulewidth)
 	max_width = rulewidth;
 
+#define MAX_TABWIDTH 10000
+    if (max_width > MAX_TABWIDTH)
+	max_width = MAX_TABWIDTH;
+
     check_maximum_width(t);
 
 #ifdef MATRIX
@@ -2102,7 +2148,7 @@ check_minimum0(struct table *t, int min)
 }
 
 static int
-setwidth0(struct table *t, struct table_mode *mode)
+setwidth0(struct table *t)
 {
     int w;
     int width = t->tabcontentssize;
@@ -2130,7 +2176,7 @@ setwidth0(struct table *t, struct table_mode *mode)
 static void
 setwidth(struct table *t, struct table_mode *mode)
 {
-    int width = setwidth0(t, mode);
+    int width = setwidth0(t);
     if (width < 0)
 	return;
 #ifdef NOWRAP
@@ -2195,7 +2241,8 @@ check_rowcol(struct table *tbl, struct table_mode *mode)
 
     if (!(tbl->flag & TBL_IN_ROW)) {
 	tbl->flag |= TBL_IN_ROW;
-	tbl->row++;
+	if (tbl->row + 1 < MAXROW_LIMIT)
+	    tbl->row++;
 	if (tbl->row > tbl->maxrow)
 	    tbl->maxrow = tbl->row;
 	tbl->col = -1;
@@ -2212,6 +2259,8 @@ check_rowcol(struct table *tbl, struct table_mode *mode)
 	if (tbl->col < MAXCOL)
 	    break;
 	tbl->col = 0;
+	if (tbl->row + 1 >= MAXROW_LIMIT)
+	    break;
     }
     if (tbl->row > tbl->maxrow)
 	tbl->maxrow = tbl->row;
@@ -2223,7 +2272,7 @@ check_rowcol(struct table *tbl, struct table_mode *mode)
     tbl->flag |= TBL_IN_COL;
 }
 
-int
+static int
 skip_space(struct table *t, char *line, struct table_linfo *linfo,
 	   int checkminimum)
 {
@@ -2272,7 +2321,8 @@ skip_space(struct table *t, char *line, struct table_linfo *linfo,
 	    }
 	    if (s > 0) {
 #ifdef USE_M17N
-		if (ctype == PC_KANJI1 && prev_ctype == PC_KANJI1)
+		if (!SimplePreserveSpace &&
+		    ctype == PC_KANJI1 && prev_ctype == PC_KANJI1)
 		    skip += s;
 		else
 #endif
@@ -2323,6 +2373,8 @@ feed_table_block_tag(struct table *tbl,
     int offset;
     if (mode->indent_level <= 0 && indent == -1)
 	return;
+    if (mode->indent_level >= CHAR_MAX && indent == 1)
+	return;
     setwidth(tbl, mode);
     feed_table_inline_tag(tbl, line, mode, -1);
     clearcontentssize(tbl, mode);
@@ -2336,10 +2388,14 @@ feed_table_block_tag(struct table *tbl,
 	if (mode->indent_level < MAX_INDENT_LEVEL)
 	    tbl->indent -= INDENT_INCR;
     }
+    if (tbl->indent < 0)
+	tbl->indent = 0;
     offset = tbl->indent;
     if (cmd == HTML_DT) {
 	if (mode->indent_level > 0 && mode->indent_level <= MAX_INDENT_LEVEL)
 	    offset -= INDENT_INCR;
+	if (offset < 0)
+	    offset = 0;
     }
     if (tbl->indent > 0) {
 	check_minimum0(tbl, 0);
@@ -2511,6 +2567,8 @@ feed_table_tag(struct table *tbl, char *line, struct table_mode *mode,
 	    check_rowcol(tbl, mode);
 	return TAG_ACTION_N_TABLE;
     case HTML_TR:
+	if (tbl->row + 1 >= MAXROW_LIMIT)
+	    return TAG_ACTION_NONE;
 	if (tbl->col >= 0 && tbl->tabcontentssize > 0)
 	    setwidth(tbl, mode);
 	tbl->col = -1;
@@ -2546,8 +2604,10 @@ feed_table_tag(struct table *tbl, char *line, struct table_mode *mode,
 	    }
 	}
 #ifdef ID_EXT
-	if (parsedtag_get_value(tag, ATTR_ID, &p))
+	if (parsedtag_get_value(tag, ATTR_ID, &p)) {
+	    check_row(tbl, tbl->row);
 	    tbl->tridvalue[tbl->row] = Strnew_charp(p);
+	}
 #endif				/* ID_EXT */
 	tbl->trattr = align | valign;
 	break;
@@ -2564,6 +2624,8 @@ feed_table_tag(struct table *tbl, char *line, struct table_mode *mode,
 	}
 	if (tbl->col == -1) {
 	    if (!(tbl->flag & TBL_IN_ROW)) {
+		if (tbl->row + 1 >= MAXROW_LIMIT)
+		    return TAG_ACTION_NONE;
 		tbl->row++;
 		tbl->flag |= TBL_IN_ROW;
 	    }
@@ -2572,7 +2634,7 @@ feed_table_tag(struct table *tbl, char *line, struct table_mode *mode,
 	}
 	tbl->col++;
 	check_row(tbl, tbl->row);
-	while (tbl->tabattr[tbl->row][tbl->col]) {
+	while (tbl->col < MAXCOL && tbl->tabattr[tbl->row][tbl->col]) {
 	    tbl->col++;
 	}
 	if (tbl->col > MAXCOL - 1) {
@@ -2594,18 +2656,23 @@ feed_table_tag(struct table *tbl, char *line, struct table_mode *mode,
 	else
 	    valign = HTT_MIDDLE;
 	if (parsedtag_get_value(tag, ATTR_ROWSPAN, &rowspan)) {
-	    if(rowspan > ATTR_ROWSPAN_MAX) {
+	    if (rowspan < 0 || rowspan > ATTR_ROWSPAN_MAX)
 		rowspan = ATTR_ROWSPAN_MAX;
-	    }
-	    if ((tbl->row + rowspan) >= tbl->max_rowsize)
-		check_row(tbl, tbl->row + rowspan);
+	    if ((tbl->row + rowspan - 1) >= MAXROW_LIMIT)
+		rowspan = MAXROW_LIMIT - tbl->row;
+	    if ((tbl->row + rowspan) > tbl->max_rowsize)
+		check_row(tbl, tbl->row + rowspan - 1);
 	}
+	if (rowspan < 1)
+	    rowspan = 1;
 	if (parsedtag_get_value(tag, ATTR_COLSPAN, &colspan)) {
 	    if ((tbl->col + colspan) >= MAXCOL) {
 		/* Can't expand column */
 		colspan = MAXCOL - tbl->col;
 	    }
 	}
+	if (colspan < 1)
+	    colspan = 1;
 	if (parsedtag_get_value(tag, ATTR_ALIGN, &i)) {
 	    switch (i) {
 	    case ALIGN_LEFT:
@@ -2708,9 +2775,6 @@ feed_table_tag(struct table *tbl, char *line, struct table_mode *mode,
 	for (i = 0; i < rowspan; i++) {
 	    check_row(tbl, tbl->row + i);
 	    for (j = 0; j < colspan; j++) {
-#if 0
-		tbl->tabattr[tbl->row + i][tbl->col + j] &= ~(HTT_X | HTT_Y);
-#endif
 		if (!(tbl->tabattr[tbl->row + i][tbl->col + j] &
 		      (HTT_X | HTT_Y))) {
 		    tbl->tabattr[tbl->row + i][tbl->col + j] |=
@@ -2878,6 +2942,14 @@ feed_table_tag(struct table *tbl, char *line, struct table_mode *mode,
 	tmp = process_input(tag);
 	feed_table1(tbl, tmp, mode, width);
 	break;
+    case HTML_BUTTON:
+       tmp = process_button(tag);
+       feed_table1(tbl, tmp, mode, width);
+       break;
+    case HTML_N_BUTTON:
+       tmp = process_n_button();
+       feed_table1(tbl, tmp, mode, width);
+       break;
     case HTML_SELECT:
 	tmp = process_select(tag);
 	if (tmp)
@@ -3010,7 +3082,6 @@ feed_table_tag(struct table *tbl, char *line, struct table_mode *mode,
 	break;
     case HTML_TABLE_ALT:
 	id = -1;
-	w = 0;
 	parsedtag_get_value(tag, ATTR_TID, &id);
 	if (id >= 0 && id < tbl->ntable) {
 	    struct table *tbl1 = tbl->tables[id].ptr;
@@ -3042,7 +3113,7 @@ feed_table_tag(struct table *tbl, char *line, struct table_mode *mode,
 		    tbl->fixed_width[tbl->col] = w;
 	    }
 #endif
-	    setwidth0(tbl, mode);
+	    setwidth0(tbl);
 	    clearcontentssize(tbl, mode);
 	}
 	break;
@@ -3183,8 +3254,9 @@ feed_table(struct table *tbl, char *line, struct table_mode *mode,
 			break;
 		    default:
 			r = conv_entity(ec);
-			if (r != NULL && strlen(r) == 1 &&
-			    ec == (unsigned char)*r) {
+			if (!r || !*r)
+			    break;
+			if (strlen(r) == 1 && ec == (unsigned char)*r) {
 			    Strcat_char(tmp, *r);
 			    break;
 			}
@@ -3286,15 +3358,19 @@ pushTable(struct table *tbl, struct table *tbl1)
     int col;
     int row;
 
+    if (tbl->ntable < 0 || tbl->ntable >= MAX_TABLE_N_LIMIT)
+	return;
     col = tbl->col;
     row = tbl->row;
 
     if (tbl->ntable >= tbl->tables_size) {
 	struct table_in *tmp;
 	tbl->tables_size += MAX_TABLE_N;
+	if (tbl->tables_size <= 0 || tbl->tables_size > MAX_TABLE_N_LIMIT)
+	    tbl->tables_size = MAX_TABLE_N_LIMIT;
 	tmp = New_N(struct table_in, tbl->tables_size);
 	if (tbl->tables)
-	    bcopy(tbl->tables, tmp, tbl->ntable * sizeof(struct table_in));
+	    memmove(tmp, tbl->tables, tbl->ntable * sizeof(struct table_in));
 	tbl->tables = tmp;
     }
 
@@ -3352,7 +3428,7 @@ correct_table_matrix2(struct table *t, int col, int cspan, double s, double b)
 }
 
 static void
-correct_table_matrix3(struct table *t, int col, char *flags, double s,
+correct_table_matrix3(struct table *t, int col, const char *flags, double s,
 		      double b)
 {
     int i, j;
@@ -3379,7 +3455,7 @@ correct_table_matrix3(struct table *t, int col, char *flags, double s,
 }
 
 static void
-correct_table_matrix4(struct table *t, int col, int cspan, char *flags,
+correct_table_matrix4(struct table *t, int col, int cspan, const char *flags,
 		      double s, double b)
 {
     int i, j;
@@ -3408,17 +3484,12 @@ correct_table_matrix4(struct table *t, int col, int cspan, char *flags,
 static void
 set_table_matrix0(struct table *t, int maxwidth)
 {
-    int size = t->maxcol + 1;
+    size_t size = t->maxcol + 1;
     int i, j, k, bcol, ecol;
     int width;
     double w0, w1, w, s, b;
-#ifdef __GNUC__
-    double we[size];
-    char expand[size];
-#else				/* not __GNUC__ */
-    double we[MAXCOL];
-    char expand[MAXCOL];
-#endif				/* not __GNUC__ */
+    double *we = GC_malloc(size * sizeof(*we));
+    char *expand = GC_malloc(size * sizeof(*expand));
     struct table_cell *cell = &t->cell;
 
     w0 = 0.;
@@ -3490,7 +3561,7 @@ set_table_matrix0(struct table *t, int maxwidth)
     }
 }
 
-void
+static void
 check_relative_width(struct table *t, int maxwidth)
 {
     int i;
