@@ -1,9 +1,8 @@
-/* $Id: fb_imlib2.c,v 1.10 2004/08/04 17:32:28 ukai Exp $ */
+/* vi: set sw=4 ts=8 ai sm noet : */
 /**************************************************************************
                 fb_imlib2.c 0.3 Copyright (C) 2002, hito
  **************************************************************************/
 
-#include <X11/Xlib.h>
 #include <Imlib2.h>
 #include "fb.h"
 #include "fb_img.h"
@@ -12,7 +11,7 @@ static void draw(FB_IMAGE * img, Imlib_Image image);
 static Imlib_Image resize_image(Imlib_Image image, int width, int height);
 
 void
-fb_image_init()
+fb_image_init(void)
 {
     return;
 }
@@ -85,6 +84,9 @@ draw(FB_IMAGE * img, Imlib_Image image)
     imlib_context_set_image(image);
     data = imlib_image_get_data_for_reading_only();
 
+    if (data == NULL)
+	return;
+
     for (j = 0; j < img->height; j++) {
 	offset = img->width * j;
 	for (i = 0; i < img->width; i++) {
